/************************************************************************

   File:		PGM.cpp

   Author(s):		Robert Sim

   Created:		 6 Dec 1997

   Last Revision:	$Date: 1998/04/08 11:35:47 $

   Description:	

   $Revision: 1.3 $

   $Log: PGM.cpp,v $
   Revision 1.3  1998/04/08 11:35:47  robbo
   *** empty log message ***

   Revision 1.2  1998/04/07 23:38:27  simra
   Changes to compile on sgi w/ CC. It works- huge speedup!

   Revision 1.1.1.1  1998/02/24 02:12:11  robbo
   First fie import.


	
   Copyright (c) 1997 by Robert Sim, Centre for Intelligent Machines,
   McGill University, Montreal, QC.  Please see the copyright notice
   included in this distribution for full details.

 ***********************************************************************/

#define PGM_CPP
#include "PGM.h"
PGM::PGM(int c, int r, gray d) {
  cols=c;
  rows=r;
  depth=d;
  if (rows*cols) {
    data = pgm_allocarray(cols, rows);
    int y,x;
    for (y=0; y<rows; y++) {
      for (x=0; x<cols; x++) {
	data[y][x]=0;
      }
    }
  } else {
    data=0;
  }
}

PGM::PGM(const PGM &I) {
  rows=I.rows;
  cols=I.cols;
  depth=I.depth;
  draw_colour=I.draw_colour;
  if (rows*cols) {
    data = pgm_allocarray(cols, rows);
    int y,x;
    for (y=0; y<rows; y++) {
      for (x=0; x<cols; x++) {
	data[y][x]=I.data[y][x];
      }
    }
  } else {
    data=0;
  }
  
}

PGM&
PGM::operator=(const PGM& I) {
  if (data) {
    pgm_freearray(data,rows);
  }
  rows=I.rows;
  cols=I.cols;
  depth=I.depth;
  draw_colour=I.draw_colour;
  if (rows*cols) {
    data = pgm_allocarray(cols, rows);
    int y,x;
    for (y=0; y<rows; y++) {
      for (x=0; x<cols; x++) {
	data[y][x]=I.data[y][x];
      }
    }
  } else {
    data=0;
  }
  return *this;
}

PGM::~PGM() {
  if (data) {
    pgm_freearray(data,rows);
  }
}

bool
PGM::load(char * filename) {
  if (data) {
    pgm_freearray(data,rows);
  }
  FILE *FPI=0;
  if (filename) {
    verbprintf(1, "PGM.load %s\n", filename); 
    FPI=pm_openr(filename);
    if (!FPI) {
      perror(filename);
      return false;
    }
  } else {
    FPI=stdin;
  }
  data=pgm_readpgm(FPI, &cols, &rows, &depth);

  if (filename)
    pm_close(FPI);
  
  return true;
}


bool
PGM::save(char * filename) const {
  FILE *FPO=0;

  if (filename) {
    verbprintf(1, "PGM.save %s\n", filename);
    FPO=pm_openw(filename);
    if (!FPO) {
      perror(filename);
      return false;
    }
  } else {
    FPO=stdout;
  }
  pgm_writepgm(FPO, data, cols, rows, depth, 0);
  if (filename)
    pm_close(FPO);
  return true;
}

matrix*
PGM::makeMatrix(const PGM &inPGM)
{
  int numRows = inPGM.Rows();
  int numCols = inPGM.Cols();

  PRECISION **values;

  values = new PRECISION*[numRows];  
  for (int i=0; i<numRows; i++)
    values[i] = new PRECISION[numCols];

  // So the first [] leads to the row, and the second [] leads to
  // the column

  for (int i=0; i<numRows; i++)
    for (int j=0; j<numCols; j++)
      values[i][j] = (PRECISION)inPGM(j,i);

  matrix *m = new matrix(numRows, numCols, values);

  for (int i=0; i<numRows; i++)
    {
      delete[] values[i];
      values[i] = NULL;
    }

  delete []values;

  values = NULL;

  return m;  
}

void 
PGM::drawWindow(PGM &inPGM, int x1, int y1, int x2, int y2)
{


  if (x2 < x1 || y2 < y1)
    {
      fprintf(stderr,"Window doesn't make sense\n");
      return;
    }

  // this is all hacked
  for (int i=x1; i<=x2; i++)
    inPGM(i,y1) = 255;

  for (int i=x1; i<=x2; i++)
    inPGM(i,y2) = 255;

  for (int i=y1; i<=y2; i++)
    inPGM(x1,i) = 255;

  for (int i=y1; i<=y2; i++)
    inPGM(x2,i) = 255;

}

void 
PGM::drawWindow(PGM &inPGM, interestPt &P, int width, int height)
{
  int x1, x2, y1, y2;
  if (!getCorners(inPGM,P,width,height,&x1,&y1,&x2,&y2)) 
    return;
  drawWindow(inPGM,x1,y1,x2,y2);
}

matrix* 
PGM::makeMatrix(const PGM &inPGM, int x1, int y1, int x2, int y2)
{


  if (x2 < x1 || y2 < y1)
    {
      fprintf(stderr,"Window doesn't make sense\n");
      return NULL;
    }

  int colIndex=x1;
  int rowIndex=y1;
  int numRows = (y2 - y1)+1;
  int numCols = (x2 - x1)+1;


  PRECISION **values;

  values = new PRECISION*[numRows]; 
  for (int i=0; i<numRows; i++)
    values[i] = new PRECISION[numCols];


  for (int i=0; i<numRows; i++)
    {
      for (int j=0; j<numCols; j++)
	{
	  //	  printf("colindex=%d, rowindex=%d\n",colIndex,rowIndex);
	  values[i][j] = (PRECISION)inPGM(colIndex,rowIndex);
	  colIndex++;
	  if (colIndex > x2)
	    colIndex = x1;
	}
      rowIndex++;
      if (rowIndex > y2)
	rowIndex = y1;
    }

  matrix *m = new matrix(numRows, numCols, values);

  for (int i=0; i<numRows; i++)
    {
      delete[] values[i];
      values[i] = NULL;
    }

  delete []values;

  values = NULL;

  return m;  

}


// Given an interest point, and a width and a height, this returns a matrix of size widthXheight.
matrix* 
PGM::makeMatrix(const PGM &inPGM, const interestPt &P, int width, int height)
{
  int x1, x2, y1, y2;
  if (!getCorners(inPGM,P,width,height,&x1,&y1,&x2,&y2)) 
    return NULL;
  matrix *m = makeMatrix(inPGM, x1, y1, x2, y2);
  return m;
}


// Given an interest point, and a width and a height, this returns a matrix of size widthXheight.
matrix* 
PGM::makeGaussianMatrix(const PGM &inPGM, const interestPt &P, int width, int height, PRECISION sigma=0)
{

  int x1,y1,x2,y2;

  if (sigma<=0) {
    if (width > height)
      sigma = width/4.0;
    else
      sigma = height/4.0;
  }
   
  getCorners(inPGM,P,width,height,&x1,&y1,&x2,&y2);

  matrix *m = makeMatrix(inPGM, x1, y1, x2, y2);
  
  // Now I want to multiply it by a Gaussian
  // Allocate space in 2d
  PRECISION **gaussianVals;
  gaussianVals = new PRECISION*[height];
  for (int i=0; i<height; i++)
    gaussianVals[i] = new PRECISION[width]; 
  
  // fill in the values and maintain the largest value
  PRECISION largest=0;

  int rowIndex=0;
  int colIndex=0;
  // first compute the gaussian with given sigma and keep the max value
  int imin=-(height/2);
  int imax=((height/2) -1 + (height%2));
  int jmin=-(width/2);
  int jmax=((width/2) -1 + (width%2));
  for (int i=imin; i<=imax; i++)
    {
      for (int j=jmin; j<=jmax; j++)
	{
	  gaussianVals[rowIndex % height][colIndex % width]=pow(E_VAL,-(((i*i)+(j*j)) / (2*sigma*sigma)));
	  if (gaussianVals[rowIndex % height][colIndex % width] >= largest)
	    largest = gaussianVals[rowIndex % height][colIndex % width];
	  colIndex++;
	}
      rowIndex++;
    }

  if (largest != 1) {
    cout << "Dividing through by " << largest << "\n";
    // need to normalise, divide by largest value
    for (int i=0; i<height; i++)
      for (int j=0; j<width; j++)
	gaussianVals[i][j] /= largest;
  }


  // multiply each value in the matrix by the gaussianVals.  Note that there are 3 times more cols due
  // to rgb so every three values of m get multiplied by the same value in gaussianVals
  rowIndex=0;
  colIndex=0;  
  for (int i=0; i<height; i++) 
    {
      for (int j=0; j<width; j++)
	{	
	  (*m)[i][j] *= gaussianVals[rowIndex % height][colIndex % width];
	  colIndex++;
	}
      rowIndex++;	  
    }


  // clean up the gaussian mem
  for (int i=0; i<height; i++)
    {
      delete[] gaussianVals[i];
      gaussianVals[i] = NULL;
    }

  delete []gaussianVals;

  gaussianVals = NULL;

  return m;
}

int
PGM::getCorners(const PGM &inPGM, const interestPt &P, int width, int height, int* a1, int* b1, int* a2, int* b2)
{

  int numRows = inPGM.Rows();
  int numCols = inPGM.Cols();

  int x, y, x1, x2, y1, y2;

  x=P.getx();
  y=P.gety();

  // Check window size
  if (width >= numCols || height >= numRows)
    {
      fprintf(stderr,"SubMatrix too big\n");
      return 0;
    }

  // Should this be strict or not!!!!!!
  if (x < 0 || x > numCols || y < 0 || y > numRows)
    {
      fprintf(stderr,"InterestPoint not within image\n");
      return 0;
    }

  if (width % 2 == 0)
    {
      x1=(int)x-(width/2);
      x2=(int)x+(width/2) - 1;
    }
  else
    {
      x1=(int)x-(width/2);
      x2=(int)x+(width/2);
    }

  if (height % 2 == 0)
    {
      y1=(int)y-(height/2);
      y2=(int)y+(height/2) - 1;
    }
  else
    {
      y1=(int)y-(height/2);
      y2=(int)y+(height/2);
    }

  if (x1 < 0) {
    x1 = 0;
    x2 = width-1;
  }

  if (y1 < 0) {
    y1 = 0;
    y2 = height-1;
  }

  if (x2 >= numCols) {
    x2 = numCols - 1;
    x1 = (numCols - 1) - (width-1);
  }

  if (y2 >= numRows) {
    y2 = numRows - 1;
    y1 = (numRows - 1) - (height-1);
  }

  *a1=x1;
  *b1=y1;
  *a2=x2;
  *b2=y2;

  return 1;
}


PGM::PGM(matrix& m) {
  rows=m.dim1();
  cols=m.dim2();
  depth=255;
  draw_colour=-200;
  if (rows*cols) {
    data = pgm_allocarray(cols, rows);
  } else {
    data=0;
  }
  PRECISION min=0, max=0;
  bool set=false;
  int y,x;
  for (y=0; y<rows; y++) {
    for (x=0; x<cols; x++) {
      //      if (m(y,x)) {
      if (!set) {
	min=max=m(y,x);
	set=true;
      } else {
	if (min>=m(y,x)) min=m(y,x);
	if (max<=m(y,x)) max=m(y,x);
	//      }
      }
    }
  }
  verbprintf(2, "PGM: Max: %f Min: %f\n", max,min);
  for (y=0; y<rows; y++) {
    for (x=0; x<cols; x++) {
      if (m(y,x))
	data[y][x]=(gray)((m(y,x)-min)*255.0/(max-min));
      else data[y][x]=0;
    }
  }
}

// PGM 
// PGM::flip() {
//   PGM p(*this);
//   for (int y=0; y<rows; y++) {
//     for (int x=0; x<cols; x++) {
//       p(x,rows-y-1)=data[y][x];
//     }
//   }
//   return p;
// }














