/************************************************************************

   File:		PPM.C

   Author(s):		Deeptiman Jugessur

   Created:		 4 Jun 1998

   Last Revision:	$Date$

   Description:	

   $Revision$

   $Log$

	
   Copyright (c) 1998 by Deeptiman Jugessur, Centre for Intelligent Machines,
   McGill University, Montreal, QC.  Please see the copyright notice
   included in this distribution for full details.

 ***********************************************************************/

#define PPM_C

/**********************
     Include Files
 **********************/
#include "PPM.h"

PPM::PPM()
{
  verbprintf(2,"Using default cons with NO init vals\n");
  cols=rows=0;
  depth=0;
  data=0;
}


// allocates space for the image and initialises the pixels to 0
PPM::PPM(int r, int c, pixval d) 
{
  verbprintf(2,"Using default cons with init vals\n");
  rows = r;
  cols = c;
  depth = d;
  if (rows*cols) {
      data = ppm_allocarray(cols, rows);
      int y,x;
      for (y=0; y<rows; y++) {
	for (x=0; x<cols; x++) {
	  data[y][x].r = 0;
	  data[y][x].g = 0;
	  data[y][x].b = 0;
	}
      }
  }
  else {
    data = 0;
  }
}

PPM::PPM(const PPM &I)
{

  verbprintf(2,"Using copy cons with init const PPM object\n");
  rows = I.rows;
  cols = I.cols;
  depth = I.depth;

  if (rows*cols) {
    data = ppm_allocarray(cols,rows);
    int y,x;
    for (y=0; y<rows; y++) {
      for (x=0; x<cols; x++) {
	data[y][x].r = I.data[y][x].r;
	data[y][x].g = I.data[y][x].g;
	data[y][x].b = I.data[y][x].b;
      }
    }
  }
  else {
    data = 0;
  }
}


PPM&
PPM::operator=(const PPM& I) {
  
  verbprintf(2,"Operator overloaded = for copying used\n");

  if (data) {
    ppm_freearray(data,rows);
  }
  rows=I.rows;
  cols=I.cols;
  depth=I.depth;

  if (rows*cols) {
    data = ppm_allocarray(cols, rows);
    int y,x;
    for (y=0; y<rows; y++) {
      for (x=0; x<cols; x++) {
	data[y][x].r=I.data[y][x].r;
	data[y][x].g=I.data[y][x].g;
	data[y][x].b=I.data[y][x].b;
      }
    }
  } else {
    data=0;
  }
  return *this;
}


PPM::~PPM() {
  verbprintf(2,"Destructor called\n");
  if (data) {
    ppm_freearray(data,rows);
  }
}



bool
PPM::load(char * filename) {
  if (data) {
    ppm_freearray(data,rows);
  }
  FILE *FPI=0;
  if (filename) {
    verbprintf(1, "PPM.load %s\n", filename); 
    FPI=pm_openr(filename);
    if (!FPI) {
      perror(filename);
      return false;
    }
  } else {
    FPI=stdin;
  }
  data=ppm_readppm(FPI, &cols, &rows, &depth);

  if (filename)
    pm_close(FPI);
  
  return true;
}


bool
PPM::save(char * filename) const {
  FILE *FPO=0;

  if (filename) {
    verbprintf(1, "PPM.save %s\n", filename);
    FPO=pm_openw(filename);
    if (!FPO) {
      perror(filename);
      return false;
    }
  } else {
    FPO=stdout;
  }
  ppm_writeppm(FPO, data, cols, rows, depth, 0);
  if (filename)
    pm_close(FPO);
  return true;
}


void
PPM::printPixel(pixel inPixel)
{
  cout << "R=" << inPixel.r << " G=" << inPixel.g << " B=" << inPixel.b << "\n";
}


// This assumes that the input is 3 matrices, the first being the red
// values, the second the green values and the third the blue values, with their corresponding
// depth values (ie max value for colour).  It returns a PPM such that the values are made up
// of the three input matrices linearised according to the input depth values with the allowable
// range ie PPM_MAXMAXVAL, so scaled to 0..PPM_MAXMAXVAL from x..(rgb)Depth
PPM::PPM(matrix& mRed, pixval rDepth, matrix& mGreen, pixval gDepth, matrix& mBlue, pixval bDepth) 
{

  // First check that they are all the same size


  if (mRed.dim1() == mGreen.dim1() && mGreen.dim1() == mBlue.dim1() &&
      mRed.dim2() == mGreen.dim2() && mGreen.dim2() == mBlue.dim2())
    {
      // Everyting cool

      rows=mRed.dim1();
      cols=mRed.dim2();
      depth=PPM_MAXMAXVAL;
      if (rows*cols) {
	data = ppm_allocarray(cols, rows); 
	
      }
      else {
	data=0;
      }
      
      
      PRECISION minRed=0, maxRed=0, minGreen=0, maxGreen=0, minBlue=0, maxBlue=0;
      bool setRed=false, setGreen=false, setBlue=false;
      int y,x;
      for (y=0; y<rows; y++) 
	{
	  for (x=0; x<cols; x++) 
	    {
	      if (!setRed) 
		{
		  minRed=maxRed=mRed(y,x);
		  setRed=true;
		} 
	      else 
		{
		  if (minRed>mRed(y,x)) minRed=mRed(y,x);
		  if (maxRed<mRed(y,x)) maxRed=mRed(y,x);
		}
	      if (!setGreen) 
		{
		  minGreen=maxGreen=mGreen(y,x);
		  setGreen=true;
		} 
	      else 
		{
		  if (minGreen>mGreen(y,x)) minGreen=mGreen(y,x);
		  if (maxGreen<mGreen(y,x)) maxGreen=mGreen(y,x);
		}
	      if (!setBlue) 
		{
		  minBlue=maxBlue=mBlue(y,x);
		  setBlue=true;
		} 
	      else 
		{
		  if (minBlue>mBlue(y,x)) minBlue=mBlue(y,x);
		  if (maxBlue<mBlue(y,x)) maxBlue=mBlue(y,x);
		}
	    } // inner for
	} // outer for
      //printf("PPM: MaxRed: %f MinRed: %f\n", maxRed,minRed);
      //printf("PPM: MaxGreen: %f MinGreen: %f\n", maxGreen,minGreen);
      //printf("PPM: MaxBlue: %f MinBlue: %f\n", maxBlue,minBlue);


      for (y=0; y<rows; y++) 
	{
	  for (x=0; x<cols; x++) 
	    {
	      if (rDepth != 0) 
		{	     
		  if (minRed != maxRed)
		    data[y][x].r=(gray)((mRed(y,x)-minRed)*depth/rDepth);
		  else 
		    data[y][x].r=(gray)(mRed(y,x)*depth/rDepth);
		} 
	      else
		data[y][x].r=0;

	      if (gDepth != 0)
		{
		  if (minGreen != maxGreen)
		    data[y][x].g=(gray)((mGreen(y,x)-minGreen)*depth/gDepth);
		  else 
		    data[y][x].g=(gray)(mGreen(y,x)*depth/gDepth);
		}
	      else
		data[y][x].g=0;
		  
	      if (bDepth != 0)
		{
		  if (minBlue != maxBlue)
		    data[y][x].b=(gray)((mBlue(y,x)-minBlue)*depth/bDepth);
		  else 
		    data[y][x].b=(gray)(mBlue(y,x)*depth/bDepth);
		}
	      else
		data[y][x].b=0;
	    }
	}
    }
  else // first if
    {
      verbprintf(1,"INPUT MATRICES ARE DIFFERENT SIZES");
      return;
    }
  
}


void 
PPM::drawWindow(PPM &inPPM, int x1, int y1, int x2, int y2)
{


  if (x2 < x1 || y2 < y1)
    {
      fprintf(stderr,"Window doesn't make sense\n");
      return;
    }

  // this is all hacked
  //  cout << y1 << " from " << x1 << " to " << x2 << "\n";
  for (int i=x1; i<=x2; i++) {
    // white line at correct spot
    inPPM(i,y1).r = 255;
    inPPM(i,y1).g = 255;
    inPPM(i,y1).b = 255;

    // black line below
    inPPM(i,y1+1).r = 0;
    inPPM(i,y1+1).g = 0;
    inPPM(i,y1+1).b = 0;

    // black line above
    inPPM(i,y1-1).r = 0;
    inPPM(i,y1-1).g = 0;
    inPPM(i,y1-1).b = 0;
  }

  //  cout << y2 << " from " << x1 << " to " << x2 << "\n";
  for (int i=x1; i<=x2; i++) {
    // white line at correct spot
    inPPM(i,y2).r = 255;
    inPPM(i,y2).g = 255;
    inPPM(i,y2).b = 255;

    // black line below
    inPPM(i,y2+1).r = 0;
    inPPM(i,y2+1).g = 0;
    inPPM(i,y2+1).b = 0;

    // black line above
    inPPM(i,y2-1).r = 0;
    inPPM(i,y2-1).g = 0;
    inPPM(i,y2-1).b = 0;
  }

  //  cout << x1 << " from " << y1 << " to " << y2 << "\n";
  for (int i=y1; i<=y2; i++) {
    // white line at correct spot
    inPPM(x1,i).r = 255;
    inPPM(x1,i).g = 255;
    inPPM(x1,i).b = 255;

    // black line besides
    inPPM(x1+1,i).r = 255;
    inPPM(x1+1,i).g = 255;
    inPPM(x1+1,i).b = 255;

    // black line besides
    inPPM(x1-1,i).r = 255;
    inPPM(x1-1,i).g = 255;
    inPPM(x1-1,i).b = 255;
  }

  //  cout << x2 << " from " << y1 << " to " << y2 << "\n";
  for (int i=y1; i<=y2; i++) {
    // white line at correct spot
    inPPM(x2,i).r = 255;
    inPPM(x2,i).g = 255;
    inPPM(x2,i).b = 255;

    // black line besides
    inPPM(x2+1,i).r = 255;
    inPPM(x2+1,i).g = 255;
    inPPM(x2+1,i).b = 255;

    // black line besides
    inPPM(x2-1,i).r = 255;
    inPPM(x2-1,i).g = 255;
    inPPM(x2-1,i).b = 255;
  }

}

void 
PPM::drawWindow(PPM &inPPM, interestPt &P, int width, int height)
{
  int x1, x2, y1, y2;
  if (!getCorners(inPPM,P,width,height,&x1,&y1,&x2,&y2)) 
    return;
  drawWindow(inPPM,x1,y1,x2,y2);
}


matrix*
PPM::makeMatrix(const PPM &inPPM)
{
  int colIndex=0;
  int numRows = inPPM.getNumRows();
  int numCols = inPPM.getNumCols();

  PRECISION **values;

  values = new PRECISION*[numRows];  
  for (int i=0; i<numRows; i++)
    values[i] = new PRECISION[numCols*3];  // To account for RGB

  // So the first [] leads to the row, and the second [] leads to
  // the column


  for (int i=0; i<numRows; i++)
    for (int j=0; j<numCols*3; j+=3)
      {
	values[i][j] = (PRECISION)inPPM(colIndex,i).r;
	values[i][j+1] = (PRECISION)inPPM(colIndex,i).g;
	values[i][j+2] = (PRECISION)inPPM(colIndex,i).b;
	colIndex = (colIndex + 1) % numRows;
      }

  matrix *m = new matrix(numRows, numCols*3, values);

  for (int i=0; i<numRows; i++)
    {
      delete[] values[i];
      values[i] = NULL;
    }

  delete []values;

  values = NULL;

  return m;  
}

matrix* 
PPM::makeMatrix(const PPM &inPPM, char colour)
{
  int numRows = inPPM.getNumRows();
  int numCols = inPPM.getNumCols();

  PRECISION **values;

  values = new PRECISION*[numRows];  
  for (int i=0; i<numRows; i++)
    values[i] = new PRECISION[numCols]; 

  // So the first [] leads to the row, and the second [] leads to
  // the column


  for (int i=0; i<numRows; i++)
    for (int j=0; j<numCols; j++)
      {
	switch (colour)
	  {
	  case 'r':
	    values[i][j] = (PRECISION)inPPM(j,i).r;
	    break;
	  case 'g':
	    values[i][j] = (PRECISION)inPPM(j,i).g;
	    break;
	  case 'b':
	    values[i][j] = (PRECISION)inPPM(j,i).b;
	    break;
	  default:
	    fprintf(stderr,"Unknown colour %c\n",colour);
	    return NULL;
	  }
      }

  matrix *m = new matrix(numRows, numCols, values);

  for (int i=0; i<numRows; i++)
    {
      delete[] values[i];
      values[i] = NULL;
    }

  delete []values;

  values = NULL;

  return m;  
}

 
matrix* 
PPM::makeMatrix(const PPM &inPPM, int x1, int y1, int x2, int y2)
{


  if (x2 < x1 || y2 < y1)
    {
      fprintf(stderr,"Window doesn't make sense\n");
      return NULL;
    }

  int colIndex=x1;
  int rowIndex=y1;
  int numRows = (y2 - y1)+1;
  int numCols = (x2 - x1)+1;

  /*
  printf("x1=%d,y1=%d,x2=%d,y2=%d\n",x1,y1,x2,y2);  
  printf("colIndex=%d, rowIndex=%d\n",colIndex,rowIndex);
  printf("numCols/3 = %d\tnumRows = %d\n",numCols,numRows);
  printf("ppmnumcols=%d, ppmnumrows=%d\n",inPPM.getNumCols(), inPPM.getNumRows());
  for (int i=0;i<inPPM.getNumRows(); i++)
    for (int j=0;j<inPPM.getNumCols(); j++)
      printf("(%d,%d) value = %lf\n",(PRECISION)inPPM(j,i).r);
  printf("!!!value = %lf\n",(PRECISION)inPPM(colIndex,rowIndex).r);
  */

  PRECISION **values;

  values = new PRECISION*[numRows]; 
  for (int i=0; i<numRows; i++)
    values[i] = new PRECISION[numCols*3]; // because of RGB


  for (int i=0; i<numRows; i++)
    {
      for (int j=0; j<numCols*3; j+=3)
	{
	  //	  printf("colindex=%d, rowindex=%d\n",colIndex,rowIndex);
	  values[i][j] = (PRECISION)inPPM(colIndex,rowIndex).r;
	  values[i][j+1] = (PRECISION)inPPM(colIndex,rowIndex).g;
	  values[i][j+2] = (PRECISION)inPPM(colIndex,rowIndex).b;
	  colIndex++;
	  if (colIndex > x2)
	    colIndex = x1;
	}
      rowIndex++;
      if (rowIndex > y2)
	rowIndex = y1;
    }

  matrix *m = new matrix(numRows, numCols*3, values);

  for (int i=0; i<numRows; i++)
    {
      delete[] values[i];
      values[i] = NULL;
    }

  delete []values;

  values = NULL;

  return m;  

}




// Given an interest point, and a width and a height, this returns a matrix of size widthXheight.
matrix* 
PPM::makeMatrix(const PPM &inPPM, const interestPt &P, int width, int height)
{
  int x1, x2, y1, y2;
  if (!getCorners(inPPM,P,width,height,&x1,&y1,&x2,&y2)) 
    return NULL;
  matrix *m = makeMatrix(inPPM, x1, y1, x2, y2);
  return m;
}


matrix* 
PPM::makeMatrix(const PPM &inPPM, const interestPt &P, int width, int height, char colour)
{
  int x1, x2, y1, y2;
  if (!getCorners(inPPM,P,width,height,&x1,&y1,&x2,&y2))
    return NULL;
  matrix *m = makeMatrix(inPPM, x1, y1, x2, y2, colour);
  return m;
}


// Given an interest point, and a width and a height, this returns a matrix of size widthXheight.
matrix* 
PPM::makeGaussianMatrix(const PPM &inPPM, const interestPt &P, int width, int height, PRECISION sigma=0)
{

  int x1,y1,x2,y2;

  if (sigma <= 0) {
    if (width > height)
      sigma = width/4.0;
    else
      sigma = height/4.0;
  }
  
  getCorners(inPPM,P,width,height,&x1,&y1,&x2,&y2);
  
  matrix *m = makeMatrix(inPPM, x1, y1, x2, y2);
  
  // Now I want to multiply it by a Gaussian
  // Allocate space in 2d
  PRECISION **gaussianVals;
  gaussianVals = new PRECISION*[height];
  for (int i=0; i<height; i++)
    gaussianVals[i] = new PRECISION[width]; 
  
  // fill in the values and maintain the largest value
  PRECISION largest=0;

  int rowIndex=0;
  int colIndex=0;
  // first compute the gaussian with given sigma and keep the max value
  int imin=-(height/2);
  int imax=((height/2) -1 + (height%2));
  int jmin=-(width/2);
  int jmax=((width/2) -1 + (width%2));
  for (int i=imin; i<=imax; i++)
    {
      for (int j=jmin; j<=jmax; j++)
	{
	  /*
	  printf("h=%d, w=%d, sigma = %lf (%d,%d) = %lf [%d,%d]\n",height,width,sigma,rowIndex % height,colIndex % width,
		 pow(E_VAL,-(((i*i)+(j*j)) / (2*sigma*sigma))),i,j); 
	  */
	  gaussianVals[rowIndex % height][colIndex % width]=pow(E_VAL,-(((i*i)+(j*j)) / (2*sigma*sigma)));
	  if (gaussianVals[rowIndex % height][colIndex % width] >= largest)
	    largest = gaussianVals[rowIndex % height][colIndex % width];
	  colIndex++;
	}
      rowIndex++;
    }

  if (largest != 1) {
    cout << "Dividing through by " << largest << "\n";
    // need to normalise, divide by largest value
    for (int i=0; i<height; i++)
      for (int j=0; j<width; j++)
	gaussianVals[i][j] /= largest;
  }


  // multiply each value in the matrix by the gaussianVals.  Note that there are 3 times more cols due
  // to rgb so every three values of m get multiplied by the same value in gaussianVals
  rowIndex=0;
  colIndex=0;  
  for (int i=0; i<height; i++) 
    {
      for (int j=0; j<width*3; j+=3)
	{	
	  (*m)[i][j] *= gaussianVals[rowIndex % height][colIndex % width];
	  (*m)[i][j+1] *= gaussianVals[rowIndex % height][colIndex % width];
	  (*m)[i][j+2] *= gaussianVals[rowIndex % height][colIndex % width];
	  colIndex++;
	}
      rowIndex++;	  
    }


  // clean up the gaussian mem
  for (int i=0; i<height; i++)
    {
      delete[] gaussianVals[i];
      gaussianVals[i] = NULL;
    }

  delete []gaussianVals;

  gaussianVals = NULL;


  return m;
}


matrix* 
PPM::makeMatrix(const PPM &inPPM, int x1, int y1, int x2, int y2, char colour)
{

  if (x2 < x1 || y2 < y1)
    {
      fprintf(stderr,"Window doesn't make sense\n");
      return NULL;
    }

  int colIndex=x1;
  int rowIndex=y1;
  int numRows = (y2 - y1)+1;
  int numCols = (x2 - x1)+1;

  PRECISION **values;

  values = new PRECISION*[numRows]; 
  for (int i=0; i<numRows; i++)
    values[i] = new PRECISION[numCols]; // because of RGB


  for (int i=0; i<numRows; i++)
    {
      for (int j=0; j<numCols; j++)
	{
	  switch (colour)
	    {
	    case 'r':
	      values[i][j] = (PRECISION)inPPM(colIndex,rowIndex).r;
	      break;
	    case 'g':
	      values[i][j] = (PRECISION)inPPM(colIndex,rowIndex).g;
	      break;
	    case 'b':
	      values[i][j] = (PRECISION)inPPM(colIndex,rowIndex).b;
	      break;
	    default:
	      fprintf(stderr,"Unknown colour %c\n",colour);
	      return NULL;
	    }
	  colIndex++;
	  if (colIndex > x2)
	    colIndex = x1;
	}
      rowIndex++;
      if (rowIndex > y2)
	rowIndex = y1;
    }


  matrix *m = new matrix(numRows, numCols, values);

  for (int i=0; i<numRows; i++)
    {
      delete[] values[i];
      values[i] = NULL;
    }

  delete []values;

  values = NULL;

  return m;  

}


int
PPM::getCorners(const PPM &inPPM, const interestPt &P, int width, int height, int* a1, int* b1, int* a2, int* b2)
{

  int numRows = inPPM.getNumRows();
  int numCols = inPPM.getNumCols();

  int x, y, x1, x2, y1, y2;

  x=P.getx();
  y=P.gety();

  // Check window size
  if (width >= numCols || height >= numRows)
    {
      fprintf(stderr,"SubMatrix too big\n");
      return 0;
    }

  // Should this be strict or not!!!!!!
  if (x < 0 || x > numCols || y < 0 || y > numRows)
    {
      fprintf(stderr,"InterestPoint not within image\n");
      return 0;
    }

  if (width % 2 == 0)
    {
      x1=(int)x-(width/2);
      x2=(int)x+(width/2) - 1;
    }
  else
    {
      x1=(int)x-(width/2);
      x2=(int)x+(width/2);
    }

  if (height % 2 == 0)
    {
      y1=(int)y-(height/2);
      y2=(int)y+(height/2) - 1;
    }
  else
    {
      y1=(int)y-(height/2);
      y2=(int)y+(height/2);
    }

  if (x1 < 0) {
    x1 = 0;
    x2 = width-1;
  }

  if (y1 < 0) {
    y1 = 0;
    y2 = height-1;
  }

  if (x2 >= numCols) {
    x2 = numCols - 1;
    x1 = (numCols - 1) - (width-1);
  }

  if (y2 >= numRows) {
    y2 = numRows - 1;
    y1 = (numRows - 1) - (height-1);
  }

  *a1=x1;
  *b1=y1;
  *a2=x2;
  *b2=y2;

  return 1;
}


