/**************************************************************************

   File:                PPM.h

   Author(s):           Deeptiman Jugessur

   Created:              4 Jun 1998

   Last Revision:       $Date$

   Description: 

   $Revision$

   $Log$

	
   Copyright (c) 1998 by Deeptiman Jugessur, Centre for Intelligent Machines,
   McGill University, Montreal, QC.  Please see the copyright notice
   included in this distribution for full details.

 **************************************************************************/


// NEED TO ADD MAKE FUNKY MATRIX FUNCTION


#ifndef PPM_H
#  define PPM_H
/**********************
     Include Files
 **********************/
#define E_VAL 2.7182818
#include <iostream.h>
#include <math.h>
#include "VObject.h"
#include "matrix.h"
#include "interestPt.h"
extern "C" {
#include <ppm.h>
}

class PPM : public VObject
{
protected:
  int rows;
  int cols;
  pixval depth;  // maxval for each color of any pixel
  pixel ** data;

public:
  /*  the const here means that if this class is instantiated in a
      const manner, then only these const functions can be accessed  */
  int getNumRows()  const { return rows;  }  
  int getNumCols()  const { return cols;  }
  pixval getDepth() const { return depth; }


  // data is a "horizontal" array of pointers such that one has to
  // first index the column then the row ie y then x to get a
  // cartesian feel ie x,y is cartesian coords ie x is col, y is row
  // (0,0)----->x
  //   |
  //   |
  //   |
  //   v
  //   y
  pixel &operator()(int x, int y) {return data[y][x]; }
  pixel operator()(int x, int y) const {return data[y][x]; }

  PPM();                                     // default constructor
  PPM(int r, int c, pixval d=PPM_MAXMAXVAL); // default constructor
  PPM(const PPM &I); // copy constructor used when "PPM y; PPM x = y;"
  PPM & operator=(const PPM &I); // copier used when "PPM x;PPM y;x=y;

  PPM(matrix &mRed, pixval rDepth, matrix &mgreen, pixval gDepth,
      matrix &mBlue, pixval bDepth);               // constructs a PPM
						   // given three
						   // matrices

                  
  pixel ** getData() { return data; }
  

  ~PPM();
  
  bool load(char * filename=0);
  bool save(char * filename=0) const;

  void printPixel(pixel inPixel);

  matrix* makeMatrix(const PPM &inPPM);
  matrix* makeMatrix(const PPM &inPPM, char colour);
  // x1,y1 is top left corner, x2,y2 is bottom right corner
  matrix* makeMatrix(const PPM &inPPM, int x1, int y1, int x2, int y2);
  matrix* makeMatrix(const PPM &inPPM, int x1, int y1, int x2, int y2, char colour);

  // Width and height should be odd numbers but can handle even's except the size of the matrix
  // obtained is not necessarily widthXheight
  matrix* makeMatrix(const PPM &inPPM, const interestPt &P, int width, int height);
  matrix* makeMatrix(const PPM &inPPM, const interestPt &P, int width, int height, char colour);
  matrix* makeGaussianMatrix(const PPM &inPPM, const interestPt &P, int width, int height, PRECISION sigma=0);
  int getCorners(const PPM &inPPM, const interestPt &c, int width, int height, int* a1, int* b1, int* a2, int* b2);
  void drawWindow(PPM &inPPM, int x1, int y1, int x2, int y2);
  void drawWindow(PPM &inPPM, interestPt &P, int width, int height);
};

#endif  /* PPM_H */




