/**************************************************************************

   File:                PrincipalComponents.cpp

   Author(s):           Robert Sim

   Created:             11 Nov 1997

   Last Revision:       $Date: 1998/07/21 14:20:50 $

   Description: 

   $Revision: 2.0 $

   $Log: PrincipalComponents.cpp,v $
   Revision 2.0  1998/07/21 14:20:50  simra
   Final Thesis submitted version.

   Revision 1.7  1998/05/19 13:03:53  robbo
   glibc2.0 setup

   Revision 1.6  1998/04/07 23:39:55  simra
   Changes to compile on sgi w/ CC. It works- huge speedup!

   Revision 1.5  1998/03/17 19:46:35  simra
   alpha compilation

   Revision 1.4  1998/03/16 20:47:28  simra
   Lots of work to get the cross platform stuff debugged- I think it's good now.
   Upped the number of iterations for LandmarkSet uncertainty to 100- there was
   too much variation in the output. I hope it all works.

   Revision 1.3  1998/03/10 21:49:22  simra
   Cross platform support- part I

   Revision 1.2  1998/02/24 13:35:07  robbo
   Changed eigentrack extensively to update the pca on the fly- best
   distinguishing the tracked sets. This required some major overhauling.
   Also changed some input file conventions. landmark outputs a clist,
   eigentrack inputs the same clist. eigenland is no longer useful for
   building the database.  The separate clist prevents the clist in the
   database from having to be recomputed every time (saving a lot of
   trouble.)

   Revision 1.1.1.1  1998/02/24 02:12:11  robbo
   First fie import.


	
   Copyright (c) 1997 by Robert Sim, Centre for Intelligent Machines,
   McGill University, Montreal, QC.  Please see the copyright notice
   included in this distribution for full details.

 **************************************************************************/

#define PRINCIPALCOMPONENTS_CPP
#include "PrincipalComponents.h"
extern "C" {
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <math.h>
#include <errno.h>
#include <values.h>
}
PrincipalComponents::PrincipalComponents() {
  K=N=0;
  normalized=false;
}

PrincipalComponents::PrincipalComponents(const matrix& in_pc, 
					 const vector& in_avg,
					 bool in_normalized) 
{
  pc=in_pc;
  average=in_avg;
  normalized=in_normalized;
  K=pc.dim1();
  N=pc.dim2();
  assert(average.dim()==N);
}
PrincipalComponents::PrincipalComponents(const matrix& in_pc, 
					 const vector& in_avg, 
					 const vector& evals,
					 bool in_normalized) {
  pc=in_pc;
  average=in_avg;
  ev=evals;
  K=pc.dim1();
  N=pc.dim2();
  normalized=in_normalized;
  assert(average.dim()==N && ev.dim()==K);
}

PrincipalComponents::PrincipalComponents(const PrincipalComponents& P) {
  pc=P.pc;
  average=P.average;
  ev=P.ev;
  K=P.K;
  N=P.N;
  normalized=P.normalized;
}

PrincipalComponents &
PrincipalComponents::operator=(const PrincipalComponents& P) {
  pc=P.pc;
  average=P.average;
  ev=P.ev;
  K=P.K;
  N=P.N;
  normalized=P.normalized;
  return *this;
}

// all the ifdefs handle endianness
bool 
PrincipalComponents::read(istream& I) {
  char token[80];
  if (nextGMLToken(I,token,80)) {
    verbprintf(5, "Token: %s\n", token);
    if (strcmp(token, "<PrincipalComponents>")) {
      cerr << "Failed PrincipalComponents read"<<endl;
      return false;
    }
    K=N=0;
    while (nextGMLToken(I,token,80)) {
      verbprintf(5, "Token: %s\n", token);
      if (!strcmp(token, "K")) {
	I>>K;
	closeGMLField(I);
      }	else if (!strcmp(token, "N")) {
	I>>N;
	closeGMLField(I);
      }	else if (!strcmp(token, "Normalized")) {
	normalized=true;
	closeGMLField(I);
      }	else if (!strcmp(token, "Eigenvalues")) {
	if (!K||!N) {
	  cerr<<"Missing K or N"<<endl;
	  return false;
	}
	ev=vector(K);
	I.get();
	for (int i=0; i<K; i++) {
	  char v1[sizeof(double)];
	  char v2[sizeof(double)];
	  I.read(v1,sizeof(double));
#if !(defined i386)
	  //(defined sgi || defined sun)
	  char * swap1=v1, *swap2=v2+sizeof(double)-1;
	  for (unsigned int fix=0; fix<sizeof(double); fix++) {
	    *swap2=*swap1;
	    swap2--;swap1++;
	  }
#else 
	  bcopy(v1,v2,sizeof(double));
#endif
	  ev[i]=*(double*)v2;
	}
	closeGMLField(I);
      }	else if (!strcmp(token, "Average")) {
	if (!K||!N) {
	  cerr<<"Missing K or N"<<endl;
	  return false;
	}
	average=vector(N);
	I.get();
	for (int i=0; i<N; i++) {
	  char v1[sizeof(double)];
	  char v2[sizeof(double)];
	  I.read(v1,sizeof(double));
#if !(defined i386)
	  //(defined sgi || defined sun)
	  char * swap1=v1, *swap2=v2+sizeof(double)-1;
	  for (unsigned int fix=0; fix<sizeof(double); fix++) {
	    *swap2=*swap1;
	    swap2--;swap1++;
	  }
#else 
	  bcopy(v1,v2,sizeof(double));
#endif
	  average[i]=*(double*)v2;
	}
	closeGMLField(I);
      }	else if (!strcmp(token, "Eigenvectors")) {
	if (!K||!N) {
	  cerr<<"Missing K or N"<<endl;
	  return false;
	}
	pc=matrix(K,N);
	I.get();
	for (int k=0; k<K; k++) {
	  for (int n=0; n<N; n++) {
	    char v1[sizeof(double)];
	    char v2[sizeof(double)];
	    I.read(v1,sizeof(double));
#if !(defined i386)
	  //(defined sgi || defined sun)
	    char * swap1=v1, *swap2=v2+sizeof(double)-1;
	    for (unsigned int fix=0; fix<sizeof(double); fix++) {
	      *swap2=*swap1;
	      swap2--;swap1++;
	    }
#else 
	    bcopy(v1,v2,sizeof(double));
#endif
	    pc(k,n)=*(double*)v2;
	  }
	}
	closeGMLField(I);
      } else {
	cerr<<"Unknown token--ignored"<<endl;
	closeGMLField(I);
      }
    }
  }
  return true;
}

istream& operator>>(istream& I, PrincipalComponents& pc) {
  pc.read(I);
  return I;
}

// All the ifdefs handle endianness.
ostream& operator<<(ostream& O, const PrincipalComponents& pc) {
  O <<"(<PrincipalComponents> "
    <<"(K "<<pc.K<<")"
    <<"(N "<<pc.N<<")";
  if (pc.K) {
    O <<"(Eigenvalues ";
    for (int i=0; i<pc.K; i++) {
      double val=pc.ev[i];
#if defined i386
      O.write(&val,sizeof(double));
#else
      char * pchar=(char *)&val+sizeof(double)-1;
      for (unsigned int fix=0; fix<sizeof(double); fix++) {
	O.write(pchar, sizeof(char));
	pchar--;
      }
#endif
    }
    O <<")";
  }
  if (pc.N) {
    O <<"(Average ";
    for (int i=0; i<pc.N; i++) {
      double val=pc.average[i];
#if defined i386
      //#if !(defined sgi || defined sun)
      O.write(&val,sizeof(double));
#else
      char * pchar=(char *)&val+sizeof(double)-1;
      for (unsigned int fix=0; fix<sizeof(double); fix++) {
	O.write(pchar, sizeof(char));
	pchar--;
      }
#endif
    }
    O <<")";
  }
  if (pc.normalized) {
    O <<"(Normalized )"<<endl;
  }
  if (pc.K && pc.N) {
    O <<"(Eigenvectors ";
    for (int k=0; k<pc.K; k++) {
      for (int n=0; n<pc.N; n++) {
#if defined i386
	//#if !(defined sgi || defined sun)
	double d=pc.pc(k,n);
	O.write(&d,sizeof(double));
#else
	double d=pc.pc(k,n);
	char * pchar=(char *)(&d)+sizeof(double)-1;
	for (unsigned int fix=0; fix<sizeof(double); fix++) {
	  O.write(pchar, sizeof(char));
	  pchar--;
	}
#endif
      }
    }
    O <<")"<<endl;
  }
  O <<")";
  return O;
}


// in_vector=N out_vector=K
bool
PrincipalComponents::Project(const vector& in_vector, vector& out_vector) const {
  assert(in_vector.dim()==N);
  assert(out_vector.dim()<=K);
  
  //  float mag=0;
  // normalize the input nrvector
  /*  for (int i=1; i<=N; i++)
      mag+=in_vector(i)*in_vector(i);
      mag=sqrt(mag);
      for (int i=1; i<=N; i++) 
      in_vector(i)/=mag;
      */
  vector comp_vector;
  if (normalized)
    comp_vector=in_vector.norm()-average;
  else 
    comp_vector=in_vector-average;
  // project
  //out_vector=pc*in_vector;
  for (int i=0; i<out_vector.dim(); i++) {
    out_vector[i]=pc.row(i)*comp_vector;
  }
  return true;
}


// in_matrix=MxN, out_matrix = MxK
bool 
PrincipalComponents::Project(const matrix& in_matrix, matrix& out_matrix) const {
  //  cout << "In matrix: " << in_matrix.dim1() << " by " << in_matrix.dim2() << endl;
  //  cout << "Out matrix: " << out_matrix.dim1() << " by " << out_matrix.dim2() << endl;
  //  cout << "K = " << K << endl;
  assert(in_matrix.dim2()==N);
  assert(out_matrix.dim2()<=K && out_matrix.dim1()==in_matrix.dim1());
  
  matrix comp_matrix=in_matrix;
  for (int r=0; r<comp_matrix.dim1(); r++) { 
    // normalize
    if (normalized)
      comp_matrix.row(r)=comp_matrix.row(r).norm()-average;
    else 
      comp_matrix.row(r)=comp_matrix.row(r)-average;
    //    in_matrix.row(r).print();
  }    
  //pc.row(0).print();
  //  matrix out=pc*in_matrix.trans();
  for (int c=0; c<comp_matrix.dim1(); c++) {  // upper bound is == M
    for (int r=0; r<out_matrix.dim2(); r++) { // upper bound is <= K
      out_matrix(c,r)=pc.row(r)*comp_matrix.row(c);
    }
  }
  //  out_matrix=out.trans();
  return true;
};


vector 
PrincipalComponents::Reconstruct(const vector& in_vector) const {
  assert(in_vector.dim()<=K);
  vector result(N);
  for (int i=0; i<in_vector.dim(); i++) {
    result+=pc.row(i)*in_vector[i];
  }
  return result+average;
}


PGM
PrincipalComponents::draw_components(int num) {
  int i,j, k,l;
  if (num<0) num=K;
  double maxcomp=-MAXDOUBLE;
  double mincomp=MAXDOUBLE;
  int wsize=(int)sqrt(N);
  PGM components(wsize,num*wsize);
  for(k=0;k<num;k++)
    for(l=0;l<N;l++) {
      if (pc(k,l)>maxcomp)
	maxcomp=pc(k,l);
      if (pc(k,l)<mincomp)
	mincomp=pc(k,l);
    }
  for (k=0; k<num ;k++)
    for (i=0; i<wsize; i++)
      for (j=0; j<wsize; j++)
	components(i,k*wsize+j) = 
	  (gray) ((pc(k,i*wsize+j)-mincomp) * 255.0 / (maxcomp-mincomp));
  return components;
}

