/**************************************************************************

   File:                PrincipalComponents.h

   Author(s):           Robert Sim

   Created:             11 Nov 1997

   Last Revision:       $Date: 1998/07/21 14:20:51 $

   Description: 

   $Revision: 2.0 $

   $Log: PrincipalComponents.h,v $
   Revision 2.0  1998/07/21 14:20:51  simra
   Final Thesis submitted version.

   Revision 1.3  1998/04/07 23:39:57  simra
   Changes to compile on sgi w/ CC. It works- huge speedup!

   Revision 1.2  1998/02/24 13:35:07  robbo
   Changed eigentrack extensively to update the pca on the fly- best
   distinguishing the tracked sets. This required some major overhauling.
   Also changed some input file conventions. landmark outputs a clist,
   eigentrack inputs the same clist. eigenland is no longer useful for
   building the database.  The separate clist prevents the clist in the
   database from having to be recomputed every time (saving a lot of
   trouble.)

   Revision 1.1.1.1  1998/02/24 02:12:11  robbo
   First fie import.


	
   Copyright (c) 1997 by Robert Sim, Centre for Intelligent Machines,
   McGill University, Montreal, QC.  Please see the copyright notice
   included in this distribution for full details.

 **************************************************************************/

#ifndef PRINCIPALCOMPONENTS_H
#  define PRINCIPALCOMPONENTS_H
#include <iostream.h>
#include <math.h>
#include "matrix.h"
#include "vector.h"
#include "VObject.h"
#include "PGM.h"
#include "utilityclasses.h"

class PrincipalComponents : public VObject, public GMLObject {

protected:
  matrix pc;  // KxN
  vector average;
  vector ev; // eigenvalues
  int K;
  int N;
  bool normalized;
public:
  PrincipalComponents();
  PrincipalComponents(const matrix& in_pc, 
		      const vector& in_avg,
		      bool in_normalized=false);
  PrincipalComponents(const matrix& in_pc, const vector& in_avg, 
		      const vector& evals, bool in_normalized=false);
  PrincipalComponents(const PrincipalComponents& pc);
  PrincipalComponents& operator=(const PrincipalComponents &P);

  matrix getPc() const {return pc;}
  bool read(istream& I);
  void read() { read(cin); }
  friend ostream& operator<<(ostream& O, const PrincipalComponents& pc);
  friend istream& operator>>(istream& I, PrincipalComponents& pc);

  // Both of these are destructive on the arguments-- returns in_ normalized
  bool Project(const vector& in_vector, vector& out_vector) const; // the projection of in_vector onto pc;
  bool Project(const matrix& in_matrix, matrix& out_matrix) const; // the projection of in_nrmatrix onto pc;
  vector Reconstruct(const vector& in_vector) const;
  int GetK() const {return K;}
  int GetN() const {return N;}
  vector GetAverage() const {return average;}
  bool IsNormalized() const {return normalized;}
  vector GetEigenvalues() const {return ev;}
  PGM draw_components(int num=-1);
};
#endif  /* PRINCIPALCOMPONENTS_H */







