/**************************************************************************

   File:                fft.h

   Author(s):           Deeptiman Jugessur

   Created:             13 Jul 1999

   Last Revision:       $Date$

   Description: 

   $Revision$

   $Log$

	
   Copyright (c) 1999 by Deeptiman Jugessur, Centre for Intelligent Machines,
   McGill University, Montreal, QC.  Please see the copyright notice
   included in this distribution for full details.

 **************************************************************************/

#ifndef FFT_H
#  define FFT_H

#include <stdio.h>
#include <math.h>
#include "matrix.h"

extern "C" {
#include "nrutil.h"
}


class fft 
{
private:
  matrix *re;
  matrix *im;
  matrix *shiftRe;
  matrix *shiftIm;
  matrix *amplitude;
  matrix *phase;

  // all the nr stuff
  void fourn(float data[], unsigned long nn[], int ndim, int isign);
  void rlft3(float ***data, float **speq, unsigned long nn1, unsigned long nn2, unsigned long nn3, int isign);

public:
  // constructors and destructors
  fft();
  fft(const matrix& in);
  ~fft();
  void computeAmplitude();
  void computeLogAmplitude();
  void computePhase();
  void computeAmplitudeAndPhase();
  static double recoverPhase(matrix& M1, matrix& M2);

  // getter functions
  matrix *getReal() {return re;}
  matrix *getIm() {return im;}
  matrix *getShiftReal() {return shiftRe;}
  matrix *getShiftIm() {return shiftIm;}
  matrix *getAmplitude() {return amplitude;}
  matrix *getPhase() {return phase;}

  // setter functions
  void setReal(matrix *in) {re = in;}
  void setIm(matrix *in) {im = in;}
  void setShiftReal(matrix *in) {shiftRe = in;}
  void setShiftIm(matrix *in) {shiftIm = in;}
  void setAmplitude(matrix *in) {amplitude = in;}
  void setPhase(matrix *in) {phase = in;}
};

#endif  /* FFT_H */
