/************************************************************************

   File:		integer_vector.cpp

   Author(s):		Robert Sim

   Created:		10 Dec 1997

   Last Revision:	$Date: 1998/03/17 23:41:25 $

   Description:	

   $Revision: 1.4 $

   $Log: integer_vector.cpp,v $
   Revision 1.4  1998/03/17 23:41:25  simra
   alpha troubles.

   Revision 1.3  1998/03/17 19:46:36  simra
   alpha compilation

   Revision 1.2  1998/03/10 21:49:25  simra
   Cross platform support- part I

   Revision 1.1.1.1  1998/02/24 02:12:11  robbo
   First fie import.


	
   Copyright (c) 1997 by Robert Sim, Centre for Intelligent Machines,
   McGill University, Montreal, QC.  Please see the copyright notice
   included in this distribution for full details.

 ***********************************************************************/

#define INTEGER_VECTOR_CPP
#include "integer_vector.h"
extern "C" {
#include <string.h>
#include <assert.h>
#include <math.h>
  //#if !(defined i386)
#if (defined sun || defined sgi)
extern void bcopy(char *b1, char *b2, int length);
#endif

}
integer_vector::integer_vector() {
  v=0;
  d=0;
}

integer_vector::integer_vector(int n) {
  d=n;
  if (d) {
    v=new int[d];
    memset(v,0,d*sizeof(int));
  } else v=0;
}

integer_vector::integer_vector(const integer_vector& c) {
  d=c.d;
  v=new int[d];
  bcopy((char *)c.v,(char *)v,d*sizeof(int));
}

integer_vector::~integer_vector() {
  if (d) delete []v;
}

integer_vector&
integer_vector::operator=(const integer_vector& c) {
  if (d) delete []v;
  d=c.d;
  v=new int[d];
  bcopy((char *)c.v,(char *)v,d*sizeof(int));
  return *this;
}

int&
integer_vector::operator[](int i) {
  assert(i>=0 && i<d);
  return v[i];
}

int
integer_vector::operator[](int i) const {
  assert(i>=0 && i<d);
  return v[i];
}

int 
integer_vector::sqr_length() const {
  int mag=0;
  for (int i=0; i<d; i++) {
    mag+=v[i]*v[i];
  }
  return mag;
}

int 
integer_vector::length() const {
  return (int)sqrt(sqr_length());
}

integer_vector&
integer_vector::operator+=(const integer_vector& v1) {
  assert(d==v1.d);
  for (int i=0; i<d; i++) {
    v[i]+=v1.v[i];
  }
  return *this;
}

integer_vector& 
integer_vector::operator-=(const integer_vector& v1) {
  assert(d==v1.d);
  for (int i=0; i<d; i++) {
    v[i]-=v1.v[i];
  }
  return *this;
}

integer_vector  
integer_vector::operator+(const integer_vector& v1) const {
  assert(d==v1.d);
  integer_vector r(d);
  for (int i=0; i<d; i++) {
    r[i]=v[i]+v1.v[i];
  }
  return r;
}

integer_vector  
integer_vector::operator-(const integer_vector& v1) const {
  assert(d==v1.d);
  integer_vector r(d);
  for (int i=0; i<d; i++) {
    r[i]=v[i]-v1.v[i];
  }
  return r;
}

int  
integer_vector::operator*(const integer_vector& v1) const {
  int p=0;
  assert(d==v1.d);
  integer_vector r(d);
  for (int i=0; i<d; i++) {
    p+=v[i]*v1.v[i];
  }
  return p;
}

integer_vector  
integer_vector::operator*(int r) const {
  integer_vector v1(d);
  for (int i=0; i<d; i++) {
    v1[i]=v[i]*r;
  }
  return v1;
}

integer_vector  
integer_vector::operator-() const {
  integer_vector v1(d);
  for (int i=0; i<d; i++) {
    v1[i]=-v[i];
  }
  return v1;
}

integer_vector  
integer_vector::operator/(int r) const {
  integer_vector v1(d);
  for (int i=0; i<d; i++) {
    v1[i]=v[i]/r;
  }
  return v1;
}

bool     
integer_vector::operator==(const integer_vector& w) const {
  if (d!=w.d) return false;
  for (int i=0; i<d; i++) {
    if (v[i]!=w.v[i]) return false;
  }
  return true;
}

void
integer_vector::print(ostream& O) const {
  O << "| ";
  for (int i=0; i<d; i++) {
    O << v[i] << " ";
  }
  O << "| ";
}

void 
integer_vector::read(istream& I) {
  char buf[80];
  I>>buf;
  for (int i=0; i<d; i++) {
    I >> v[i];
  }
  I>>buf;
}

ostream& 
operator<<(ostream& O, const integer_vector& v) {
  v.print(O);
  return O;
}

istream& 
operator>>(istream& I, integer_vector& v) {
  v.read(I);
  return I;
}

