
#include "interestPt.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>


interestPt::interestPt()
{
  x=-1;
  y=-1;
  n=0;
  compare=-1;
  signature=NULL;
  
}
interestPt::interestPt(int inX, int inY)
{
  x=inX;
  y=inY;
  n=0;
  compare=-1;
  signature=NULL;
}

interestPt::interestPt(int inX, int inY, int inN)
{
  x=inX;
  y=inY;
  n=inN;
  compare=-1;
  if (n)
    {
      // allocate space and fill with -1's
      signature = new float[n];
      memset(signature,-1,n*sizeof(float));
    }
  else
    signature=NULL;
}

interestPt::interestPt(int inX, int inY, int inN, float *sigs)
{
  x=inX;
  y=inY;
  n=inN;
  compare=-1;
  if (n)
    {
      signature = new float[n];
      memset(signature,-1,n*sizeof(float));
    }
  else
    signature=NULL;

  for (int i=0; i<n; i++)
    signature[i] = sigs[i];

}

interestPt::interestPt(const interestPt& pt)
{
  x=pt.getx();
  y=pt.gety();
  n=pt.getn();
  compare=pt.getCompare();
  if (n) 
    {
      signature  = new float[n];
      bcopy((char *)pt.getSignature(), (char *)signature, n*sizeof(float));
    }
  else 
    signature = NULL;
}


interestPt::~interestPt()
{
  if (n) {
    delete []signature;
  }
}

void
interestPt::setSignature(int inN, float *sigs)
{
  // check if there was a preexisting signature
  if (signature != NULL)
    delete []signature;

  n=inN;
  
  if (n)
    {
      signature = new float[n];
      bcopy((char *)sigs,(char *)signature,n*sizeof(float));
    }
  else
    signature=NULL;
}


interestPt& 
interestPt::operator=(const interestPt& pt)
{
  if (signature != NULL) delete []signature;
  
  x=pt.getx();
  y=pt.gety();
  n=pt.getn();
  compare=pt.getCompare();
  if (n)
  {
    signature = new float[n];
    bcopy((char *)pt.getSignature(),(char *)signature,n*sizeof(float));
  }
  else signature=NULL;

  return *this;
}



void
interestPt::print(ostream& O) const
{
  // place the number of signatures first
  O << "#" << n << " ";
  O << "(" << getx() << "," << gety() << ")\t[";
  if (n) 
    {
      for (int i=0; i<n-1; i++)
	O.form("%9.9lf ",signature[i]);
      O.form("%9.9lf",signature[n-1]);
    }
  O << "]\n";
}


// NOTE!!! Assumes that the top of the file has #X where X is the number of signatures
// following the (x,y) coordinates within []
void 
interestPt::read(istream& I)
{
  char c;

  // check if there was a preexisting signature
  if (signature != NULL)
    delete []signature;
  
  // get the size of the array for the signature first
  do {
    c=I.get();
  }while (c!='#');
  if (c=='#')
    I >> n;

  // allocate space
  if (n)
    {
      signature = new float[n];
      memset(signature,-1,n*sizeof(float));
    } 

  // get x,y coords
  do {
    c=I.get();
  } while (c!='(');
  if (c=='(')
    I >> x;
  do {
    c=I.get();
  } while (c!=',');  
  if (c==',')
    I >> y;


  do {
    c=I.get();
  } while (c!='[');
  if (c=='[') 
    {
      for (int i=0; i<n; i++)
	I >> signature[i];
    }
  
  do {
    c=I.get();
  } while (c!=']');

}

//  Assumes that the buffer has the format of #X (x,y) [a b c ... X-1]\n
void
interestPt::read(char *buf)
{
  char sig[256];
  int i=0;
  int ok=0;

  while ((buf[i]!='\n') && (buf[i]!='(')) // if either a carriage return or '(' found should have seen a '#'
    {
      if (buf[i]=='#')
	{
	  ok=1;
	  break;
	}
      else
	i++;
    }
	
  // check if buf is valid (ugly but what can you do)
  if (!ok)
    {
      fprintf(stderr,"interestPt::read: Don't understand buffer format %s\n",buf);
      return;
    }
  else {
    n = atoi(&buf[i+1]); // hash sign immediately followed by number of sigs
  }

  if (signature != NULL)
    delete []signature;

  // allocate space and initialise to zero's
  if (n)
    {      
      signature = new float[n];
      memset(signature,-1,n*sizeof(float));
    } 
  else 
    {
      sscanf(buf,"#%d (%d,%d) %[^\n]",&n,&x,&y,sig);
      return;
    }

  sscanf(buf,"#%d (%d,%d) %[^\n]",&n,&x,&y,sig);

  for (int i=0; i<n; i++) {
    sscanf(sig+1,"%f%[^\n]",&signature[i],sig);
  }
}


ostream&
operator<<(ostream& O, const interestPt& p)
{
  p.print(O);
  return O;
}

istream&
operator>>(istream& I, interestPt& p)
{
  p.read(I);
  return I;
}

float 
interestPt::keyAt(int l)
{
  float val =  signature[l];
  return val;
}

// euclidean distance in signature space.
float 
interestPt::euclideanDist(interestPt& p)
{
  float dist=0;

  if (p.getn() != n)
    return -1;

  for (int i=0; i<n; i++)
    dist += (signature[i] - (p.getSignature())[i]) * (signature[i] - (p.getSignature())[i]) ;

  dist = sqrt(dist);
  return dist;
}
