/************************************************************************

   File:		IOdata.cpp

   Author(s):		Deeptiman Jugessur

   Created:		31 Oct 1998

   Last Revision:	$Date$

   Description:	

   $Revision$

   $Log$

	
   Copyright (c) 1998 by Deeptiman Jugessur, Centre for Intelligent Machines,
   McGill University, Montreal, QC.  Please see the copyright notice
   included in this distribution for full details.

 ***********************************************************************/
#define IODATA_CPP
#include <stdio.h>
#include "ioData.h"

#define ARR_SIZE 256

// This constructor is called when don't have database and array so it allocates space
ioData::ioData(int dbR, int dbC, int arrR, int arrC, fstream& dbstream, fstream& ipstream)
{
  numDatabaseRows = dbR;
  numDatabaseCols = dbC;
  numArrRows = arrR;
  numArrCols = arrC;

  // Allocate appropriate space for the database
  database = new matrix(numDatabaseRows,numDatabaseCols);

  // Allocate appropriate space for the interest Points array
  if (numArrRows && numArrCols)
    {
      array = new interestPt*[numArrRows];
      for (int i=0; i<numArrRows; i++)
	array[i] = new interestPt[numArrCols];
    }

  allocated=1;
  dbRead(dbstream);
  arrRead(ipstream);
}

// This constructor is called when DO have database and array so no new allocation
// purely for disk access.
ioData::ioData(matrix *dataB, interestPt **arr, int arrR, int arrC)
{
  // NOTE THAT THIS MAKES NO COPIES
  allocated=0;
  numDatabaseRows = dataB->dim1();
  numDatabaseCols = dataB->dim2();
  database = dataB; 
  numArrRows = arrR;
  numArrCols = arrC;
  array = arr;
}


ioData::~ioData()
{
  if (database != NULL) {
    delete database;
    database = NULL;
  }

  if (array != NULL) {
    for (int i=0; i<numArrRows; i++) {
      delete[] (array[i]);
      array[i] = NULL;
    }

    delete[] array;
    array = NULL;
  }
}


// read database from I
void
ioData::dbRead(istream& I)
{
  (*database).read(I);
}

// save database to O
void 
ioData::dbWrite(ostream& O) const
{
  (*database).print(O);
}


void
ioData::arrRead(istream& I)
{
  char streamBuf[ARR_SIZE];

  if (numArrRows && numArrCols)
    for (int i=0; i<numArrRows; i++)
      for (int j=0; j<numArrCols; j++)
	{
	  
	  I.getline(streamBuf,ARR_SIZE,'\n');
	  array[i][j].read(streamBuf);
	}
}

void
ioData::arrWrite(ostream& O) const
{
  // printf("NumArrRow=%d, numArrCols=%d\n",numArrRows, numArrCols);
  if (numArrRows && numArrCols)
    {
      for (int i=0; i<numArrRows; i++)
	for (int j=0; j<numArrCols; j++) {
	  O << array[i][j];
	  // cout << array[i][j];
	}
    }
}

void 
ioData::clearIoData()
{
  // This is a completely ugly hack, in fact the whole object sucks. 
  // Basically it clears the object for the destructor when no 
  // memory is allocated.
  if (!allocated) {
    database = NULL;
    array = NULL;
  }
  else
    printf("Delete the objects first\n");
}













