/************************************************************************

   File:		matrix.cpp

   Author(s):		Robert Sim

   Created:		10 Dec 1997

   Last Revision:	$Date: 1998/04/03 23:26:02 $

   Description:	

   $Revision: 1.4 $

   $Log: matrix.cpp,v $
   Revision 1.4  1998/04/03 23:26:02  simra
   Debugging outlier removal. Still.

   Revision 1.3  1998/03/18 17:33:27  simra
   1. alpha failed...
   2. matrix and vector reads still need work.

   Revision 1.2  1998/03/18 15:10:35  robbo
   1. Fixed matrix&vector reads.
   2. Fixed Statistic reads.
   3. Fixed the 75% bug- no need for iteration now.
   4. Added -w flag for localize matching.

   Revision 1.1.1.1  1998/02/24 02:12:11  robbo
   First fie import.


	
   Copyright (c) 1997 by Robert Sim, Centre for Intelligent Machines,
   McGill University, Montreal, QC.  Please see the copyright notice
   included in this distribution for full details.

 ***********************************************************************/

#define MATRIX_CPP
#define E_VAL 2.7182818
#include "matrix.h"
#include "integer_vector.h"
#include <math.h>
#include <assert.h>
matrix::matrix(int n, int m) {
  d1=n; 
  d2=m;
  if(d1 && d2) {
    v=new vector*[d1];
    for (unsigned int i=0; i<d1; i++) {
      v[i]=new vector(d2);
    }
  } else v=0;
}

matrix::matrix(const matrix& m) {
  d1=m.d1;
  d2=m.d2;
  if(d1 && d2) {
    v=new vector*[d1];
    for (unsigned int i=0; i<d1; i++) {
      v[i]=new vector(*m.v[i]);
    }
  } else v=0;
}

matrix::matrix(const vector& v1) {
  d1=v1.d;
  d2=1;
  if(d1 && d2) {
    v=new vector*[d1];
    for (unsigned int i=0; i<d1; i++) {
      v[i]=new vector(1);
      v[i]->v[0]=v1[i];
    }
  } else v=0;

}
    
matrix::matrix(int n, int m, PRECISION** r) {
  d1=n;
  d2=m;
  if(d1 && d2) {
    v=new vector*[d1];
    for (unsigned int i=0; i<d1; i++) {
      v[i]=new vector(d2);
      for (unsigned int j=0; j<d2; j++) {
	v[i]->v[j]=r[i][j];
      }
    }
  } else v=0;
}


matrix& 
matrix::operator=(const matrix& m) {
  // delete
  if (v) {
    for (unsigned int i=0; i<d1; i++) {
      delete v[i];
    }
  }
  delete []v;
  // realloc
  d1=m.d1;
  d2=m.d2;
  if(d1 && d2) {
    v=new vector*[d1];
    for (unsigned int i=0; i<d1; i++) {
      v[i]=new vector(m[i]);
    }
  } else v=0;
  return *this;
}

matrix::~matrix() {
  if (v) {
    for (unsigned int i=0; i<d1; i++) {
      delete v[i];
    }
  }
  delete []v;
}

vector& 
matrix::row(unsigned int i) const {
  assert(i>=0 && i<d1);
  return *(v[i]);
}

vector  
matrix::col(unsigned int i) const {
  assert(i>=0 && i<d2);
  vector r(d1);
  for (unsigned int j=0; j<d1; j++) {
    r[j]=v[j]->v[i];
  }
  return r;
}

matrix  
matrix::trans() const {
  matrix r(d2,d1);
  for (unsigned int j=0; j<d2; j++) {
    r[j]=col(j);
  }
  return r;
}

matrix::operator vector() const {
  assert(d2==1);
  return trans().row(0);
}

PRECISION& 
matrix::operator()(unsigned int i, unsigned int j) {
  assert(i>=0 && i<d1 && j>=0 && j<d2);
  return v[i]->v[j];
}

PRECISION  
matrix::operator()(unsigned int i, unsigned int j) const {
  assert(i>=0 && i<d1 && j>=0 && j<d2);
  return v[i]->v[j];
}

bool
matrix::operator==(const matrix& m) const {
  if (d1!=m.d1 || d2!=m.d2) return false;
  for (unsigned int i=0; i<d1; i++) {
    if (row(i)!=m.row(i)) return false;
  }
  return true;
}


vector
matrix::makeVector(const matrix& M)
{
  int vecSize=M.dim1()*M.dim2();
  PRECISION *vals;
  int k=0;

  vals = new PRECISION[vecSize];

  for (int i=0; i<M.dim1(); i++)
    for (int j=0; j<M.dim2(); j++)
      {
	vals[k] = M(i,j);
	k++;
      }

  vector v(vecSize,vals);
  delete vals;

  return v;
}


// This returns a matrix containing concentric annular averages
matrix
matrix::makeCircularMatrix(const matrix& M)
{
  struct radVal
  {
    float val;
    int count;
    radVal(float vin=0, int cin=0) : val(vin), count(cin) {}
  };
  
  //  int maxNumRads = (int)(0.5*sqrt(height*height + width*width/9);
  //  radVal radValArray[maxNumRads];

  int minRadius=999999;
  int radiusCount=0;
  int numRows=M.dim1();
  int numCols=M.dim2();
  int xc=numCols/2;
  int yc=numRows/2;
  int radius=0;


  for (int i=0; i<numRows; i++) {
    for (int j=0; j<numCols; j++) {
      radius = (int)(sqrt((j-xc)*(j-xc) + (i-yc)*(i-yc)));     
      if (radius < minRadius) {
	radiusCount++;
	minRadius = radius;
      }
    }
  }

  //  cout << "Num discrete radii = " << radiusCount << "\n";
  radVal radValArray[radiusCount];

  for (int i=0; i<numRows; i++) {
    for (int j=0; j<numCols; j++) {
      radius = (int)(sqrt((j-xc)*(j-xc) + (i-yc)*(i-yc)));     
      radValArray[radius].val += M(i,j);
      radValArray[radius].count++;
    }
  }

  for (int i=0; i<radiusCount; i++) {
    radValArray[i].val = radValArray[i].val / (float)radValArray[i].count;
  }

  matrix m(numRows,numCols);
  for (int i=0; i<numRows; i++)
    for (int j=0; j<numCols; j++) {
      radius = (int)(sqrt((j-xc)*(j-xc) + (i-yc)*(i-yc)));     
      //      cout << "radius at :(" << i << "," << j << ")=" << radius << "\n";
      m(i,j) = radValArray[radius].val;
    }
  
  /*
  vector v(radiusCount);
  for (int i=0; i<radiusCount; i++) {
    v[i] = radValArray[i].val;
  }
  */

  return m;
}


matrix 
matrix::makeGaussianMatrix(const matrix& M, PRECISION sigma=0)
{

  int width = M.dim2();
  int height = M.dim1();

  matrix m(height,width);
  m = M; // use the overloaded = to make a copy
  
  // Now I want to multiply it by a Gaussian
  // Allocate space in 2d
  PRECISION **gaussianVals;
  gaussianVals = new PRECISION*[height];
  for (int i=0; i<height; i++)
    gaussianVals[i] = new PRECISION[width]; 
  
  // fill in the values and maintain the largest value
  PRECISION largest=0;

  if (sigma <= 0) {
    if (width > height)
      sigma = width/4.0;
    else
      sigma = height/4.0;
  }

  int rowIndex=0;
  int colIndex=0;
  // first compute the gaussian with given sigma and keep the max value
  int imin=-(height/2);
  int imax=((height/2) -1 + (height%2));
  int jmin=-(width/2);
  int jmax=((width/2) -1 + (width%2));
  for (int i=imin; i<=imax; i++)
    {
      for (int j=jmin; j<=jmax; j++)
	{
	  /*
	  printf("h=%d, w=%d, sigma = %lf (%d,%d) = %lf [%d,%d]\n",height,width,sigma,rowIndex % height,colIndex % width,
		 pow(E_VAL,-(((i*i)+(j*j)) / (2*sigma*sigma))),i,j); 
	  */
	  gaussianVals[rowIndex % height][colIndex % width]=pow(E_VAL,-(((i*i)+(j*j)) / (2*sigma*sigma)));
	  if (gaussianVals[rowIndex % height][colIndex % width] >= largest)
	    largest = gaussianVals[rowIndex % height][colIndex % width];
	  colIndex++;
	}
      rowIndex++;
    }

  if (largest != 1) {
    cout << "Dividing through by " << largest << "\n";
    // need to normalise, divide by largest value
    for (int i=0; i<height; i++)
      for (int j=0; j<width; j++)
	gaussianVals[i][j] /= largest;
  }


  // multiply each value in the matrix by the gaussianVals.  Note that there are 3 times more cols due
  // to rgb so every three values of m get multiplied by the same value in gaussianVals
  rowIndex=0;
  colIndex=0;  
  for (int i=0; i<height; i++) 
    {
      for (int j=0; j<width; j++)
	{	
	  m(i,j) *= gaussianVals[rowIndex % height][colIndex % width];
	  colIndex++;
	}
      rowIndex++;	  
    }


  // clean up the gaussian mem
  for (int i=0; i<height; i++)
    {
      delete[] gaussianVals[i];
      gaussianVals[i] = NULL;
    }

  delete []gaussianVals;

  gaussianVals = NULL;


  return m;
}


/*
  Creates an r,theta matrix from M and returns it.  The dimesionality
  is done such that theta ranges from 0 to 2*pi, in (2*pi)/numT increments 
  and r ranges from 0 to maxR in M, in maxR/numR increments.
*/
matrix 
matrix::makePolarMatrix(const matrix& M, int numT, int numR)
{
  // First thing to do is to compute the maximum allowed radius.
  //  float maxR = 0.5*sqrt(M.dim1()*M.dim1()+M.dim2()*M.dim2());
  float maxR = 0.5*M.dim1(); // lose some resolution here as points beyond circumscribed circle are ignored
  // Get the step increment of r
  float stepR = maxR/numR;
  // Get the step increment of theta
  float stepT = 2*PI/numT;
  
  // Allocate space
  matrix polar(numR, numT);
  float curR=0;
  float curT=0;
  float fx=0;
  float fy=0;
  int x=0;
  int y=0;
  int actualX=0;
  int actualY=0;
  for (int i=0; i<numR; i++) {
    for (int j=0; j<numT; j++) {
      // find the corresponding x,y coordinates
      fx = curR*cos(curT);
      fy = curR*sin(curT);
      // rounding to nearest integer value
      fx > 0 ? x = (int)(fx+0.5) : x = (int)(fx-0.5);
      fy > 0 ? y = (int)(fy+0.5) : y = (int)(fy-0.5);
      // now these coords are from the center of the matrix so have to transform
      // them such that 0,0 is the top left corner.
      actualX = (x + M.dim2()/2) % M.dim2();
      actualY = (y + M.dim1()/2) % M.dim1();
      if (actualX < 0) actualX = 0;
      if (actualX > M.dim2()-1) actualX = M.dim2()-1;
      if (actualY < 0) actualY = 0;
      if (actualY > M.dim1()-1) actualY = M.dim1()-1;
      //      cout << "i=" << i << ",j=" << j << ",fx=" << fx << ",fy=" << fy << ",x=" 
      //	   << x << ",y=" << y << ",ax=" << actualX << ",ay=" << actualY << " curT=" << curT << " curR=" << curR << "\n" ;
      polar(i,j) = M(actualY,actualX); // reverse because y rows, x cols
      curT += stepT;
    }
    curT = 0;
    curR += stepR;
  }
  return polar;
}


matrix 
matrix::makeBiggerMatrix(const matrix&M, int numR, int numC)
{
  assert(numR >= M.dim1() && numC >= M.dim2());

  // allocate space for the bigger return matrix
  matrix bigger(numR, numC);
  float fx=0;
  float fy=0;
  int x=0;
  int y=0;
  float xrem=0;
  float yrem=0;
  float xratio=(M.dim2()-1)/(float)(numC-1);
  float yratio=(M.dim1()-1)/(float)(numR-1);
  PRECISION value=0;

  for (int i=0; i<numR; i++)
    for (int j=0; j<numC; j++)
      {
	fx = j*xratio;
	fy = i*yratio;
	x=(int)fx;
	y=(int)fy;
	xrem = fx - x;
	yrem = fy - y;
	//	cout << "xratio=" << xratio << " yratio=" << yratio << " fx=" << fx << " fy=" << fy << " x=" << x << " y=" << y << "\n";
	if (x==(M.dim2()-1) && y==(M.dim1()-1))
	  value = (1-xrem)*(1-yrem)*M(y,x);
	else if (x==(M.dim2()-1))
	  value = (1-xrem)*(1-yrem)*M(y,x) + yrem*(1-xrem)*M(y+1,x);
	else if (y==(M.dim1()-1))
	  value = (1-xrem)*(1-yrem)*M(y,x) + xrem*(1-yrem)*M(y,x+1);
	else
	  value = (1-xrem)*(1-yrem)*M(y,x) + xrem*(1-yrem)*M(y,x+1) + yrem*(1-xrem)*M(y+1,x) + xrem*yrem*M(y+1,x+1);	  
	//	cout << "xrem=" << xrem << " yrem=" << yrem << " value=" << value << "\n";
	//	bigger(j,i) = value;
	bigger(i,j) = value;
      }
  
  return bigger;
}

//  Crops info from this matrix onto the empty matrix from x1->x2 & y1->y2
//  For bounds not within the info matrix, padding is performed by wrapping 
//  info matrix around itself.
matrix 
matrix::crop(int x1, int y1, int x2, int y2)
{
  // make sure that corners make sense
  assert(x2 > x1 && y2 > y1);

  // allocate space for returned matrix
  matrix empty((y2-y1)+1, (x2-x1)+1);

  int colIndex = x1;
  int rowIndex = y1;
  int numCols = (x2-x1)+1;
  int numRows = (y2-y1)+1;

  int width = d2;
  int height = d1;
  
  // cout << "Width = " << width << ", Height = " << height << "\n"; 

  for (int i=0; i<numRows; i++)
    {
      for (int j=0; j<numCols; j++) 
        {
          if (colIndex<0 || colIndex>=width)
            colIndex = (colIndex + width) % width;
          
          if (rowIndex<0 || rowIndex>=height)
            rowIndex = (rowIndex + height) % height;
        
          /*
          cout << "e(" << i << "," << j << ")=";
          cout << "i(" << colIndex << "," << rowIndex << ")=";
          cout << info(rowIndex,colIndex) << " numR=" << numRows << ",numC=" << numCol
s << "\n";
          */
          
          empty(i,j) = (*this)(rowIndex,colIndex);
          colIndex++;
        }
      colIndex = x1;
      rowIndex++;
    }

  return empty;
}


PRECISION
matrix::leastSquares(const matrix& M, int xoffset=0, int yoffset=0)
{
  assert(d1==M.d1 && d2==M.d2);
  PRECISION result=0;
  int rowIndex=0;
  int colIndex=0;

  for (unsigned int i=0; i<d1; i++)
    for (unsigned int j=0; j<d2; j++) {
      colIndex = (j+xoffset) % d2;
      rowIndex = (i+yoffset) % d1;
      result += (M(rowIndex,colIndex) - operator()(i,j)) * (M(rowIndex,colIndex) - operator()(i,j));
    }

  return sqrt(result);
}

matrix 
matrix::operator+(const matrix& M1) {
  assert(d1==M1.d1 && d2==M1.d2);
  matrix R(d1,d2);
  for (unsigned int i=0; i<d1; i++) {
    R[i]=row(i)+M1[i];
  }
  return R;
}

matrix 
matrix::operator-(const matrix& M1) {
  assert(d1==M1.d1 && d2==M1.d2);
  matrix R(d1,d2);
  for (unsigned int i=0; i<d1; i++) {
    R[i]=row(i)-M1[i];
  }
  return R;
}

matrix 
matrix::operator-() {
  matrix R(d1,d2);
  for (unsigned int i=0; i<d1; i++) {
    R[i]=-row(i);
  }
  return R;
}


matrix& 
matrix::operator-=(const matrix& M1) {
  assert(d1==M1.d1 && d2==M1.d2);
  for (unsigned int i=0; i<d1; i++) {
    row(i)-=M1[i];
  }
  return *this;
}

matrix& 
matrix::operator+=(const matrix& M1) {
  assert(d1==M1.d1 && d2==M1.d2);
  for (unsigned int i=0; i<d1; i++) {
    row(i)+=M1[i];
  }
  return *this;
}

matrix 
matrix::operator*(const matrix& M1) {
  assert(d2==M1.d1);
  matrix R(d1,M1.d2);
  for (unsigned int i=0; i<d1; i++) {
    for (unsigned int j=0; j<M1.d2; j++) {
      R[i][j]=row(i)*M1.col(j);
    }
  }
  return R;
}

matrix 
matrix::operator*(PRECISION x) {
  matrix R(d1,d2);
  for (unsigned int i=0; i<d1; i++) {
    R[i]=row(i)*x;
  }
  return R;
}

void 
matrix::print(ostream& O) const {
  O << "(";
  for (unsigned int i=0; i<d1; i++) {
    O << row(i);
  }
  O << ")\n";
}


void 
matrix::printBinary(ostream& O, int flag) const {
  char *openparen = "(";
  char *closeparen = ")\n";

  O.write((char *)&d1,sizeof(int));
  O.write((char *)&d2,sizeof(int));
  O.write(openparen,sizeof(char));
  for (unsigned int i=0; i<d1; i++) {
    row(i).printBinary(O,flag);
  }
  O.write(closeparen,2*sizeof(char));
}

void 
matrix::read(istream& I) {
  char c;
  do {
    c=I.get();   
  } while (c!='(');
  
  if (c=='(') {
    for (unsigned int i=0; i<d1; i++) {
      I >> row(i);
      // cout << "i=" << i << row(i) << "\n";
    }
    do {
      c=I.get();
    } while (c!=')');
  }
}

void
matrix::readBinary(istream& I, int flag) {
  char openparen;
  char closeparen;
  char carriage;
  int nr;
  int nc;

  if (v) {
    for (unsigned int i=0; i<d1; i++) {
      delete v[i];
    }
  }
  delete []v;

  I.read((char *)&nr,sizeof(int));
  // cout << "Read " << nr << "\n";
  I.read((char *)&nc,sizeof(int));
  // cout << "Read " << nc << "\n";
  
  d1=nr;
  d2=nc;
  v=new vector*[d1];
  for (unsigned int i=0; i<d1; i++) 
    v[i]=new vector(d2);
  
  I.read((char *)&openparen,sizeof(char));
  // cout << "Read " << openparen << "\n";
  for (unsigned int i=0; i<d1; i++)
    row(i).readBinary(I,flag);

  I.read((char *)&closeparen,sizeof(char));
  I.read((char *)&carriage,sizeof(char));
}

matrix&
matrix::ones() {
  assert(d1==d2);
  matrix &a=*this;
  for (unsigned int k=0; k<d1; k++) {
    for (unsigned int j=0; j<d2; j++) {
      if (k==j) a(k,j)=1.0;
      else a(k,j)=0.0;
    }
  }
  return a;
}

ostream& operator<<(ostream& O, const matrix& M) {
  M.print(O);
  return O;
}


istream& operator>>(istream& I, matrix& M) {
  M.read(I);
  return I;
}

#ifdef SWAP
#error "Can't define swap"
#endif
#define SWAP(a, b) {PRECISION temp=(a);(a)=(b);(b)=temp;}
matrix
matrix::inv() const {
  // Invert a matrix from numerical recipes.. using gauss-jordan elim.
  // non-destructive-makes a copy of this.
  matrix b(d1, d2);  
  //  b.ones();
  //matrix &a=*this;
  matrix a=*this;
  unsigned int irow,icol;
  integer_vector indxc(d1);
  integer_vector indxr(d1);
  integer_vector ipiv(d1);

  for (unsigned int p=0; p<d1; p++) {
    PRECISION big=0.0;
    for (unsigned int j=0; j<d1; j++) 
      if (ipiv[j]!=1)
	for (unsigned int k=0; k<d1; k++) {
	  if (ipiv[k]==0) {
	    if (fabs(a(j,k))>=big) {
	      big=fabs(a(j,k));
	      irow=j;
	      icol=k;
	    }
	  } else if (ipiv[k]>1)  {
	    cerr<<"GAUSSJ: Singular Matrix-1"<<endl;
	    return matrix(0,0);
	  }
	}
    ++(ipiv[icol]);
    if (irow!=icol) {
      for (unsigned int l=0; l<d1; l++) SWAP(a(irow,l), a(icol,l));
      //for (int l=0; l<d1; l++) SWAP(b(irow,l), b(icol,l));
    }
    indxr[p]=irow;
    indxc[p]=icol;
    if (a(icol,icol)==0.0) {
      cerr<<"GAUSSJ: Singular Matrix-2"<<endl;
      return matrix(0,0);
    }
    PRECISION pivinv=1.0/a(icol,icol);
    a(icol,icol)=1.0;
    for (unsigned int l=0;l<d1;l++) a(icol,l) *= pivinv;
    //for (int l=0;l<d1;l++) b(icol,l) *= pivinv;
    for (unsigned int ll=0;ll<d1;ll++)
      if (ll != icol) {
	PRECISION dum=a(ll,icol);
	a(ll,icol)=0.0;
	for (unsigned int l=0;l<d1;l++) a(ll,l) -= a(icol,l)*dum;
	//for (int l=0;l<d1;l++) b(ll,l) -= b(icol,l)*dum;
      }
  }
  for (int l=d1-1;l>=0;l--) {
    if (indxr[l] != indxc[l])
      for (unsigned int k=0;k<d1;k++)
	SWAP(a(k,indxr[l]),a(k,indxc[l]));
  }
  return a;
}
#undef SWAP

#if (defined PYTHAG || defined MAX || defined SIGN)
#error "Can't define PYTHAG or MAX or SIGN"
#endif
/*
static float at,bt,ct;
#define PYTHAG(a,b) ((at=fabs(a)) > (bt=fabs(b)) ? \
(ct=bt/at,at*sqrt(1.0+ct*ct)) : (bt ? (ct=at/bt,bt*sqrt(1.0+ct*ct)): 0.0))
*/

static float maxarg1,maxarg2;
#define MAX(a,b) (maxarg1=(a),maxarg2=(b),(maxarg1) > (maxarg2) ?\
	(maxarg1) : (maxarg2))
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))
static inline double dsqr(double a) {return a*a;}

static unsigned int minarg1, minarg2;
#define IMIN(a,b) (minarg1=(a),minarg2=(b),(minarg1) < (minarg2) ?\
	(minarg1) : (minarg2))

double pythag(double a, double b) {
  double absa, absb;
  absa=fabs(a);
  absb=fabs(b);
  if (absa > absb) return (absa*sqrt(1.0+dsqr(absb/absa)));
  else return (absb == 0.0 ? 0.0 : absb*sqrt(1.0+dsqr(absa/absb)));
}

void
matrix::svdcmp(matrix& U, vector& W, matrix& V) {
  // singular value decompostion from numerical recipes-non destructive.
  matrix a=*this;
  cout << "MATRIX DIMS ARE " << d1 << " by " << d2 << "\n";
  unsigned int i,its,j,jj,k,l=0,nm=0;
  PRECISION c,s,f,h,x,y,z;
  PRECISION anorm=0,g=0,scale=0;
  vector rv1(d2);
  
  if (d1 < d2) {
    cerr<<"SVDCMP: You must augment A with extra zero rows"<<endl;
    //    U=a;
    W=vector(0);
    V=matrix(0,0);
    return;
  }
  
  W=vector(d2);
  V=matrix(d2,d2);
  for (i=0; i<d2; i++) {
#ifndef NDEBUG
    cerr<< "Pass 1 (0 to d2) : Col: "<<i<<"\r"<<flush;
#endif
    l=i+1;
    rv1[i]=scale*g;
    g=s=scale=0;
    if (i<d1) {
      for (k=i; k<d1; k++) scale+=(PRECISION)fabs(a(k,i));
      if (scale) {
	for (k=i; k<d1; k++) {
	  a(k,i)/=scale;
	  s+=a(k,i)*a(k,i);
	}
	f=a(i,i);
	g= (PRECISION) -SIGN(sqrt(s),f);
	h=f*g-s;
	a(i,i)=f-g;
	if (i!=d2) {
	  for (j=l; j<d2; j++) {
	    for (s=0,k=i;k<d1;k++) s+=a(k,i)*a(k,j);
	    f=s/h;
	    for (k=i;k<d1;k++) a(k,j)+=f*a(k,i);
	  }
	}
	for (k=i; k<d1; k++) a(k,i) *=scale;
      }
    }
    W[i]=scale*g;
    g=s=scale=0;
    
    if (i<d1 && i!=(d2-1)) {
      for (k=l; k<d2; k++) scale+=(PRECISION)fabs(a(i,k));
      if (scale) {
	for (k=l;k<d2;k++) {
	  a(i,k)/=scale;
	  s+=a(i,k)*a(i,k);
	}
	f=a(i,l);
	g= (PRECISION) -SIGN(sqrt(s),f);
	h=f*g-s;
	a(i,l)=f-g;
	for (k=l; k<d2; k++) rv1[k]=a(i,k)/h;
	for (j=l; j<d1; j++) {
	  for (s=0, k=l; k<d2; k++) s+=a(j,k)*a(i,k);
	  for (k=l; k<d2; k++) a(j,k)+=s*rv1[k];
	}
	for (k=l; k<d2; k++) a(i,k)*=scale;
      }
    }
    anorm=(PRECISION)MAX(anorm,(fabs(W[i])+fabs(rv1[i])));
  }
  
  for (i=d2-1; i<d2; i--) {
#ifndef NDEBUG
    cerr<<"Pass 2: Col: "<<i<<"\r"<<flush;
#endif
    if (i<(d2-1)) {
      if (g) {
	for (j=l; j<d2; j++) 
	  V(j,i) = (a(i,j)/a(i,l))/g;
	for (j=l; j<d2; j++) {
	  for (s=0,k=l; k<d2; k++) s+=a(i,k)*V(k,j);
	  for (k=l;k<d2; k++) V(k,j)+=s*V(k,i);
	}
      }
      for (j=l; j<d2; j++) V(i,j)=V(j,i)=0;
    }
    V(i,i)=1;
    g=rv1[i];
    l=i;
  }

  for (i=IMIN(d1-1,d2-1); i<IMIN(d1,d2); i--) {
#ifndef NDEBUG
    cerr<<"Pass 3: Col: "<<i<<"\t\r"<<flush;
#endif
    l=i+1;
    g=W[i];
    for (j=l; j<d2; j++) a(i,j)=0;
    if (g) {
      g=(PRECISION)(1.0/g);
      for (j=l; j<d2; j++) {
	for (s=0, k=l; k<d1; k++) s+=a(k,i)*a(k,j);
	f=(s/a(i,i))*g;
	for (k=i; k<d1; k++) a(k,j)+=f*a(k,i);
      }
      for (j=i; j<d1; j++) a(j,i)*=g;
    } else for (j=i; j<d1; j++) a(j,i)=0;
    ++a(i,i);
  }
  for (k=d2-1; k<d2; k--) {
    for (its=1; its<=30; its++) {
#ifndef NDEBUG
      cerr<< "Pass 4: Col: "<<k<<" Iteration: "<<its<<"\r"<<flush;
#endif
      bool flag=true;
      for (l=k; l<k+1; l--) {
	nm=l-1;
	if ((float)(fabs(rv1[l])+anorm)==anorm) {
	  flag=false;
	  break;
	}
	if ((float)(fabs(W[nm])+anorm)==anorm) break;
      }
      if (flag) {
	c=0;
	s=1;
	for (i=l; i<=k; i++) {
	  f=s*rv1[i];
	  rv1[i]=c*rv1[i];
	  if ((float)(fabs(f)+anorm)==anorm) break;
	  g=W[i];
	  h=(PRECISION)pythag(f,g);
	  W[i]=h;
	  h=(PRECISION)(1.0/h);
	  c=g*h;
	  s= -f*h;
	  for (j=0; j<d1; j++) {
	    y=a(j,nm);
	    z=a(j,i);
	    a(j,nm)=y*c+z*s;
	    a(j,i)=z*c-y*s;
	  }
	}
      }
      z=W[k];
      if (l==k) {
	if (z<0) {
	  W[k]= -z;
	  for (j=0; j<d2; j++) V(j,k)= -V(j,k);
	}
	break;
      }
      if (its==30) {
	cerr<<"No convergence in 30 SVDCMP iterations"<<endl;
	return;
      }
      x=W[l];
      nm=k-1;
      y=W[nm];
      g=rv1[nm];
      PRECISION h=rv1[k];
      PRECISION f=(PRECISION) (((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y));
      g=(PRECISION)pythag(f,1.0);
      f=(PRECISION)((x-z)*(x+z)+h*((y/(f+SIGN(g,f)))-h))/x;
      c=s=1;
      for (j=l;j<=nm;j++) {
	i=j+1;
	g=rv1[i];
	y=W[i];
	h=s*g;
	g=c*g;
	z=(PRECISION)pythag(f,h);
	rv1[j]=z;
	c=f/z;
	s=h/z;
	f=x*c+g*s;
	g=g*c-x*s;
	h=y*s;
	y *= c;
	for (jj=0;jj<d2;jj++) {
	  x=V(jj,j);
	  z=V(jj,i);
	  V(jj,j)=x*c+z*s;
	  V(jj,i)=z*c-x*s;
	}
	z=(PRECISION)pythag(f,h);
	W[j]=z;
	if (z) {
	  z=(PRECISION)(1.0/z);
	  c=f*z;
	  s=h*z;
	}
	f=c*g+s*y;
	x=c*y-s*g;
	for (jj=0;jj<d1;jj++) {
	  y=a(jj,j);
	  z=a(jj,i);
	  a(jj,j)=y*c+z*s;
	  a(jj,i)=z*c-y*s;
	}
      }
      rv1[l]=0;
      rv1[k]=f;
      W[k]=x;
    }
  }
  U=a;
#ifndef NDEBUG
  cerr<<endl;
#endif
}


/*
void
matrix::svdcmp(matrix& U, vector& W, matrix& V) {
  // singular value decompostion from numerical recipes-non destructive.
  matrix a=*this;
  int i,its,j,jj,k,l=1,nm=d2-2;
  PRECISION c,f,h,s,x,y,z;
  PRECISION anorm=0.0,g=0.0,scale=0.0;
  vector rv1(d2);
  
  if (d1 < d2) {
    cerr<<"SVDCMP: You must augment A with extra zero rows"<<endl;
    //    U=a;
    W=vector(0);
    V=matrix(0,0);
    return;
  }
  
  W=vector(d2);
  V=matrix(d2,d2);
  for (i=0; i<d2; i++) {
    verbprintf(3, "Pass 1: Col: %d\r", i);
    l=i+1;
    rv1[i]=scale*g;
    g=scale=0.0;
    s=0.0;
    if (i<d1) {
      for (k=i; k<d1; k++) scale+=fabs(a(k,i));
      if (scale) {
	for (k=i; k<d1; k++) {
	  a(k,i)/=scale;
	  s+=a(k,i)*a(k,i);
	}
	f=a(i,i);
	g=-SIGN(sqrt(s),f);
	h=f*g-s;
	a(i,i)=f-g;
	if (i!=d2) {
	  for (j=l; j<d2; j++) {
	    for (s=0.0,k=i;k<d1;k++) s+=a(k,i)*a(k,j);
	    f=s/h;
	    for (k=i;k<d1;k++) a(k,j)+=f*a(k,i);
	  }
	}
	for (k=i; k<d1; k++) a(k,i) *=scale;
      }
    }
    W[i]=scale*g;
    g=s=scale=0.0;
    
    if (i<d1 && i!=(d2-1)) {
      for (k=l; k<d2; k++) scale+=fabs(a(i,k));
      if (scale) {
	for (k=l;k<d2;k++) {
	  a(i,k)/=scale;
	  s+=a(i,k)*a(i,k);
	}
	f=a(i,l);
	g= -SIGN(sqrt(s),f);
	h=f*g-s;
	a(i,l)=f-g;
	for (k=l; k<d2; k++) rv1[k]=a(i,k)/h;
	if (i!=d1) {
	  for (j=l; j<d1; j++) {
	    for (s=0.0, k=l; k<d2; k++) s+=a(j,k)*a(i,k);
	    for (k=l; k<d2; k++) a(j,k)+=s*rv1[k];
	  }
	}
	for (k=l; k<d2; k++) a(i,k)*=scale;
      }
    }
    anorm=MAX(anorm,(fabs(W[i])+fabs(rv1[i])));
  }
  
  for (i=d2-1; i>=0; i--) {
    verbprintf(3, "Pass 2: Col: %d\r", i);
    if (i<(d2-1)) {
      if (g) {
	for (j=l; j<d2; j++) 
	  V(j,i) = (a(i,j)/a(i,l))/g;
	for (j=l; j<d2; j++) {
	  for (s=0.0,k=l; k<d2; k++) s+=a(i,k)*V(k,j);
	  for (k=l;k<d2; k++) V(k,j)+=s*V(k,i);
	}
      }
      for (j=l; j<d2; j++) V(i,j)=V(j,i)=0.0;
    }
    V(i,i)=1.0;
    g=rv1[i];
    l=i;
  }

  for (i=d2-1; i>=0; i--) {
    verbprintf(3, "Pass 3: Col: %d\t\r", i);
    l=i+1;
    g=W[i];
    if (i<(d2-1)) for (j=l; j<d2; j++) a(i,j)=0.0;
    if (g) {
      g=1.0/g;
      if (i!=(d2-1)) {
	for (j=l; j<d2; j++) {
	  for (s=0.0, k=l; k<d1; k++) s+=a(k,i)*a(k,j);
	  f=(s/a(i,i))*g;
	  for (k=i; k<d1; k++) a(k,j)+=f*a(k,i);
	}
      }
      for (j=i; j<d1; j++) a(j,i)*=g;
    } else {
      for (j=i; j<d1; j++) a(j,i)=0.0;
    }
    ++a(i,i);
  }
  for (k=d2-1; k>=0; k--) {
    for (its=1; its<=30; its++) {
      verbprintf(3, "Pass 4: Col: %d Iteration: %d\r", k, its);
      bool flag=true;
      for (l=k; l>=0; l--) {
	nm=l-1;
	if (fabs(rv1[l])+anorm==anorm) {
	  flag=false;
	  break;
	}
	if (fabs(W[nm])+anorm==anorm) break;
      }
      if (flag) {
	c=0.0;
	s=1.0;
	for (i=l; i<=k; i++) {
	  f=s*rv1[i];
	  if (fabs(f)+anorm!=anorm) {
	    g=W[i];
	    h=PYTHAG(f,g);
	    W[i]=h;
	    h=1.0/h;
	    c=g*h;
	    s=(-f*h);
	    for (j=0; j<d1; j++) {
	      y=a(j,nm);
	      z=a(j,i);
	      a(j,nm)=y*c+z*s;
	      a(j,i)=z*c-y*s;
	    }
	  }
	}
      }
      z=W[k];
      if (l==k) {
	if (z<0.0) {
	  W[k]=-z;
	  for (j=0; j<d2; j++) V(j,k)=(-V(j,k));
	}
	break;
      }
      if (its==30) {
	cerr<<"No convergence in 30 SVDCMP iterations"<<endl;
	return;
      }
      x=W[l];
      nm=k-1;
      y=W[nm];
      g=rv1[nm];
      PRECISION h=rv1[k];
      PRECISION f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
      g=PYTHAG(f,1.0);
      f=((x-z)*(x+z)+h*((y/(f+SIGN(g,f)))-h))/x;
      c=s=1.0;
      for (j=l;j<=nm;j++) {
	i=j+1;
	g=rv1[i];
	y=W[i];
	h=s*g;
	g=c*g;
	z=PYTHAG(f,h);
	rv1[j]=z;
	c=f/z;
	s=h/z;
	f=x*c+g*s;
	g=g*c-x*s;
	h=y*s;
	y=y*c;
	for (jj=0;jj<d2;jj++) {
	  x=V(jj,j);
	  z=V(jj,i);
	  V(jj,j)=x*c+z*s;
	  V(jj,i)=z*c-x*s;
	}
	z=PYTHAG(f,h);
	W[j]=z;
	if (z) {
	  z=1.0/z;
	  c=f*z;
	  s=h*z;
	}
	f=(c*g)+(s*y);
	x=(c*y)-(s*g);
	for (jj=0;jj<d1;jj++) {
	  y=a(jj,j);
	  z=a(jj,i);
	  a(jj,j)=y*c+z*s;
	  a(jj,i)=z*c-y*s;
	}
      }
      rv1[l]=0.0;
      rv1[k]=f;
      W[k]=x;
    }
  }
  U=a;
  verbprintf(3, "\n");
}
*/


#undef SIGN
#undef MAX
#undef PYTHAG


void 
matrix::svbksb(const vector &W, const matrix& V, const vector& b,
	       vector& x) const {
  int jj,j,i;
  PRECISION s;
  vector tmp(W.dim());
  x=vector(W.dim());
  const matrix &U=*this;
  for (j=0; j<W.dim(); j++) {
    s=0.0;
    if (W[j]) {
      for (i=0; i<U.dim1(); i++) s+=U(i,j)*b[i];
      s/=W[j];
    }
    tmp[j]=s;
  }
  for (j=0; j<x.dim(); j++) {
    s=0.0;
    for (jj=0; jj<tmp.dim(); jj++) s+=V(j,jj)*tmp[jj];
    x[j]=s;
  }
}


#define TOL 1.0e-5
void 
matrix::svdfit(const vector& x, const vector& y, const vector& sig,
	      vector& a, matrix& V, vector& w, PRECISION* chisq,
	      void (*funcs)(PRECISION, vector&))
{
  int j,i;
  PRECISION wmax,tmp,thresh,sum;
  int ma=a.dim();
  int ndata=x.dim();
  vector b(ndata);
  matrix u(ndata,ma);
  vector afunc(ma);
  for (i=0;i<ndata;i++) {
    (*funcs)(x[i],afunc);
    tmp=1.0/sig[i];
    for (j=0;j<ma;j++) u(i,j)=afunc[j]*tmp;
    b[i]=y[i]*tmp;
  }
  matrix u2;
  u.svdcmp(u2,w,V);
  wmax=0.0;
  for (j=0;j<ma;j++)
    if (w[j] > wmax) wmax=w[j];
  thresh=TOL*wmax;
  for (j=0;j<ma;j++)
    if (w[j] < thresh) w[j]=0.0;
  u2.svbksb(w,V,b,a);
  *chisq=0.0;
  for (i=0;i<ndata;i++) {
    (*funcs)(x[i],afunc);
    for (sum=0.0,j=0;j<ma;j++) sum += a[j]*afunc[j];
    *chisq += (tmp=(y[i]-sum)/sig[i],tmp*tmp);
  }
}

#undef TOL
