/*******************************************************************************
+
+
+  matrix.h  A sort of but not really LEDA 3.5.1 compatible matrix class
+
+ 
*******************************************************************************/

#ifndef MATRIX_H
#define MATRIX_H

#ifndef PI
#define PI 3.1415927
#endif
//------------------------------------------------------------------------------
//  matrices
//------------------------------------------------------------------------------

#include "vector.h"
#include "math.h"
#include "VObject.h"
class matrix : public VObject {
  vector** v;
  unsigned int  d1; // number of rows
  unsigned int  d2; // number of columns

public:

  matrix(int n=0, int m=0);
  matrix(const matrix&);
  matrix(const vector&);
  matrix(int,int,PRECISION**);
  matrix& operator=(const matrix&);
  ~matrix();
  int dim1()  const  {  return d1; }
  int dim2()  const  {  return d2; }
  vector& row(unsigned int i) const;
  vector  col(unsigned int i) const;
  matrix  trans() const;
  //matrix  inv()   const;
  //PRECISION  det()   const;
  //matrix solve(const matrix&) const;
  //vector solve(const vector& b) const { return vector(solve(matrix(b))); }
  operator vector() const; 
  vector& operator[](int i)    const { return row(i); }
  PRECISION& operator()(unsigned int i, unsigned int j);
  PRECISION  operator()(unsigned int,unsigned int) const;
  bool operator==(const matrix&)    const;
  bool operator!=(const matrix& x)  const { return !(*this == x); }

  vector makeVector(const matrix& M);

  matrix makeCircularMatrix(const matrix& M);
  matrix makeGaussianMatrix(const matrix& M, PRECISION sigma=0);
  matrix makePolarMatrix(const matrix& M, int numT, int numR);
  matrix makeBiggerMatrix(const matrix& M, int numR, int numC);

  PRECISION leastSquares(const matrix& M, int xoffset=0, int yoffset=0);
  matrix crop(int x1, int y1, int x2, int y2);

  matrix operator+(const matrix& M1);
  matrix operator-(const matrix& M1);
  matrix operator-(); // unary
  matrix& operator-=(const matrix&);
  matrix& operator+=(const matrix&);
  matrix operator*(const matrix& M1);
  vector operator*(const vector& vec) { return vector(*this * matrix(vec)); }
  matrix operator*(PRECISION x);
  void print(ostream& O) const;
  void printBinary(ostream& O, int flag=0) const; // flag values:0 for PRECISIONs, 1 for chars ie 8 bits
  void print() { print(cout); }
  void read(istream& I);
  void readBinary(istream& I, int flag=0);        // flag values:0 for PRECISIONs, 1 for chars ie 8 bits
  void read() { read(cin); }
  friend ostream& operator<<(ostream& O, const matrix& M);
  friend istream& operator>>(istream& I, matrix& M);
  matrix inv() const;
  matrix& ones();
  void svdcmp(matrix& U, vector& W, matrix& V);
  void svbksb(const vector &W, const matrix& V, const vector& b,
	      vector& x) const;
  static void svdfit(const vector& x, const vector& y, const vector& sig,
	      vector& a, matrix& V, vector& w, PRECISION* chisq,
	      void (*funcs)(PRECISION, vector&));
};
#endif


