/**************************************************************************

   File:                utilityclasses.h

   Author(s):           Robert Sim

   Created:              6 Dec 1997

   Last Revision:       $Date: 1998/07/21 14:21:58 $

   Description: 

   $Revision: 2.0 $

   $Log: utilityclasses.h,v $
   Revision 2.0  1998/07/21 14:21:58  simra
   Final Thesis submitted version.

   Revision 1.5  1998/04/21 21:20:28  simra
   More LocalizationInfo debugging.
   Fixed RAND_MAX to MY_RAND_MAX
   Plotting uncert ellipse.

   Revision 1.4  1998/04/07 23:40:48  simra
   Changes to compile on sgi w/ CC. It works- huge speedup!

   Revision 1.3  1998/04/02 22:43:30  simra
   Abstracted LocalizationResults away and rewrote the plotter to get a better
   look at what's going on.  We need to separate basic constellation data from
   hard-core data.  Also point3f's etc should be aliased vectors to make life
   nice.

   Revision 1.2  1998/03/15 18:39:02  robbo
   1. Rewrote purging and set it up to be on by default in eigentrack.
   2. Implemented Sm and Cheesman-like uncertainty measures for lms.
      This needs to be implemented into eigentrack or RecomputeStats for
      precomputation.  It takes some time but I think it should be useful
      in the end.
   3. uncertainty implementation required some point3f vs vector
      reimplementation.  I'm not entirely satisfied w/ the results. Too
      much is tied to the LocalizationWindow way of looking at the data,
      and so I need to rewrite all the linf, votedat objects.  A lot of
      the old localization methods are obsolete now.

   Revision 1.1.1.1  1998/02/24 02:12:11  robbo
   First fie import.


	
   Copyright (c) 1997 by Robert Sim, Centre for Intelligent Machines,
   McGill University, Montreal, QC.  Please see the copyright notice
   included in this distribution for full details.

 **************************************************************************/

#ifndef UTILITYCLASSES_H
#  define UTILITYCLASSES_H
#include <iostream.h>
#include <math.h>

#ifdef sgi
#include "bool.h"
#endif

// RAND_MAX is different between sgi and linux, and sun doesn't define
// it.  Luckily, random(3) has the same max val for all three.
#define MY_RAND_MAX  2147483647

template<class T>
inline T my_sqr(T t) {return t*t;}
//#define my_sqr(x) ((x)*(x))

class point2i {
public:
  int x;
  int y;
  point2i(int i=0, int j=0) {x=i;y=j;}
  bool operator ==(const point2i& p) const {return (x==p.x && y==p.y);}
  bool operator!=(const point2i& p) const {return !(operator==(p));}
  point2i operator+(const point2i& p) {return point2i(x+p.x, y+p.y);}
  point2i operator-(const point2i& p) {return point2i(x-p.x, y-p.y);}
  point2i& operator+=(const point2i& p) {x+=p.x; y+=p.y; return *this;}
  friend ostream& operator<<(ostream& O, point2i p);
  friend istream& operator>>(istream& I, point2i& p);
  double length() {return sqrt((double)((x*x)+(y*y)));}
};

class point2f {
public:
  double x;
  double y;
  point2f(double i=0, double j=0) {x=i;y=j;}
  bool operator ==(const point2f& p) const {return (x==p.x && y==p.y);}
  point2f operator+(const point2f& p) {return point2f(x+p.x, y+p.y);}
  point2f& operator+=(const point2f& p) {x+=p.x; y+=p.y; return *this;}
  friend ostream& operator<<(ostream& O, point2f p);
  friend istream& operator>>(istream& I, point2f& p);
  double length() {return sqrt((x*x)+(y*y));}
};

class point3f {
public:
  double x;
  double y;
  double z;
  bool operator ==(const point3f& p) const {return (x==p.x && y==p.y && z==p.z);}
  bool operator !=(const point3f& p) const {return !operator==(p);}
  point3f(float i=0, float j=0, float k=0) {x=i;y=j;z=k;}
  point3f operator+(const point3f& p) {return point3f(x+p.x, y+p.y,
						      z+p.z);}
  point3f operator+=(const point3f& p) {x+=p.x; y+=p.y; z+=p.z; return *this;}
  point3f operator-(const point3f& p) {return point3f(x-p.x, y-p.y,
						      z-p.z);}
  point3f operator-=(const point3f& p) {x-=p.x; y-=p.y, z-=p.z; return *this;}
  point3f operator*(double d) { return point3f(x*d, y*d, z*d);}
  point3f operator*=(double d) { x*=d; y*=d; z*=d; return *this;}
  point3f operator/(double d) { return point3f(x/d, y/d, z/d);}
  point3f operator/=(double d) { x/=d; y/=d; z/=d; return *this;}
  double length() {return sqrt((x*x)+(y*y)+(z*z));}
  friend ostream& operator<<(ostream& O, const point3f& p);
  friend istream& operator>>(istream& I, point3f& p);
};

class GMLObject {
protected:
  int nextGMLToken(istream &I, char * tok, int len);
  int closeGMLField(istream &I);
};


#endif  /* UTILITYCLASSES_H */
