/************************************************************************

   File:		vector.cpp

   Author(s):		Robert Sim

   Created:		10 Dec 1997

   Last Revision:	$Date: 1998/03/18 17:33:27 $

   Description:	

   $Revision: 1.8 $

   $Log: vector.cpp,v $
   Revision 1.8  1998/03/18 17:33:27  simra
   1. alpha failed...
   2. matrix and vector reads still need work.

   Revision 1.7  1998/03/18 15:10:36  robbo
   1. Fixed matrix&vector reads.
   2. Fixed Statistic reads.
   3. Fixed the 75% bug- no need for iteration now.
   4. Added -w flag for localize matching.

   Revision 1.6  1998/03/17 23:41:26  simra
   alpha troubles.

   Revision 1.5  1998/03/17 19:46:36  simra
   alpha compilation

   Revision 1.4  1998/03/10 21:49:26  simra
   Cross platform support- part I

   Revision 1.3  1998/02/24 22:26:42  simra
   Did some parameter estimation with localizer, but it's slow.  LandmarkSet's
   now compute the stats on demand- no more hold/release- it uses a stale flag to
   decide when--- BUG: if a landmark is accessed and altered using [] then the
   stale flag won't be set.  Adjusted library depends- libnr is no longer in the
   distribution.

   Revision 1.2  1998/02/24 13:35:15  robbo
   Changed eigentrack extensively to update the pca on the fly- best
   distinguishing the tracked sets. This required some major overhauling.
   Also changed some input file conventions. landmark outputs a clist,
   eigentrack inputs the same clist. eigenland is no longer useful for
   building the database.  The separate clist prevents the clist in the
   database from having to be recomputed every time (saving a lot of
   trouble.)

   Revision 1.1.1.1  1998/02/24 02:12:11  robbo
   First fie import.


	
   Copyright (c) 1997 by Robert Sim, Centre for Intelligent Machines,
   McGill University, Montreal, QC.  Please see the copyright notice
   included in this distribution for full details.

 ***********************************************************************/

#define VECTOR_CPP
#include "vector.h"
extern "C" {
#include <string.h>
#include <assert.h>
#include <math.h>
#if (defined sun || defined sgi)
extern void bcopy(char *b1, char *b2, int length);
#endif
}
vector::vector() {
  v=0;
  d=0;
}

vector::vector(unsigned int n) {
  d=n;
  if (d) {
    v=new PRECISION[d];
    memset(v,0,d*sizeof(PRECISION));
  } else v=0;
}

vector::vector(unsigned int n, PRECISION *values) {
  d=n;
  if (d) {
    v=new PRECISION[d];
    memset(v,0,d*sizeof(PRECISION));
  } else v=0;

  for (unsigned int i=0; i<n; i++)
    v[i]=values[i];
}

vector::vector(const vector& c) {
  d=c.d;
  v=new PRECISION[d];

  bcopy((char *)c.v,(char *)v,d*sizeof(PRECISION));
}

vector::vector(PRECISION x, PRECISION y) {
  d=2;
  v=new PRECISION[2];
  v[0]=x;
  v[1]=y;
}

vector::~vector() {
  if (d) {
    //cout << "DESTRUCTOR for " << this << "\n";
    //cout << "CONTENT =  " << *v << "  last = " << v[d-1] << "\n";
    if (v == NULL) cout << "SEVERE DESTRUCTOR ERROR!!!" << endl;
    delete []v;
  }
}

vector&
vector::operator=(const vector& c) {
  if (d) delete []v;
  d=c.d;
  if (d) {
    v=new PRECISION[d];
    bcopy((char *)c.v,(char *)v,d*sizeof(PRECISION));
  }
  return *this;
}



void
vector::copyVector(vector *v1){
  if (d) delete []v;
  d = v1->d;
  if (d){
    v=new PRECISION[d];
    memcpy((char *)v,(char*)v1->v,d*sizeof(PRECISION));
  }
}


PRECISION&
vector::operator[](unsigned int i) {
  if (!(i>=0 && i <d)) {
    cout << "!!!!!!!! i = " << i << "\n";
    assert(i>=0 && i<d);
  }
  assert(i>=0 && i<d);
  return v[i];
}

PRECISION
vector::operator[](unsigned int i) const {
  assert(i>=0 && i<d);
  return v[i];
}

PRECISION 
vector::sqr_length() const {
  PRECISION mag=0;
  for (unsigned int i=0; i<d; i++) {
    mag+=v[i]*v[i];
  }
  return mag;
}

PRECISION 
vector::length() const {
  return sqrt(sqr_length());
}

vector
vector::norm() const {
  assert(sqr_length()!=0); 
  return *this/length(); 
}

vector&
vector::operator+=(const vector& v1) {
  assert(d==v1.d);
  for (unsigned int i=0; i<d; i++) {
    v[i]+=v1.v[i];
  }
  return *this;
}

vector& 
vector::operator-=(const vector& v1) {
  assert(d==v1.d);
  for (unsigned int i=0; i<d; i++) {
    v[i]-=v1.v[i];
  }
  return *this;
}

vector  
vector::operator+(const vector& v1) const {
  assert(d==v1.d);
  vector r(d);
  for (unsigned int i=0; i<d; i++) {
    r[i]=v[i]+v1.v[i];
  }
  return r;
}


vector*
vector::assignDifference(vector const *v1,vector const *v2){
  assert(v1->d == v2->d);
  if (d!=v1->d){
    d = v1->d;
    if (v!=NULL) 
      delete []v;
    v = new PRECISION[d];
  }
  for (unsigned int i=0;i<d;i++) 
    v[i] = v1->v[i] - v2->v[i];
  return this;
}


vector  
vector::operator-(const vector& v1) const {
  assert(d==v1.d);
  vector r(d);
  for (unsigned int i=0; i<d; i++) {
    r[i]=v[i]-v1.v[i];
  }

  return r;
}

PRECISION  
vector::operator*(const vector& v1) const {
  PRECISION p=0;
  assert(d==v1.d);
  //vector r(d);
  for (unsigned int i=0; i<d; i++) {
    p+=v[i]*v1.v[i];
  }
  return p;
}

vector  
vector::operator*(PRECISION r) const {
  vector v1(d);
  for (unsigned int i=0; i<d; i++) {
    v1[i]=v[i]*r;
  }
  return v1;
}

vector  
vector::operator-() const {
  vector v1(d);
  for (unsigned int i=0; i<d; i++) {
    v1[i]=-v[i];
  }
  return v1;
}

vector  
vector::operator/(PRECISION r) const {
  vector v1(d);
  for (unsigned int i=0; i<d; i++) {
    v1[i]=v[i]/r;
  }
  return v1;
}

bool     
vector::operator==(const vector& w) const {
  if (d!=w.d) return false;
  for (unsigned int i=0; i<d; i++) {
    if (v[i]!=w.v[i]) return false;
  }
  return true;
}

void
vector::print(ostream& O) const {
  O << "| ";
  for (unsigned int i=0; i<d; i++) {
    O << v[i] << " ";
  }
  O << "|";
}

void 
vector::printBinary(ostream& O, int flag) const {
  char *pipe="|";
  

  O.write((char *)&d,sizeof(int));      // put the size of the vector
  O.write(pipe,sizeof(char));           // put the pipe
  if (flag) {
    for (unsigned int i=0; i<d; i++) {
      char x = (unsigned char)v[i];
      O.write(&x, sizeof(unsigned char)); // put the vector info
    }
  }
  else
    O.write((char *)v, d*sizeof(PRECISION)); // put the vector info
  O.write(pipe,sizeof(char));           // put the finishing pipe 
}


// note that read assumes the vector is already initialized.
void 
vector::read(istream& I) {
  char c;
  do {
    c=I.get();
  } while (c!='|');
  if (c=='|') {
    for (unsigned int i=0; i<d; i++) {
      I >> v[i];
    }
    do {
      c=I.get();
    } while (c!='|');
  }
}

void
vector::readBinary(istream& I, int flag) {
  char pipe;
  int vecDim;
  PRECISION *PRECISIONVals;
  unsigned char *charVals;

  if (d) delete []v;  

  I.read((char *)&vecDim,sizeof(int));
  d = vecDim;
  v = new PRECISION[vecDim];
  memset(v,0,d*sizeof(PRECISION));

  if (flag)
    charVals = new unsigned char[vecDim];
  else
    PRECISIONVals = new PRECISION[vecDim];

  I.read((char *)&pipe,sizeof(char));

  if (flag) {
    for (unsigned int i=0; i<d; i++)
      I.read((char *)&charVals[i],sizeof(unsigned char));
  }
  else
    I.read((char *)PRECISIONVals,vecDim*sizeof(PRECISION));

  for (int i=0; i<vecDim; i++) {
    if (flag)
      v[i] = (PRECISION)charVals[i];
    else
      v[i] = PRECISIONVals[i];
  }
  I.read((char *)&pipe,sizeof(char));
}


ostream& 
operator<<(ostream& O, const vector& v) {
  v.print(O);
  return O;
}

istream& 
operator>>(istream& I, vector& v) {
  v.read(I);
  return I;
}







