function marg = addmargin(im, left, up, right, down)
% ADDMARGIN	Add compact support (i.e., add zero-padded margins) to an 
%		image.
%		
%		ADDMARGIN(IM, LEFT, UP, RIGHT, DOWN)
%		IM - image.
%		LEFT, UP, RIGHT, DOWN - left, up, right, down margin sizes,
%		                        respectively.
%		
%		ADDMARGIN(IM, XMARG, YMARG)
%		IM - image.
%		XMARG - size of horizontal margins.
%		YMARG - size of vertical margins.
%		
%		ADDMARGIN(IM, MARG)
%		IM - image.
%		MARG - margin size. Margins are equal in all sides.

%
% Written by Ariel Tankus, 20.9.96.
%


if nargin == 2
    up    = left;
    right =  left;
    down  = left;
end
if nargin == 3
    right = left;
    down  = up;
end

[s1 s2] = size(im);

marg = [zeros(up,s2+left+right); ...
	zeros(s1,left), im, zeros(s1,right); ...
	zeros(down,s2+left+right)];
