function [a, mag, dx, dy] =  autoarg(im, xRadius, yRadius, thresh_factor)
% AUTOARG	Return the argument of image `im'.
%
%		[A, MAG, DX, DY] = AUTOARG(IM, XRADIUS, YRADIUS, ...
%			THRESH_FACTOR)
%		IM      - image
%		XRADIUS - half width of derivation vicinity.
%		YRADIUS - half height of derivation vicinity.
%		THRESH_FACTOR - Factor of the threshold.
%		
%		A   - Gradient argument.
%		MAG - Gradient magnitude (optional).
%		DX  - X-derivative (optional).
%		DY  - Y-derivative (optional).

%
% Written by Ariel Tankus, 19.9.96.
%


if nargin < 3, yRadius = xRadius; end

if nargin > 4
    [mag, dx, dy] = grad(im, xRadius, yRadius, thresh_factor);
else
    [mag, dx, dy] = grad(im, xRadius, yRadius);
end

a = arg(dx, dy, mag);
