function [mag, dx, dy] = grad(im, xRadius, yRadius, thresh_factor)
% GRAD		Return the gradient megnitude of the given image.
% 
%		[MAG, DX, DY] = GRAD(IM, XRADIUS, YRADIUS, THRESH_FACTOR)
%		IM - image
%		XRADIUS - half width of derivation vicinity.
%		YRADIUS - half height of derivation vicinity.
%		THRESH_FACTOR - Factor of the threshold.
%		
%		MAG - Gradient magnitude.
%		DX - X-derivative (optional).
%		DY - Y-derivative (optional).

%
% Written by Ariel Tankus, 19.9.96.
%


if nargin < 3, yRadius = xRadius; end

dx1 = xderiv(im, xRadius, yRadius);
dy1 = yderiv(im, xRadius, yRadius);

if nargout > 1; dx = dx1; end
if nargout > 2; dy = dy1; end

m = dx1.^2 + dy1.^2;

if nargin > 4
    % apply threshold
    m =  m .* ( m > ((max2(dx1)+max2(dy1)) ./ thresh_factor) );
end

mag = sqrt(m);
