// Author: Ariel Tankus.
// Created: Sept. 1996.

#include <iostream.h>

template<class T>
inline mymatrix<T>::mymatrix()
{
    c = r = 0;
    mat = NULL;
}

template<class T>
inline mymatrix<T>::mymatrix(int r1, int c1)
{
    c = c1;
    r = r1;
    mat = new T[c*r];
}

template<class T>
inline mymatrix<T>::mymatrix(int r1, int c1, T *m, transMthd t)
{
    c = c1;
    r = r1;

    mat = ((t == COPY) ? (T *)memcpy(new T[r*c], m, r*c*sizeof(T)) : m);
}

template<class T>
inline mymatrix<T>::mymatrix(const mymatrix& m)
{
    c = m.c;
    r = m.r;
    mat = (T *)memcpy(new T[r*c], m.mat, c*r*sizeof(T));
}

template<class T>
inline mymatrix<T>::~mymatrix()
{
    if (mat)
	delete [] mat;
}

template<class T>
inline T& mymatrix<T>::operator()(int r1, int c1) const
{
#ifdef RUNTIME_CHECK
    if (c1 < 0 || c1 >= c || r1 < 0 || r1 >= r) {
	cerr<<"inline T& mymatrix<T>::operator()(int r1, int c1) const: Error:\n"
	    <<"Attempt to access element ("<<r1<<","<<c1<<") in a "
	    <<r<<"x"<<c<<" mymatrix.\n";
	if (c1 == c || r1 == r)
	    cerr<<"Legal mymatrix indices are in: {0,...,"<<r-1<<"} x {0,...,"
		<<c-1<<"}\n";
	exit(1);
    }
#endif

    return mat[r1 + r*c1];
}


//
// Set all mymatrix elements to the specified scalar.
//
template<class T>
inline mymatrix<T>& mymatrix<T>::setAllElem(const T& scalar)
{
    register int size = r*c;
    register int i;
    for (i = 0; i < size; i++)
	this->mat[i] = scalar;

    return *this;
}

template<class T>
inline mymatrix<T>& mymatrix<T>::operator=(const mymatrix<T>& m)
{
#ifdef RUNTIME_CHECK
    if (c != m.c || r != m.r) {
	cerr<<"operator=: Error: Incorrect mymatrix dimensions\n";
	exit(1);
    }
#endif
    if (this != &m)
	memcpy(mat, m.mat, r*c*sizeof(T));
    return *this;
}

template<class T>
inline mymatrix<T>& mymatrix<T>::operator+=(const mymatrix<T>& m)
{
#ifdef RUNTIME_CHECK
    if (c != m.c || r != m.r) {
	cerr<<"mymatrix.operator+=(): Unequal mymatrix dimensions: ("<<c<<","<<r;
	cerr<<") vs. ("<<m.c<<","<<m.r<<") \n";
	exit(1);
    }
#endif

    int i;
    register int size = c*r;
    for (i = 0; i < size; i++)
	    mat[i] += m.mat[i];

    return *this;
}

template<class T>
inline mymatrix<T>& mymatrix<T>::operator+(const mymatrix<T>& m) const
{
    mymatrix<T>& p = *new mymatrix<T>(*this);
    p += m;

    return p;
}

//
// Sum the current mymatrix and mymatrix `m'. Put the result in `result'.
// result = (*this) + m .
//
// In-place calculation is allowed (i.e., maybe: result == *this).
//
template<class T>
mymatrix<T>& mymatrix<T>::add(const mymatrix<T>& m, mymatrix<T>& result) const
{
#ifdef RUNTIME_CHECK
    if (c != m.c || r != m.r) {
	cerr<<"mymatrix.add(): Unequal mymatrix dimensions: ("<<r<<","<<c;
	cerr<<") vs. ("<<m.r<<","<<m.c<<") \n";
	exit(1);
    }
    if (c != result.c || r != result.r) {
	cerr<<"mymatrix.add(): Wrong dimension for result mymatrix dimensions: (";
	cerr<<result.c<<","<<result.r<<") instead of ("<<c<<","<<r<<") \n";
	exit(1);
    }
#endif

    int i;
    register int size = c*r;
    for (i = 0; i < size; i++)
	result.mat[i] = mat[i] + m.mat[i];

    return result;
}

//
// Sum the current mymatrix and mymatrix `m'. Put the result in `result'.
// (Scalar is on the left hand side of the `+' operator. Might be important if
// summation is not associative.)
//
// In-place calculation is allowed (i.e., maybe: result == *this).
//
template<class T>
mymatrix<T>& mymatrix<T>::add(const T& scalar, mymatrix<T>& result) const
{
#ifdef RUNTIME_CHECK
    if (c != result.c || r != result.r) {
	cerr<<"mymatrix.add(const T&, mymatrix<T>&): Wrong dimension for result mymatrix dimensions: (";
	cerr<<result.c<<","<<result.r<<") instead of ("<<c<<","<<r<<") \n";
	exit(1);
    }
#endif

    int i;
    register int size = c*r;
    for (i = 0; i < size; i++)
	result.mat[i] = scalar + mat[i];

    return result;
}

//
// Subtract `m' from the current mymatrix. Put the result in `result'.
// result = (*this) - m .
//
// In-place calculation is allowed (i.e., maybe: result == *this).
//
template<class T>
mymatrix<T>& mymatrix<T>::sub(const mymatrix<T>& m, mymatrix<T>& result) const
{
#ifdef RUNTIME_CHECK
    if (c != m.c || r != m.r) {
	cerr<<"mymatrix.sub(): Unequal mymatrix dimensions: ("<<r<<","<<c;
	cerr<<") vs. ("<<m.r<<","<<m.c<<") \n";
	exit(1);
    }
    if (c != result.c || r != result.r) {
	cerr<<"mymatrix.sub(): Wrong dimension for result mymatrix dimensions: (";
	cerr<<result.c<<","<<result.r<<") instead of ("<<c<<","<<r<<") \n";
	exit(1);
    }
#endif

    int i;
    register int size = c*r;
    for (i = 0; i < size; i++)
	result.mat[i] = mat[i] - m.mat[i];

    return result;
}

//
// Subtract the current mymatrix and mymatrix `m' from the mymatrix 
// scalar*ones(size(m)). (actually: result = scalar - (*this)).
// Put the result in `result'.
// (Scalar is on the left hand side of the `-' operator !)
//
// In-place calculation is allowed (i.e., maybe: result == *this).
//
template<class T>
mymatrix<T>& mymatrix<T>::sub(const T& scalar, mymatrix<T>& result) const
{
#ifdef RUNTIME_CHECK
    if (c != result.c || r != result.r) {
	cerr<<"mymatrix.sub(const T&, mymatrix<T>&): Wrong dimension for result mymatrix dimensions: (";
	cerr<<result.c<<","<<result.r<<") instead of ("<<c<<","<<r<<") \n";
	exit(1);
    }
#endif

    int i;
    register int size = c*r;
    for (i = 0; i < size; i++)
	result.mat[i] = scalar - mat[i];

    return result;
}

//
// Multiply 2 matrices, element by element.
//
template<class T>
inline mymatrix<T>& mymatrix<T>::elemMult(const mymatrix<T>& m,
				      mymatrix<T>& result) const
{
#ifdef RUNTIME_CHECK
    if (c != m.c || r != m.r) {
	cerr<<"mymatrix.elemMult(): Error: Unequal mymatrix dimensions: ("<<r;
	cerr<<","<<c<<") vs. ("<<m.r<<","<<m.c<<") \n";
	exit(1);
    }
    if (c != result.c || r != result.r) {
	cerr<<"mymatrix.elemMult(): Error: Wrong result mymatrix dimensions: (";
	cerr<<result.r<<","<<result.c<<") instead of ("<<r<<","<<c<<") \n";
	exit(1);
    }
#endif

    int i;
    register int size = c*r;
    for (i = 0; i < size; i++)
	    result.mat[i] = mat[i] * m.mat[i];

    return result;
}

//
// Multiply the mymatrix by a scalar, element by element.
// (Scalar is the left multiplicand. Might be important if
// summation is not associative.)
//
// In-place calculation is allowed (i.e., maybe: result == *this).
//
template<class T>
inline mymatrix<T>& mymatrix<T>::mult(const T& scalar, mymatrix<T>& result) const
{
#ifdef RUNTIME_CHECK
    if (c != result.c || r != result.r) {
	cerr<<"mymatrix.mult(T, mymatrix<T>&): Error: Wrong result mymatrix dimensions: (";
	cerr<<result.r<<","<<result.c<<") instead of ("<<r<<","<<c<<") \n";
	exit(1);
    }
#endif

    int i;
    register int size = c*r;
    for (i = 0; i < size; i++)
	result.mat[i] = scalar * mat[i];

    return result;
}

//
// Naive algorithm for mymatrix multiplication.
//
template<class T>
inline mymatrix<T>& mymatrix<T>::mult(const mymatrix<T>& m, mymatrix<T>& result) const
{
#ifdef RUNTIME_CHECK
    if (c != m.r) {
	cerr<<"mymatrix.mult(): Error: Inappropriate mymatrix dimensions: \n";
	cerr<<"               left mymatrix has "<<c<<" columns, \n";
	cerr<<"               while right mymatrix has "<<m.r<<" rows\n";
	exit(1);
    }
    if (m.c != result.c || r != result.r) {
	cerr<<"mymatrix.mult(): Error: Wrong result mymatrix dimensions: (";
	cerr<<result.r<<","<<result.c<<") instead of ("<<r<<","<<m.c<<")\n";
	exit(1);
    }
#endif

    register int i, j, k;
    register T tmp;
    register T *res_ptr = result.mat;
    register T *mat_ptr, *m_mat_ptr;		// optimization temporaries

    for (i = 0; i < r; i++) {

	mat_ptr = mat + c*i;

	for (j = 0; j < m.c; j++) {

	    tmp = 0;
	    m_mat_ptr = m.mat + j;
	    for (k = 0; k < c; k++) {
		// optimization for: tmp += mat[k+c*i] * m.mat[j+m.c*k];
		tmp += (*(mat_ptr + k)) * (*(m_mat_ptr + m.c*k));
	    }
	    *(res_ptr++) = tmp;

	}
    }

    return result;
}

//
// The product of the transpose of the current mymatrix (left) and the mymatrix m
// (right).
//
//  result == transpose(*this) * m
//
template<class T>
inline mymatrix<T>& mymatrix<T>::transposeMult(const mymatrix<T>& m,
					   mymatrix<T>& result) const
{
#ifdef RUNTIME_CHECK
    if (r != m.r) {
	cerr<<"mymatrix.transposeMult():\n";
	cerr<<"Error: Inappropriate mymatrix dimensions: \n";
	cerr<<"       left mymatrix has "<<r<<" rows, \n";
	cerr<<"       while right mymatrix has "<<m.r<<" rows\n";
	exit(1);
    }
    if (result.r != c || result.c != m.c) {
	cerr<<"mymatrix.transposeMult():\n";
	cerr<<"Error: Wrong result mymatrix dimensions:\n";
	cerr<<"       ("<<result.r<<","<<result.c<<") instead of ";
	cerr<<"("<<c<<","<<m.c<<")\n";
	exit(1);
    }
#endif

    register int i, j, k;
    register T tmp;
    register T *res_ptr = result.mat;
    register T *mat_ptr, *m_mat_ptr;		// optimization temporaries

    for (i = 0; i < c; i++) {
	mat_ptr = mat + i;
	for (j = 0; j < m.c; j++) {
	    tmp = 0;
	    m_mat_ptr = m.mat + j;
	    for (k = 0; k < r; k++) {
		// optimization for: tmp += mat[i+c*k] * m.mat[j+m.c*k];
		// r == m.r
		tmp += (*(mat_ptr + c*k)) * (*(m_mat_ptr + m.c*k));
	    }
	    *(res_ptr++) = tmp;
	}
    }

    return result;
}

//
// The product of the transpose of the current mymatrix (left) and the mymatrix m
// (right).
//
//  result == (*this) * transpose(m)
//
template<class T>
inline mymatrix<T>& mymatrix<T>::multTranspose(const mymatrix<T>& m,
					   mymatrix<T>& result) const
{
#ifdef RUNTIME_CHECK
    if (c != m.c) {
	cerr<<"mymatrix.multTranspose():\n";
	cerr<<"Error: Inappropriate mymatrix dimensions: \n";
	cerr<<"       left mymatrix has "<<c<<" columns, \n";
	cerr<<"       while right mymatrix has "<<m.c<<" columns\n";
	exit(1);
    }
    if (result.r != r || result.c != m.r) {
	cerr<<"mymatrix.multTranspose():\n";
	cerr<<"Error: Wrong result mymatrix dimensions:\n";
	cerr<<"       ("<<result.r<<","<<result.c<<") instead of ";
	cerr<<"("<<r<<","<<m.r<<")\n";
	exit(1);
    }
#endif

    register int i, j, k;
    register T tmp;
    register T *res_ptr = result.mat;
    register T *mat_ptr, *m_mat_ptr;		// optimization temporaries

    for (i = 0; i < r; i++) {
	mat_ptr = mat + c*i;
	for (j = 0; j < m.r; j++) {
	    tmp = 0;
	    m_mat_ptr = m.mat + m.c*j;
	    for (k = 0; k < c; k++) {
		// optimization for: tmp += mat[k+c*i] * m.mat[k+m.c*j];
		// c == m.c
		tmp += (*(mat_ptr + k)) * (*(m_mat_ptr + k));
	    }
	    *(res_ptr++) = tmp;
	}
    }

    return result;
}


//
// Transpose of a mymatrix.
//
template<class T>
inline mymatrix<T>& mymatrix<T>::transpose(mymatrix<T>& result) const
{
#ifdef RUNTIME_CHECK
    if (c != result.r || r != result.c) {
	cerr<<"mymatrix.transpose(): Error: Wrong result mymatrix dimensions: (";
	cerr<<result.r<<","<<result.c<<") instead of ("<<c<<","<<r<<") \n";
	exit(1);
    }
#endif

    int i, j;

    for (i = 0; i < c; i++)
	for (j = 0; j < r; j++)
	    result.mat[j+r*i] = mat[i+c*j];

    return result;
}

//
// Divide the current mymatrix by mymatrix `m', element by element.
//
template<class T>
inline mymatrix<T>& mymatrix<T>::elemDiv(const mymatrix<T>& m,
				     mymatrix<T>& result) const
{
#ifdef RUNTIME_CHECK
    if (c != m.c || r != m.r) {
	cerr<<"mymatrix.elemDiv(): Error: Unequal mymatrix dimensions: ("<<r;
	cerr<<","<<c<<") vs. ("<<m.r<<","<<m.c<<") \n";
	exit(1);
    }
    if (c != result.c || r != result.r) {
	cerr<<"mymatrix.elemMult(): Error: Wrong result mymatrix dimensions: (";
	cerr<<result.r<<","<<result.c<<") instead of ("<<r<<","<<c<<") \n";
	exit(1);
    }
#endif

    int i;
    register int size = c*r;
    for (i = 0; i < size; i++)
	    result.mat[i] = mat[i] / m.mat[i];

    return result;
}

//
// sumRows:
// Sum the elements in each row of the mymatrix.
//
template<class T>
inline mymatrix<T>& mymatrix<T>::sumRows(mymatrix<T>& result) const
{
#ifdef RUNTIME_CHECK
    if (result.r != r || result.c != 1) {
	cerr<<"mymatrix<T>::sumRows(): Error: ";
	cerr<<"Incorrect result mymatrix dimensions;\n";
	cerr<<"should be: ("<<r<<",1)   ";
	cerr<<"(currently: ("<<result.r<<","<<result.c<<"))\n";
	exit(1);
    }
#endif

    int i, j;
    T *row_i;
    for (i = 0; i < r; i++) {
	result.mat[i] = 0;
	row_i = mat + i*c;
	for (j = 0; j < c; j++)
	    // optimization for: result(i) += mat[j+c*i];
	    result.mat[i] += row_i[j];
    }

    return result;
}

//
// sumRows:
// Sum the elements in each columns of the mymatrix.
//
template<class T>
inline mymatrix<T>& mymatrix<T>::sumColumns(mymatrix<T>& result) const
{
#ifdef RUNTIME_CHECK
    if (result.r != 1 || result.c != c) {
	cerr<<"mymatrix<T>::sumColumns(): Error: ";
	cerr<<"Incorrect result mymatrix dimensions;\n";
	cerr<<"should be: (1,"<<c<<")   ";
	cerr<<"(currently: ("<<result.r<<","<<result.c<<"))\n";
	exit(1);
    }
#endif

    int i, j;
    T *column_i;
    for (i = 0; i < c; i++) {
	result.mat[i] = 0;
	column_i = mat + i;
	for (j = 0; j < r; j++)
	    // optimization for: result(i) += mat[i+c*j];
	    result.mat[i] += column_i[c*j];
    }

    return result;
}

//
// maxOnRow:
// Return the maximum value attained in the given row.
// Optionally returns the column where the maximum was attained.
//
template<class T>
inline T mymatrix<T>::maxOnRow(int row, int *col) const
{
    int maxInd = 0;		// index (i.e., col) where maximum attained.
    T maxVal = mat[row*c];	// maximal value

    for (int i = 1; i < c; i++) {
	if (maxVal < mat[i+row*c]) {
	    maxVal = mat[i+row*c];
	    maxInd = i;
	}
    }

    if (col)
	*col = maxInd;

    return maxVal;
}

//
// maxOnColumn:
// Return the maximum value attained in the given column.
// Optionally returns the row where the maximum was attained.
//
template<class T>
inline T mymatrix<T>::maxOnColumn(int col, int *row) const
{
    int maxInd = 0;		// index (i.e., row) where maximum attained.
    T maxVal = mat[col];	// maximal value

    for (int i = 1; i < r; i++) {
	if (maxVal < mat[col+i*c]) {
	    maxVal = mat[col+i*c];
	    maxInd = i;
	}
    }

    if (row)
	*row = maxInd;

    return maxVal;
}


//
// secondMaxOnColumn:
// Return the 2 maximal values attained in the given column,
// and the rows where the maximal values were attained.
//
// row1, val1 - row and value of the maximum.
// row2, val2 - row and value of the second maximum (i.e., maximum of the
//              set excluding the maximum).
//
template<class T>
inline void mymatrix<T>::secondMaxOnColumn(int col, int *row1, T *val1,
					 int *row2, T *val2) const
{
    if (r == 1) {
	*row1 = *row2 = 0;
	*val1 = *val2 = mat[col];
	return;
    }

    // here: at least 2 rows exist

    int maxInd;			// index (i.e, row) where maximum attained.
    T maxVal;			// maximal value.
    int max2Ind;		// index (i.e, row) where 2nd maximum attained.
    T max2Val;			// 2nd maximal value.

    // initialization
    if (mat[col] > mat[col+1]) {

	maxInd = 0;
	maxVal = mat[col];

	max2Ind = 1;
	max2Val = mat[col+c];

    } else {

	maxInd = 1;		// index (i.e., row) where maximum attained.
	maxVal = mat[col+c];	// maximal value

	max2Ind = 0;		// index (i.e., row) where 2nd maximum attained
	max2Val = mat[col];	// 2nd maximal value

    }

    for (int i = 2; i < r; i++) {
	if (maxVal < mat[col+i*c]) {

	    max2Val = maxVal;
	    max2Ind = maxInd;

	    maxVal = mat[col+i*c];
	    maxInd = i;

	} else if (max2Val < mat[col+i*c]) {

	    max2Val = mat[col+i*c];
	    max2Ind = i;

	}
    }

    // return results (requested ones only)
    if (row1)
	*row1 = maxInd;

    if (val1)
	*val1 = maxVal;

    if (row2)
	*row2 = max2Ind;

    if (val2)
	*val2 = max2Val;
}

template<class T>
inline T mymatrix<T>::sumSqr() const
{
    register int size = r*c;
    T sum = 0;
    for (int i = 0; i < size; i++)
	sum += mat[i] * mat[i];

    return sum;
}

template<class T>
inline T mymatrix<T>::meanSqr() const
{
    register int size = r*c;
    T sum = 0;
    for (int i = 0; i < size; i++)
	sum += mat[i] * mat[i];

    return sum / size;
}

//
// splitVert:
// (*this) is refered to as if it were an interleaving of 2 matrices,
// where each `interval1' columns from the first mymatrix are followed by
// `interval2' columns from the second mymatrix.
//
// The interleaving is considered exact, i.e.: the number of rows in (*this)
// is a multiplicand of (interval1 + interval2).
//
// These original mymatrix are returned in `mat1' and `mat2' respectively.
// (`mat1' and `mat2' be of the correct sizes to hold resulting matrices!)
//
template<class T>
void mymatrix<T>::splitVert(int interval1, int interval2, mymatrix<T>& mat1,
			  mymatrix<T>& mat2) const
{

    int cycle = interval1 + interval2;         // cycle of adjacent columns
                                               // from mat1 & mat2 in (*this)

#ifdef RUNTIME_CHECK
    if (c % cycle != 0) {
	// inexact split
	cerr<<"inline T mymatrix<T>::splitVert(int interval1, int interval2,\n"
	    <<"                              mymatrix<double>& mat1, "
	    <<"mymatrix<double>& mat2) const\n"
	    <<"Error: Attempt of inexact split: Interval sizes cannot be\n"
	    <<"used for an exact split\n";
	exit(1);
    }
    if (mat1.rows() != r || mat1.columns() != c / cycle * interval1 ||
	mat2.rows() != r || mat2.columns() != c / cycle * interval2) {
	cerr<<"inline T mymatrix<T>::splitVert(int interval1, int interval2,\n"
	    <<"                              mymatrix<double>& mat1, "
	    <<"mymatrix<double>& mat2) const\n"
	    <<"Error: Result mymatrix have inappropriate size\n";
	exit(1);
    }
#endif

    int i, j;
    mymatrix<double> *currentMat;                // points to either mat1 or mat2

    for (i = 0; i < c/cycle; i++) {
	// decide the mymatrix for current column
	currentMat = (i % cycle < interval1 ? &mat1 : &mat2);
	for (j = 0; j < r; j++)
	    mat1.mat[i+j*c] = mat[i+j*c];
    }
}

//
// Output mymatrix to stream.
//
template<class T>
inline ostream& operator<<(ostream& s, const mymatrix<T>& m)
{
    int i, j;
    for (i = 0; i < m.r; i++) {
	for (j = 0; j < m.c; j++)
	    s<<m(i,j)<<" ";
	s<<"\n";
    }

    return s;
}

//
// Output mymatrix header to stream.
//
template<class T>
inline ostream& header(ostream& s, const mymatrix<T>& m)
{
    s<<m.r<<" "<<m.c<<"\n";
}

//
// Output mymatrix in raw format to stream. Header containing size of mymatrix
// is also printed.
//
template<class T>
ostream& rawWrite(ostream& s, const mymatrix<T>& m)
{
    s.write((char *)&m.r, sizeof(int));
    s.write((char *)&m.c, sizeof(int));
    if (s.fail())
	cerr<<"rawWrite(): Error: Writing stream header failed\n";

    s.write((char *)m.mat, m.r * m.c * sizeof(T));
    if (s.fail())
	cerr<<"rawWrite(): Warning: Writing stream raw data failed\n";

    return s;
}

//
// Input mymatrix in raw format to stream. Header containing size of mymatrix
// is also printed.
//
template<class T>
istream& rawRead(istream& s, mymatrix<T>& m)
{
    s.read((char *)&m.r, sizeof(int));
    s.read((char *)&m.c, sizeof(int));
    if (m.r <= 0 || m.c <= 0 || s.fail()) {
	cerr<<"rawRead(): Warning: Reading stream header failed\n";
	return s;
    }

    if (m.mat)
	delete [] m.mat;
    m.mat = new T[m.r*m.c];
    s.read((char *)m.mat, m.r * m.c * sizeof(T));

    if (s.fail()) {
	delete [] m.mat;
	m.mat = NULL;
	cerr<<"rawRead(): Warning: Reading stream of raw data failed\n";
    }

    return s;
}
