// mymatrix.h - mymatrix declaration header file

// Author: Ariel Tankus.
// Created: Sept. 1996.

#ifndef __MYMATRIX_H_
#define __MYMATRIX_H_

#include <iostream.h>

//  #if defined linux
//  #include </usr/include/g++/iostream.h>
//  #endif

//  #if defined sgi
//  #include <iostream.h>
//  #endif

template<class T>
class vector;

enum transMthd {COPY, MOVE};

//
// The indices of elements in the mymatrix begins at 0.
//
template<class T>
class mymatrix {
protected:
    int c;		// #columns.
    int r;		// #rows.
    T *mat;

public:
    mymatrix();
    mymatrix(int r1, int c1);
    mymatrix(int r1, int c1, T *m, transMthd t = COPY);
    mymatrix(const mymatrix& m);

    ~mymatrix();

    T& operator()(int r1, int c1) const;
    int columns(void) const { return c; }
    int rows(void)    const { return r; }
    int size(void)    const { return c*r; }
    bool isEqualDim(const mymatrix<T>& m) const
        { return ((r == m.r) && (c == m.c)); }

    mymatrix<T>& setAllElem(const T& scalar);

    inline mymatrix<T>& operator=(const mymatrix<T>& m);
    inline mymatrix<T>& operator+=(const mymatrix<T>& m);
    inline mymatrix<T>& operator+(const mymatrix<T>& m) const;
    bool operator!() const { return (!mat || !c || !r); }
    operator const T *() const { return mat; }
    operator T *() const
        { return (T *)memcpy(new T[r*c], mat, r*c*sizeof(T)); }

    inline mymatrix<T>& add(const mymatrix<T>& m, mymatrix<T>& result) const;
    inline mymatrix<T>& add(const T& scalar, mymatrix<T>& result) const;

    inline mymatrix<T>& sub(const mymatrix<T>& m, mymatrix<T>& result) const;
    // result = scalar - (*this)
    inline mymatrix<T>& sub(const T& scalar, mymatrix<T>& result) const;

    inline mymatrix<T>& elemMult(const mymatrix<T>& m, mymatrix<T>& result) const;
    inline mymatrix<T>& mult(const T& scalar, mymatrix<T>& result) const;
    inline mymatrix<T>& mult(const mymatrix<T>& m, mymatrix<T>& result) const;
    inline mymatrix<T>& transposeMult(const mymatrix<T>& m,
				    mymatrix<T>& result) const;
    inline mymatrix<T>& multTranspose(const mymatrix<T>& m,
				    mymatrix<T>& result) const;
    inline mymatrix<T>& transpose(mymatrix<T>& result) const;
    inline mymatrix<T>& elemDiv(const mymatrix<T>& m, mymatrix<T>& result) const;

    inline mymatrix<T>& sumRows(mymatrix<T>& result) const;
    inline mymatrix<T>& sumColumns(mymatrix<T>& result) const;
    inline T maxOnRow(int row, int *col) const;
    inline T maxOnColumn(int col, int *row) const;
    inline void secondMaxOnColumn(int col, int *row1, T *val1, int *row2,
				  T *val2) const;

    inline T sumSqr() const;
    inline T meanSqr() const;

    void splitVert(int interval1, int interval2, mymatrix<T>& mat1,
		   mymatrix<T>& mat2) const;

    friend inline ostream& header(ostream& s, const mymatrix<T>& m);
    friend inline ostream& operator<<(ostream& s, const mymatrix<T>& m);
    friend ostream& rawWrite(ostream& s, const mymatrix<T>& m);
    friend istream& rawRead(istream& s, mymatrix<T>& m);
};

#include "mymatrix.c"

#endif		// ndef __MYMATRIX_H_
