// Author: Ariel Tankus.
// Created: Sept. 1996.
// Last modified: 26.3.1999.

#include <math.h>
#include <values.h>
#include "mex.h"
#include "general.h"

/* Input Arguments */
#define	DX_IN		prhs[0]
#define	DY_IN		prhs[1]
#define	MAG_IN		prhs[2]
/* Output Arguments */
#define	THIN_OUT	plhs[0]


static void non_maximal_suppression(double *thin, double *Dx, double *Dy,
				    double *mag, int height, int size)
{
    register double F = tan(M_PI/8);
    
    do {
	*(thin++) = *mag &&
	    (fabs(*Dy) < F*fabs(*Dx) ? (*mag>=mag[height] && *mag>=mag[-height]) :
	     fabs(*Dx) < F*fabs(*Dy) ? (*mag>=mag[1] && *mag>=mag[-1]) :
	     *Dx * *Dy  > 0 ? (*mag>=mag[height+1] && *mag>=mag[-height-1]) :
	     (*mag>=mag[1-height] && *mag>=mag[height-1])) ? *mag : 0.0;
	Dx++;
	Dy++;
	mag++;
    } while (--size);
    
    return;
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    double *dx, *dy, *mag;	/* input arguments. */
    double *thin;		/* output argument. */
    unsigned int dx_m, dx_n, dy_m, dy_n, mag_m, mag_n;	/* dimensions. */
    
    /* Check for proper number of arguments */
    if (nrhs != 3)
	mexErrMsgTxt("NMS requires three input arguments.");
    else if (nlhs > 1)
	mexErrMsgTxt("NMS requires one output argument.");

    /* Check the dimensions.  All input arguments must be of the same size. */
    dx_m = mxGetM(DX_IN);
    dx_n = mxGetN(DX_IN);
    dy_m = mxGetM(DY_IN);
    dy_n = mxGetN(DY_IN);
    mag_m = mxGetM(MAG_IN);
    mag_n = mxGetN(MAG_IN);
    if (!isNormal(DX_IN) || !isNormal(DY_IN) || !isNormal(MAG_IN) ||
	(dx_m != dy_m) || (dx_m != mag_m) || 
	(dx_n != dy_n) || (dx_n != mag_n))
	mexErrMsgTxt("NMS requires that DX, DY and MAG be vectors of the same size.");


    /* Create a matrix for the return argument */
    THIN_OUT = mxCreateDoubleMatrix(dx_m, dx_n, mxREAL);


    /* Assign pointers to the various parameters */

    thin = mxGetPr(THIN_OUT);

    dx  = mxGetPr(DX_IN);
    dy  = mxGetPr(DY_IN);
    mag = mxGetPr(MAG_IN);


    /* Do the actual computations in a subroutine */
    non_maximal_suppression(thin, dx, dy, mag, dx_m, dx_m*dx_n);

    return;
}
