function [im]=readpgm(fname)
% READPGM Read pgm files.
%	im=readpgm('filename'), read the pgm file and store it in
%	MATLAB mat format, which enable display the image using IMAGE.
%	im=readpgm, will ask the file name from the user.
%
%	it is possible to read binary/ascii image, hence it is not recommended
%	to use ascii pgm due to its size.
%
%   Please note that the default resolution on matlab is poor (and in
%     color) so before (or after) image you should set the color 
%     map - colormap(gray(256))
%
%	See also SAVEPGM.
 
%	O. Mandel 23.11.93
%	School  of  Mathematics and Computer Science, 
%	Tel-Aviv University, Israel.
	
if nargin == 0,
  clc
  fname = input('input pgm-filename : ','s');
  if isempty(fname), 
        return;
  end
end
%
fp=fopen(fname);
magicnum=fscanf(fp,'%s',1);     % P2 - ascii,   P5 - binary
word=fscanf(fp,'%s',1);
while word(1)=='#',	% remark  skip it
  x=0;
  while x~=10,
    x=fscanf(fp,'%c',1);
  end
  word=fscanf(fp,'%s',1);
end
width=sscanf(word,'%d',1);
word=fscanf(fp,'%s',1);
while word(1)=='#',	% remark  skip it
  x=0;
  while x~=10,
    x=fscanf(fp,'%c',1);
  end
  word=fscanf(fp,'%s',1);
end
hight=sscanf(word,'%d',1);
word=fscanf(fp,'%s',1);
while word(1)=='#',	% remark  skip it
  x=0;
  while x~=10,
    x=fscanf(fp,'%c',1);
  end
  word=fscanf(fp,'%s',1);
end
maxgray=sscanf(word,'%d',1);
word=fscanf(fp,'%c',1);		% skip newline
if magicnum=='P5'               % binary
  tmp=fread(fp,[width,hight],'uchar');
  fclose(fp);
  im=tmp';
else
  if magicnum=='P2'		% ascii
    tmp=fscanf(fp,'%d',[width,hight]);
    fclose(fp);
    im=tmp';
  else
    fprintf('Illegal magicnum\n');
    return;
  end
end

