function d = xDeriv(im, xRadius, yRadius)
% XDERIV	Returns the X-derivative of image `im'.
%		D = XDERIV(IM, XRADIUS, YRADIUS)
%		IM      - Input image.
%		XRADIUS - half the width of the vicinity in which the 
%		          derivative is calculated.
%		YRADIUS - half the height of the vicinity in which the 
%		          derivative is calculated (default: equal to XRADIUS).
%

%
% Written by Ariel Tankus, 19.9.96.
%


if nargin < 3, yRadius = xRadius; end

if ((xRadius < 0) | (yRadius < 0))
    error('Radii must be positive!')
end

xRadius = 2*xRadius;
yRadius = 2*yRadius;

x = -1:(2/xRadius):1;
y = -1:(2/yRadius):1;

% Gaussian at [-1,1] with mean = 0; std. dev.=1/3, sampling it in 
% `xRadius' positions.  Same applies to dgauss.
xDGauss = dgauss(x, 1/3, 0);
yGauss = gauss(y, 1/3, 0);

d = conv2(conv2(im, xDGauss, 'valid'), yGauss', 'valid');
d = addmargin(d, ceil(xRadius/2), ceil(yRadius/2), floor(xRadius/2), ...
    floor(yRadius/2));
