function d = yDeriv(im, xRadius, yRadius)
% YDERIV	Returns the Y-derivative of image `im'.
%		D = YDERIV(IM, XRADIUS, YRADIUS)
%		IM      - Input image.
%		XRADIUS - half the width of the vicinity in which the 
%		          derivative is calculated.
%		YRADIUS - half the height of the vicinity in which the 
%		          derivative is calculated.
%                    	     

%
% Written by Ariel Tankus, 19.9.96.
%

if nargin < 3, yRadius = xRadius; end

if ((xRadius <= 0) | (yRadius <= 0))
    error('Radii must be positive!')
end

xRadius = 2*xRadius + 1;
yRadius = 2*yRadius + 1;

x = -1:(2/xRadius):1;
y = -1:(2/yRadius):1;

% Gaussian at [-1,1] with mean = 0; std. dev.=1/3, sampling it in 
% `yRadius' positions.  Same applies to dgauss.
xGauss = gauss(x, 1/3, 0);
yDGauss = dgauss(y, 1/3, 0);

d = conv2(conv2(im, xGauss, 'valid'), yDGauss', 'valid');
d = addmargin(d, ceil(xRadius/2), ceil(yRadius/2), floor(xRadius/2), ...
    floor(yRadius/2));
