


/* This file is basically a modification of   */
/*      $Revision: 1.1 $ */
/*
 *	engdemo.c
 *
 *	This is a simple program that illustrates how to call the MATLAB
 *	Engine functions from a C program.
 *
 *      Copyright (c) 1996 The MathWorks, Inc.
 *      All rights reserved
 */

/* 

   This should get execed with the correct arguments to run the matlab symmetry operator 
   such that it knows where the input pgm file is and where it can spew all the interest
   point to.  It can also recieve another argument which says where all the matlab code 
   is.

   Needs to be compiled with a gcc supporting libc5.  However should be able to run it 
   on any linux box.

*/
   

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
// For the line below to be happy your LD_LIBRARY_PATH should be:
// LD_LIBRARY_PATH=/local/matlab5/extern/lib/lnx86
#include "engine.h"

#define DEFAULT_OPERATOR_PATH "/home/leonardo/deepu/pcaproject/deeppca/src/symmetryOperator"
#define DEFAULT_RUN_BUF "matlab5" // ie run locally

void main(int argc, char **argv)
{
  char infile[255]="";
  char outfile[255]="";
  char operatorPath[255]=DEFAULT_OPERATOR_PATH;
  Engine *ep;
  //  mxArray *T = NULL, *result = NULL;
  char buffer[256], str[256];
  char conditionalBuf[1024];
  char runBuf[255]=DEFAULT_RUN_BUF;
  char hostname[255];
  pid_t mypid = getpid();
  
  gethostname(hostname,255);

  if (argc < 3 || argc > 5)
    {
      fprintf(stderr,"%s <inputfile> <outputfile> [remote machine (optional)] [working directory for attention operator (optional)]\n",argv[0]); 
      exit(-1);
    }
  else // argc == 3 or 4
    {
      strcpy(infile,argv[1]);
      strcpy(outfile,argv[2]);
      if (argc == 4) 
	{
	  strcpy(runBuf,"");
	  strcpy(runBuf,"rsh ");
	  strcat(runBuf,argv[3]);
	  strcat(runBuf," ");
	  strcat(runBuf,"\"/bin/csh -c 'setenv DISPLAY ");
	  strcat(runBuf,hostname);
	  strcat(runBuf,":0; /local/matlab5/bin/matlab5'\"");
	}
      if (argc == 5)
	strcpy(operatorPath,argv[4]);
    }

  /*
	 * Start the MATLAB engine locally by executing the string
	 * "matlab"
	 *
	 * To start the session on a remote host, use the name of
	 * the host as the string rather than \0
	 *
	 * For more complicated cases, use any string with whitespace,
	 * and that string will be executed literally to start MATLAB
	 */
  //  if (!(ep = engOpen("\0"))) {
  if (!(ep = engOpen(runBuf))) {
    fprintf(stderr, "\nCan't start MATLAB engine with %s\n",runBuf);
    exit(-1);
  }

  /*
   * Use engOutputBuffer to capture MATLAB output, so we can
   * echo it back.
   */
  engOutputBuffer(ep, buffer, 256);


  // Place the calls which are required by the attention operator.
  // Note that all the matlab files (ie .m files) should be present in
  // the directory where this is being run
  
  strcpy(str,"cd ");
  strcat(str,operatorPath);
  strcat(str,";");
  engEvalString(ep,str);
  strcpy(str,"im = readpgm('");
  strcat(str,infile);
  strcat(str,"');");
  engEvalString(ep,str);
  // Ugly way to check for the error but what the hell
  if (buffer[4] == '?')
    {
      fprintf(stderr,"Matlab Error: couldn't find file %s\n",infile);
      fprintf(stderr,"%s",buffer);
      exit(-1);
    }
  // scale issues should be addressed here as in 60, 60 is the size 
  // that is completely related to the size of the objects in the input
  // pgm image.
  engEvalString(ep,"[a, mag, dx, dy] = autoarg(im, 5, 5);");
  // engEvalString(ep,"[a, mag, dx, dy] = autoarg(im, 30, 30);");
  engEvalString(ep,"lmag = log1p(mag);");
  //  engEvalString(ep,"[radi, symag, symori] = radial(30, 30, 0, dx, dy, mag, lmag, a);");
  engEvalString(ep,"[radi, symag, symori] = radial(100, 100, 0, dx, dy, mag, lmag, a);");
  engEvalString(ep,"[I,J] = size(radi);");
  //  engEvalString(ep,"if (I > 10000) I=10000; end;");
  //  engEvalString(ep,"if (J > 10000) J=10000; end;");
  engEvalString(ep,"bigmat = zeros(I*J,5);");
  engEvalString(ep,"row=1;");
  strcpy(conditionalBuf,"for i=1:I");
  strcat(conditionalBuf,"  for j=1:J");
  strcat(conditionalBuf,"    if (radi(i,j) > 1)");
  strcat(conditionalBuf,"      bigmat(row,1) = j;");
  strcat(conditionalBuf,"      bigmat(row,2) = i;");
  strcat(conditionalBuf,"      bigmat(row,3) = radi(i,j);");
  strcat(conditionalBuf,"      bigmat(row,4) = symag(i,j);");
  strcat(conditionalBuf,"      bigmat(row,5) = symori(i,j);");
  strcat(conditionalBuf,"      row = row+1;");
  strcat(conditionalBuf,"    end;");
  strcat(conditionalBuf,"  end;");
  strcat(conditionalBuf,"end;");
  engEvalString(ep,conditionalBuf);
  engEvalString(ep,"bigmat(row:I*J,:) = [];");
  engEvalString(ep,"allsorted = sortrows(bigmat,3);");
  engEvalString(ep,"flippedallsorted = flipdim(allsorted,1);");
  strcpy(conditionalBuf,"if (row > 5000) ");
  strcat(conditionalBuf,"  flippedallsorted(5000:row-1,:) = [];");  // keeps 5000 points
  strcat(conditionalBuf,"end;");
  engEvalString(ep,conditionalBuf);
  strcpy(str,"data = fopen('");
  strcat(str,outfile);
  strcat(str,"','w');");
  engEvalString(ep,str);
  printf("%s",buffer);
  sprintf(str,"fprintf(data,'#3\\t(%%d,%%d)\\t[%%10.10f %%10.10f %%10.10f]\\n',flippedallsorted');");
  engEvalString(ep,str);
  engEvalString(ep,"fclose(data);");

  engClose(ep);

  exit(mypid);
}


