/************************************************************************

   File:		symVer.cpp

   Author(s):		Deeptiman Jugessur

   Created:		24 May 1999

   Last Revision:	$Date$

   Description:	

   $Revision$

   $Log$

	
   Copyright (c) 1999 by Deeptiman Jugessur, Centre for Intelligent Machines,
   McGill University, Montreal, QC.  Please see the copyright notice
   included in this distribution for full details.


   -- Have to check getNClosest
   -- Have to merge close eigenpoints.

 ***********************************************************************/

#include <stdio.h>
#include <limits.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <fstream.h>
#include <assert.h>
#include <iostream.h>
// my headers
#include <PPM.h>
#include <PCA.h>
#include <PrincipalComponents.h>
#include <interestPt.h>
#include <ioData.h>
#include <kdtree.h>
#include <fft.h>

#define ARR_SIZE 256
#define DEFAULT_NUM_INTEREST_PTS 10
#define DEFAULT_SUBWINDOW_WIDTH 40
#define DEFAULT_SUBWINDOW_HEIGHT 40
#define DEFAULT_SIGMA 0
#define DEFAULT_DATABASE_PATH "/home/leonardo/deepu/thesisCode/src/symmetryVersion/"
#define DEFAULT_INTEREST_POINT_PATH "/home/leonardo/deepu/thesisCode/src/symmetryVersion/"
#define DEFAULT_PRINC_COMP_PATH "/home/leonardo/deepu/thesisCode/src/symmetryVersion/"
#define DEFAULT_PROJECTED_DB_PATH "/home/leonardo/deepu/thesisCode/src/symmetryVersion/"
#define DEFAULT_INTEREST_OP_PATH "/home/leonardo/deepu/thesisCode/src/symmetryVersion/runSymmetry"
#define DEFAULT_PPMSFILE "/home/leonardo/deepu/pcaproject/deeppca/src/symmetryVersion/images/ppmread"
#define DEFAULT_TESTPPMSFILE "/home/leonardo/deepu/pcaproject/deeppca/src/symmetryVersion/images/ppmtest"
#define DEFAULT_TEST_IM_FNAME "/home/leonardo/deepu/pcaproject/deeppca/ppms/test.ppm"

// Change these depending on how well you want to filter and mask
#define BAD_MATCH 0.005
#define OK_MATCH 0.0005
#define GOOD_MATCH 0.00005
#define BG_THRESHOLD 45
#define WHITE_THRESHOLD 220
#define BW_BG_THRESHOLD 10
#define STD_DEV_THRESHOLD 15

#define DIST_THRESHOLD 10

#define NUM_RMACHINES 3
              
// prototypes
matrix *processTrainingImages(char *PPMStoRead, int numIntPts, int windowWidth, int windowHeight);
matrix *processTestImage(char *testImFname,int numIntPts, int windowWidth,int windowHeight);
vector *combineForColour(vector *r, vector *g, vector *b);
int processTestImageNames(char *PPMStoTEST, char*** testArray);
int sforkexec(char *opPath, char *opExe, char *input, char *output);          // sequential
int pforkexec(int numPgms, char **pgms, char *opPath);                        // parallel
void convertNames(char **in, char **out, int n, char *ext);
void convertName(char *in, char *out, char *ext);
interestPt *getInterestPoints(char *ppmName, int numIntPts, int forTest);
void recursiveFill(FILE *fp, interestPt *array, int numPoints, int dist, PPM *image);
void justFill(FILE *fp, interestPt *array, int numPoints, int dist, PPM *image);
int checkForEmpty(interestPt *array, int numPoints);
int sufficientlyDifferent(interestPt p, interestPt *array, float dist);
int sufficientlyDifferentPoints(interestPt p, interestPt q, float dist);
int checkNeighborhood(interestPt p, PPM *image, int d);

interestPt *getSpecificInterestPoints(char *ppmName, int numIntPts);
float signatureAt(const interestPt *p, int index);
float distBetween(const interestPt *p, const interestPt *q);
float vecSignatureAt(const vector *v, int index);
float vecDistBetween(const vector *v1, const vector *v2);
interestPt *getClosestPoint(kdtree<interestPt,signatureAt,distBetween> *kdt,interestPt *point);
vector *getClosestVector(kdtree<vector,vecSignatureAt,vecDistBetween> *kdt, vector *in, int *where=NULL);
vector * getNClosestVectors(kdtree<vector,vecSignatureAt,vecDistBetween> *kdt, vector *in, int *where=NULL);
void specificFill(FILE *fp, interestPt *array, int numPoints, PPM *image);
interestPt *specificFill(FILE *fp, int numPoints, PPM *image);
int sortInterest(const void *p1, const void *p2);
int sortFloat(float *a, float *b);

int getEucDist(const vector *in, const matrix *m, float *dist);
int classifier(const matrix *testImage, const matrix *database);
int *kdtClassifier(const matrix *testImage, kdtree<vector,vecSignatureAt,vecDistBetween> *kdt);
int maskAndClassify(matrix *testImage, matrix *db, matrix *outTest, matrix *outDb, PrincipalComponents *pc); 
int *maskAndClassify(matrix *testImage, matrix *db, matrix *outTest, kdtree<vector,vecSignatureAt,vecDistBetween> *kdtOutDb , PrincipalComponents *pc);
int getMaxIndex(float *array, int arrSize);
matrix *getProjectedMatrix(matrix *in, PrincipalComponents *pc, int K=-1);
vector *getProjectedVector(vector *in, PrincipalComponents *pc, int K=-1);
void getPC(matrix *db, PrincipalComponents *pc, bool val);
int loadPC(char *path,PrincipalComponents *pc);
int savePC(char *path, PrincipalComponents *pc);
int loadMatrix(char *path, matrix *mat);
int saveMatrix(char *path, matrix *mat);

int isPowerOf2(int n);
int getNextPowerOf2(int n);
void fillTrainingImageNames(char *PPMStoRead);
void correctForPhase(char *tesImageName, int *order, int windowWidth, int windowHeight, int imageOrder);


// Globals
int globalNumImages=0;
int dbNumIntPts=DEFAULT_NUM_INTEREST_PTS;    // how many interestPts one is to get from the interest operator for the db
int testNumIntPts=DEFAULT_NUM_INTEREST_PTS;  // how many interestPts one is to get from the interest operator for the test image
interestPt **dbIntPts=NULL;                  // training set interest points
interestPt *testIntPts=NULL;                 // test image interest points  
int *closestPair=NULL;                       // every test image can keep track of best db interestPt in this array

char globalResultsPath[256]=DEFAULT_DATABASE_PATH;
int numNearestNeighbors=0;
char intOpPath[ARR_SIZE]=DEFAULT_INTEREST_OP_PATH;
char remoteMachines[NUM_RMACHINES][40]={"raven","enterprise","intrepid"}; 
int useRemote=0;
int runLinear=0;
int haveDatabase=0;  // database is big matrix of training images
int haveEverything=0;
int stayAlive=0;     // keeps things looping querying for new images to recognise
int distThreshold=DIST_THRESHOLD;
int specificSig=0;
int overKill=0;
int performMasking=0;
int userK=0;
int useCircular=0;
int useFFT=0;
int useNothing=0;
int useColour=0; // use pgms by default
int drawBoxes=0;
double sigma=DEFAULT_SIGMA;
int correctPhase=0;
double epsilon=1;
char **pgmFiles=NULL;  // 2d array which will hold the strings for the pgm fnames used for training
char **ppmFiles=NULL;  // 2d array which will hold the strings for the ppm fnames used for training

// CVPR HACK
char globalTestImFname[ARR_SIZE] = DEFAULT_TEST_IM_FNAME;

// searches kdt for the signature that best resembles that held by point
interestPt *
getClosestPoint(kdtree<interestPt,signatureAt,distBetween> *kdt, interestPt *point)
{
  int numSigs=point->getn();
  float pSigs[numSigs];
  float pointSigs[numSigs];
  // initialise the max values to those of point as when finding max's for 
  // a specific fp, its possible that the max values are greater in point.
  memcpy((char *)pointSigs, (char *)point->getSignature(), numSigs*sizeof(float));

  interestPt *p = new interestPt();
  *p=*(kdt->findClosest(point)); // force to make a copy using overloaded =
  memcpy((char *)pSigs, (char *)p->getSignature(), numSigs*sizeof(float));


  float norms[numSigs];
  float average=0.0;
  for (int i=0; i<numSigs; i++)
    {
      norms[i] = (pSigs[i] - pointSigs[i]) * (pSigs[i] - pointSigs[i]);
      average += norms[i];
    }
  p->setCompare(average); // sum of the difference squared
  //  cout << "Closest by kdt: " << *p;
  //  cout << "It's " << p->euclideanDist(*point) << "units away\n";

  return p;
}

// searches kdt for the vector that best resembles that held by the vector 'in'
vector * 
getClosestVector(kdtree<vector,vecSignatureAt,vecDistBetween> *kdt, vector *in, int *where=NULL)
{
  vector *result=new vector;
  
  *result = *(kdt->findClosest(in,where)); // forces a copy to be made  
  return result;
}

// searches kdt for the N closest vectors (where N is the global variable numNearestNeighbors) 
// that best resemble that held by the vector 'in'.
vector * 
getNClosestVectors(kdtree<vector,vecSignatureAt,vecDistBetween> *kdt, vector *in, int *where=NULL)
{
  assert(numNearestNeighbors > 0);
  
  vector **allResults=NULL;
  vector difference;
  float dist=0;
  int *whereArray=NULL;
  struct trackNode {
    int index;
    int where;
    float dist;
    float minDist;
    int numVotes;
  };
  
  whereArray = new int[numNearestNeighbors];
  trackNode trackArray[globalNumImages];

  for (int i=0; i<globalNumImages; i++) {
    trackArray[i].index=-1;
    trackArray[i].where=-1;
    trackArray[i].dist=0;
    trackArray[i].minDist=0;
    trackArray[i].numVotes=0;
  }
  
  allResults = kdt->findNClosest(in,numNearestNeighbors,whereArray); 
  
  for (int i=0; i<numNearestNeighbors; i++) {
    //difference = *in - *(allResults[i]);
    difference.assignDifference(in,allResults[i]);
    dist = difference.sqr_length();
    int imageEquiv = whereArray[i]/dbNumIntPts;
    //    cout << i << ":Matches interest point: " << whereArray[i] << " with distance " << dist << ".  Image equivalent is " << imageEquiv << "\n"; 
    trackArray[imageEquiv].numVotes++;
    trackArray[imageEquiv].dist+=dist;
    // only the closest guys index and where pt is kept
    if (trackArray[imageEquiv].index == -1) { 
      trackArray[imageEquiv].index=i;  
      trackArray[imageEquiv].where=whereArray[i];
      trackArray[imageEquiv].minDist=dist;
    }
    else if (trackArray[imageEquiv].minDist <= dist){
      trackArray[imageEquiv].index=i;  
      trackArray[imageEquiv].where=whereArray[i];      
      trackArray[imageEquiv].minDist=dist;
    }
  }

  int max=trackArray[0].numVotes;
  int maxIndex=0;
  for (int i=0; i<globalNumImages; i++) {
    //    cout << i << ":(numVotes=" << trackArray[i].numVotes << ", index=" << trackArray[i].index << ", where=" << trackArray[i].where << ", dist=" << trackArray[i].dist << ", minDist = " << trackArray[i].minDist << ")\n";
    if (trackArray[i].numVotes >= max){
      max = trackArray[i].numVotes;
      maxIndex = i;
    }
  }

  float min=trackArray[maxIndex].dist;
  int minIndex=trackArray[maxIndex].index;
  *where=trackArray[maxIndex].where;
  //  traverse and average the max values and keep track of the lowest average dist.
  for (int i=0; i<globalNumImages; i++) {
    if (trackArray[i].numVotes == max) {
      if (min >= trackArray[i].dist) {
	min = trackArray[i].dist;
	minIndex = trackArray[i].index;
	*where = trackArray[i].where;
      }
    }
  }
  
  
  vector *result = new vector(in->dim());
  result->copyVector(allResults[minIndex]);

  for (int i=0; i<numNearestNeighbors; i++) {
    delete allResults[i];
    allResults[i] = NULL;
  }
  
  delete []allResults;
  delete []whereArray;
  return result;
  
}


// fills array with interesting points which are similar to those specified in dbIntPts                                        
void
specificFill(FILE *fp, interestPt *array, int numPoints, PPM *image)
{
  interestPt *p;
  interestPt bigArray[numPoints*globalNumImages];  // contains all the best matching interest points
  int k=0;
  interestPt dummy;
  dummy.setCompare(99999);
  char buf[ARR_SIZE];
  kdtree<interestPt,signatureAt,distBetween> *kdt = new kdtree<interestPt,signatureAt,distBetween>;
  float *maxSigs=NULL;
  int first=1;

  rewind(fp);

  // since the euclidean distance is NOT what I want, I insert points that are preprocessed into
  // the kdtree, so first I have to go through and get the maximum values for each coordinate of 
  // the signature.
  int numSigs=-1;
  while (fgets(buf,ARR_SIZE,fp) != NULL) {
    dummy.read(buf);
    if (first) {
      numSigs = dummy.getn();
      maxSigs = new float[numSigs];
      memcpy((char *)maxSigs, (char *)dummy.getSignature(), numSigs*sizeof(float));      
      first=0;
    }
    for (int i=0; i<numSigs; i++) {
      if ((dummy.getSignature())[i] >= maxSigs[i]) {
	maxSigs[i] = (dummy.getSignature())[i];
      }
    }
  }    
  

  // first need to build an array of interest point which the kdtree will use
  // note this should be kept in memory.  Since kd-tree returns the closest point
  // by Euclidean distance and I want it to return values which have pre normalised
  // signatures, I create new signatures for each point divided by max coord values
  rewind(fp);
  cout << "BUILDING KD-TREE\n";
  float newSig[numSigs];
  // place everything that's in 'fp' in memory
  while (fgets(buf,ARR_SIZE,fp) != NULL)
    {
      interestPt *point = new interestPt();
      point->read(buf);
      memcpy((char *)newSig, (char *)point->getSignature(), numSigs*sizeof(float));      
      for (int i=0; i<numSigs; i++) {
	newSig[i] = newSig[i] / maxSigs[i]; 
      }
      point->setSignature(numSigs,newSig);
      kdt->insert(point);
    }
  cout << "BUILT.\n";      
  //  kdt->print();

  interestPt toCompare;
  for (int i=0; i<globalNumImages; i++) 
    for (int j=0; j<numPoints; j++)
      {
	toCompare = dbIntPts[i][j]; // makes its own copy
	memcpy((char *)newSig, (char *)dbIntPts[i][j].getSignature(), numSigs*sizeof(float));      
	for (int c=0; c<numSigs; c++) {
	  newSig[c] = newSig[c] / maxSigs[c]; 
	}
	// now change the signature such that its normalised, note it changes the actual signature
	toCompare.setSignature(numSigs,newSig);
	cout << "Comparing to image #" << i << " point #" << j << " :" << toCompare;
	p = getClosestPoint(kdt, &toCompare);
	cout << "Closest point is " << *p;
	bigArray[k] = *p;       
	
	delete p;
	k++;
	// rewind(fp);
      }


  
  qsort(bigArray, numPoints*globalNumImages, sizeof(interestPt), sortInterest);
  
  printf("SORTED\n");
  //  for (int i=0; i<numPoints*globalNumImages; i++)
  //    cout << i << ":(" << bigArray[i].getCompare() << ") " << bigArray[i];
  
  // copy the best matches to the array required
  for (int i=0; i<numPoints; i++)
    array[i] = bigArray[i];

  delete kdt;

}



// fills array with interesting points which are similar to those specified in dbIntPts
interestPt *
specificFill(FILE *fp, int numPoints, PPM *image)
{
  interestPt *p;
  interestPt *bigArray = new interestPt[numPoints*globalNumImages];  // contains all the best matching interest points
  int k=0;
  interestPt dummy;
  dummy.setCompare(99999);
  char buf[ARR_SIZE];
  kdtree<interestPt,signatureAt,distBetween> *kdt = new kdtree<interestPt,signatureAt,distBetween>;
  float *maxSigs=NULL;
  int first=1;

  rewind(fp);

  // since the euclidean distance is NOT what I want, I insert points that are preprocessed into
  // the kdtree, so first I have to go through and get the maximum values for each coordinate of 
  // the signature.
  int numSigs=-1;
  while (fgets(buf,ARR_SIZE,fp) != NULL) {
    dummy.read(buf);
    if (first) {
      numSigs = dummy.getn();
      maxSigs = new float[numSigs];
      memcpy((char *)maxSigs, (char *)dummy.getSignature(), numSigs*sizeof(float));      
      first=0;
    }
    for (int i=0; i<numSigs; i++) {
      if ((dummy.getSignature())[i] >= maxSigs[i]) {
	maxSigs[i] = (dummy.getSignature())[i];
      }
    }
  }    
  

  // first need to build an array of interest point which the kdtree will use
  // note this should be kept in memory.  Since kd-tree returns the closest point
  // by Euclidean distance and I want it to return values which have pre normalised
  // signatures, I create new signatures for each point divided by max coord values
  rewind(fp);
  cout << "BUILDING KD-TREE\n";
  float newSig[numSigs];
  // place everything that's in 'fp' in memory
  while (fgets(buf,ARR_SIZE,fp) != NULL)
    {
      interestPt *point = new interestPt();
      point->read(buf);
      memcpy((char *)newSig, (char *)point->getSignature(), numSigs*sizeof(float));      
      for (int i=0; i<numSigs; i++) {
	newSig[i] = newSig[i] / maxSigs[i]; 
      }
      point->setSignature(numSigs,newSig);
      kdt->insert(point);
    }
  cout << "BUILT.\n";      
  //  kdt->print();

  interestPt toCompare;
  for (int i=0; i<globalNumImages; i++) 
    for (int j=0; j<numPoints; j++)
      {
	toCompare = dbIntPts[i][j]; // makes its own copy
	memcpy((char *)newSig, (char *)dbIntPts[i][j].getSignature(), numSigs*sizeof(float));      
	for (int c=0; c<numSigs; c++) {
	  newSig[c] = newSig[c] / maxSigs[c]; 
	}
	// now change the signature such that its normalised, note it changes the actual signature
	toCompare.setSignature(numSigs,newSig);
	cout << "Comparing to image #" << i << " point #" << j << " :" << toCompare;
	p = getClosestPoint(kdt, &toCompare);
	cout << "Closest point is " << *p;
	bigArray[k] = *p;       
	
	delete p;
	k++;
	// rewind(fp);
      }


  
  qsort(bigArray, numPoints*globalNumImages, sizeof(interestPt), sortInterest);
  
  printf("SORTED\n");
  //  for (int i=0; i<numPoints*globalNumImages; i++)
  //    cout << i << ":(" << bigArray[i].getCompare() << ") " << bigArray[i];


  delete kdt;

  return bigArray;
}



int 
sortInterest(const void *vp1, const void *vp2)
{
  double lowerBound = -0.0000001;
  double upperBound =  0.0000001;

  // assign pointers properly

  interestPt *p1;
  interestPt *p2;

  p1 = (interestPt *)vp1;
  p2 = (interestPt *)vp2;

  if (p1->getCompare() == -1 || p2->getCompare() == -1) {
    fprintf(stderr,"Cannot compare signatures\n");
    return -2;
  }

  double difference = p1->getCompare() - p2->getCompare();
 
 

  if (difference > upperBound)
    return 1;

  if (difference < lowerBound)
    return -1;

  if ((difference > lowerBound) && (difference < upperBound))
    return 0;

  return -2;
}


int 
sortFloat(float *a, float *b)
{
  float lowerBound = -0.0000001;
  float upperBound =  0.0000001;

  float difference = *a - *b;

  if (difference > upperBound)
    return 1;

  if (difference < lowerBound)
    return -1;

  if ((difference > lowerBound) && (difference < upperBound))
    return 0;

  return -2;
}


int 
getMaxIndex(float *array, int arrSize)
{
  float max=-99999;
  int index=-1;
  
  for (int i=0; i<arrSize; i++)
    {
      if (array[i] > max)
	{
	  max = array[i];
	  index = i;
	}
    }
  return index;
}

// return closest row in "m" to "in" using euclidean distance
int 
getEucDist(const vector *in, const matrix *m, float *dist)
{
  float minDistance=99999999;
  int index=-1;
  float distance;
  vector difference;

  for (int i=0; i < m->dim1(); i++)
    {
      // cout << "Vector:" << in << ", Matrix" << m; 
      //      difference = (*m)[i] - *in;
      difference.assignDifference(&((*m)[i]), in);
      distance = difference.sqr_length();
      // cout << i << ":" << distance << "\n";
      if (distance < minDistance)
	{
	  minDistance = distance;
	  *dist = distance;
	  index = i;
	}
    }

  //  cout << "Min Distance = " << minDistance << "\n";

  return index;
}


// CONST FUCK UPS IN KDT TREE!!!!!!!  I think
int *kdtClassifier(const matrix *testImage, kdtree<vector,vecSignatureAt,vecDistBetween> *kdt)
{
  float dist=-1.0;
  vector difference;
  int numNodes = kdt->getNumNodes();
  float buckets[numNodes];
  int numImages = globalNumImages;
  float values[numImages];  

  int *bestIndeces;
  bestIndeces = new int[numImages];


  vector *closest;

  // Initialise the buckets to 0
  for (int i=0; i<numNodes; i++)
    buckets[i] = 0.0;

  for (int i=0; i<numImages; i++) {
    values[i] = 0.0;
    bestIndeces[i] = -1;
  }

  int numPoints = testImage->dim1(); //number of row == number of matrices around sub-points

  // For every test Image (ie matrix around interest point) look for the closest row in 
  // the database and place a measurement metric at that index     
  int where=-1;
  vector testVector;
  for (int i=0; i<numPoints; i++)
    {
      testVector = (*testImage)[i];
      if (numNearestNeighbors)
	closest = getNClosestVectors(kdt,&testVector,&where);
      else
	closest = getClosestVector(kdt,&testVector,&where);
      closestPair[i] = where; // global array to keep track of which test interestPt matches which global one.
      //difference = testVector - *closest;
      difference.assignDifference(&testVector,closest);
      if (closest==NULL) 
	cout << "NULL Delete!!" << endl;
      delete closest;
      dist = difference.sqr_length();
      buckets[where] += 1/(dist + epsilon); 
      cout << i << ":Matches interest point: " << where << " with distance " << dist << ", weight "  << buckets[where] << 
	".  Image equivalent is " << where/dbNumIntPts << "\n"; 
    }

 
  for (int i=0; i<numNodes; i++)
    {
      // printf("BUCKET VALUE %d=%d\n",i,buckets[i]);
      if (buckets[i] != 0)
	// values[(int)(i/numPoints)] += buckets[i];
	values[(int)(i/dbNumIntPts)] += buckets[i];
    }


  
  // CVPR HACK
  
  float triples[14];
  char *names[14] = {"bear","bread","bDino","burger","cstick","clown","frog","giraffe","gDino","monkey","oDino","pepper","pot","tiger"};
  for (int i=0; i<14; i++)
    triples[i]=0;

  printf("\n");
  float sumThree=0;
  int sumCnt=0;
  for (int i=0; i<numImages; i++) {
    printf("%d: %f votes for %s\n",i,values[i],ppmFiles[i]);
    sumThree += values[i];
    if (((i+1) % 3 ) == 0) {
      printf("Sum of three = %f\n",sumThree);
      triples[sumCnt++]=sumThree;
      sumThree=0;      
    }
  }
      

  char cvprhack[256];
  strcpy(cvprhack,globalResultsPath);
  strcat(cvprhack,"/cvprhack");
  cout << "Attempting to open file " << cvprhack << "\n";

  FILE *fp;
  fp = fopen(cvprhack,"a");
  if (fp == NULL)
    {
      fprintf(stderr,"Problems openining %s\n",cvprhack);
      return NULL;
    }

  int max1 = getMaxIndex(triples,14);  
  fprintf(fp,"%d:%70s\t%f for %s\t",testNumIntPts,globalTestImFname,triples[max1],names[max1]);
  triples[max1]=-1;
  int max2 = getMaxIndex(triples,14);  
  fprintf(fp,"%f for %s\n",triples[max2],names[max2]);
  fclose(fp);

 
 

  
  // NORMAL VERSION (no triple summation)
  for (int i=0; i<numImages; i++) {
    printf("%d: %f votes for %s\n",i,values[i],ppmFiles[i]);
  }
 
  printf("\n");
  

  for (int i=0; i<numImages; i++) {
    int index = getMaxIndex(values,numImages);
    bestIndeces[i] = index;
    values[index] = -1;
  }

  //  for (int i=0; i<numImages; i++)
  //    printf("%s\n",ppmFiles[bestIndeces[i]]);

  printf("Best Match is %s\n",ppmFiles[bestIndeces[0]]);
  
  return bestIndeces;  

}
  
int 
classifier(const matrix *testImage, const matrix *database)
{

  float dist=-1.0;
  int index;
  //  vector difference;
  float buckets[database->dim1()];
  // int numImages = (int)(database->dim1()/testImage->dim1());
  int numImages = globalNumImages;
  float values[numImages];

  // Initialise the buckets to 0
  for (int i=0; i<database->dim1(); i++)
    buckets[i] = 0.0;

  for (int i=0; i<numImages; i++)
    values[i] = 0.0;

  int numPoints = testImage->dim1(); //number of row == number of matrices around sub-points

  // For every test Image (ie matrix around interest point) look for the closest row in 
  // the database and place a measurement metric at that index     
  for (int i=0; i<testImage->dim1(); i++)
    {
      vector testVector = (*testImage)[i];
      int where = getEucDist(&testVector,database,&dist);
      closestPair[i] = where; // global array to keep track of which test interestPt matches which global one.
      cout << i << ":Matches interest point: " << where << " with distance " << dist << ".  Image equivalent is " << where/numPoints << "\n";
      buckets[where] += 1/dist; 
    }

 
  for (int i=0; i<database->dim1(); i++)
    {
      // printf("BUCKET VALUE %d=%d\n",i,buckets[i]);
      if (buckets[i] != 0)
	values[(int)(i/numPoints)] += buckets[i];
    }

    
  for (int i=0; i<numImages; i++)
    printf("VALUE[%d]=%f\n",i,values[i]);

  index = getMaxIndex(values,numImages);

  printf("Best Match is with image %d\n",index);

  return index;

}

// testImage is the matrix consisting of rows which form the individual subwindows
// around the testImage interestPts.  The idea here is to find the vectors in testImage
// (not projected yet) which when projected lead to points that are close to outDb points.
// For these ambiguous points, remove the background by overlaying the training images and
// removing the background.
int 
maskAndClassify(matrix *testImage, matrix *db, matrix *outTest, matrix *outDb, PrincipalComponents *pc)
{
  float dist=-1.0;
  struct bucket
  {
    float value;
    int index;
  };

  bucket bucketArray[outTest->dim1()];
  // int numImages = (int)(db->dim1()/testImage->dim1());
  int numImages = globalNumImages;
  //  int numIntPts = dbNumIntPts;  // outTest->dim1()
  int numIntPts = outTest->dim1();  // corresponds to the number of test image interest points
  float values[numImages];
  

  // Initialise the buckets to 0
  for (int i=0; i<outTest->dim1(); i++) {
    bucketArray[i].value = 10.0;
    bucketArray[i].index = -1;
  }

  // For every test Image (ie matrix around interest point) look for the closest row in 
  // the database and place a measurement metric at that index     
  for (int i=0; i<outTest->dim1(); i++)
    {
      vector testVector = (*outTest)[i];
      int where = getEucDist(&testVector,outDb,&dist);
      closestPair[i] = where; // global array to keep track of which test interestPt matches which global one.
      // cout << "Matches image number: " << where/numPoints << "\n";
      bucketArray[i].value = dist;
      bucketArray[i].index = where;
    }

  for (int i=0; i<outTest->dim1(); i++) {
    printf("Test Image %d matches Training image %d with value %f\n",i,bucketArray[i].index, bucketArray[i].value);
  }
  printf("\n");
  
  // Now for every test image that returned a best match which was shitty, shitty being defined as 
  // having a value which is >= 0.1, remove the background by overlaying the training images (keep 
  // result which gives the smallest distance.  This is where stuff gets expensive.
  vector *projectedMask;
  vector maskToProject;
  for (int i=0; i<outTest->dim1(); i++) 
    {
      /*  
	  Now there's different ways I can go about this...make something thats a good match an even better match or
	  as I'm doing currently, which is to make a bad match into a better one.  Talk to Greg.
      */
      if (bucketArray[i].value >= BAD_MATCH) 
	{  
	  maskToProject = (*testImage)[i]; // creates a copy
	  cout << "Trying to improve test image " << i << " by removing bg\n";
	  for (int j=0; j<db->dim1(); j++) 
	    {
	      if (useColour)
		{
		  for (int k=0; k<db->dim2(); k+=3) 
		    { // account for rgb
		      if (((*db)(j,k) <= BG_THRESHOLD) && ((*db)(j,k+1) <= BG_THRESHOLD) && ((*db)(j,k+2) <= BG_THRESHOLD)) 
			{
			  // paint the potential bg black in the image to be projected
			  maskToProject[k] = (*db)(j,k);
			  maskToProject[k+1] = (*db)(j,k+1);
			  maskToProject[k+2] = (*db)(j,k+2);
			}
		    }
		}
	      else
		{
		  for (int k=0; k<db->dim2(); k++) 
		    { // account for rgb
		      if (((*db)(j,k) <= BW_BG_THRESHOLD))
			{
			  // paint the potential bg black in the image to be projected
			  maskToProject[k] = (*db)(j,k);
			}
		    }
		}
	      
	      // project this vector onto the eigenspace and get distance to closest eigen point
	      projectedMask = getProjectedVector(&maskToProject,pc,userK);
	      // get the distance in eigenspace and if it is less than that in bucketArray, keep it and
	      // update the outTest matrix.
	      int where = getEucDist(projectedMask,outDb,&dist);
	      // reset the maskToProject
	      maskToProject = (*testImage)[i];
	      if (dist < bucketArray[i].value) 
		{
		  cout << "Image " << i << " is better matched with " << where << " in " << j << "th iteration with " << dist << "\n"; 
		  closestPair[i] = where; // global array to keep track of which test interestPt matches which global one.
		  // Then update the outTest matrix;
		  (*outTest)[i] = *projectedMask;
		  // free some memory
		  delete projectedMask;
		  bucketArray[i].index = where;
		  bucketArray[i].value = dist;
		  if (dist < GOOD_MATCH) 
		    {
		      // potential speed up?! 
		      cout << "Found a good match, breaking out\n";
		      break;
		    }	  
		}
	      else 
		{
		  delete projectedMask;
		}
	    }
	}
    }
  cout << "\n";
  for (int i=0; i<outTest->dim1(); i++) {
    printf("Test Image %d matches Training image %d with value %f\n",i,bucketArray[i].index, bucketArray[i].value);
  }

  // initialise
  for (int i=0; i<numImages; i++)
    values[i]=0.0;
  // sum
  for (int i=0; i<outTest->dim1(); i++)    
    values[(int)(bucketArray[i].index/numIntPts)] += 1/bucketArray[i].value;
  for (int i=0; i<numImages; i++)
    printf("VALUE[%d]=%f\n",i,values[i]);
  int index = getMaxIndex(values,numImages);

  printf("Best Match is with image %d\n",index);

  return 1;
}



// testImage is the matrix consisting of rows which form the individual subwindows
// around the testImage interestPts.  The idea here is to find the vectors in testImage
// (not projected yet) which when projected lead to points that are close to outDb points.
// For these ambiguous points, remove the background by overlaying the training images and
// removing the background.
int *
maskAndClassify(matrix *testImage, matrix *db, matrix *outTest, kdtree<vector,vecSignatureAt,vecDistBetween> *kdtOutDb , PrincipalComponents *pc)
{
  float dist=-1.0;
  struct bucket
  {
    float value;
    int index;
  };

  bucket bucketArray[outTest->dim1()];
  vector *closest;
  vector difference;

  // Initialise the buckets
  for (int i=0; i<outTest->dim1(); i++) {
    bucketArray[i].value = 10.0;
    bucketArray[i].index = -1;
  }


  // For every test Image (ie matrix around interest point) look for the closest row in 
  // the database and place a measurement metric at that index     
  int where=-1;
  vector testVector;
  for (int i=0; i<outTest->dim1(); i++)
    {
      testVector = (*outTest)[i];
      if (numNearestNeighbors)
	closest = getNClosestVectors(kdtOutDb,&testVector,&where);
      else
	closest = getClosestVector(kdtOutDb,&testVector,&where);
      //      closest = getClosestVector(kdtOutDb,&testVector,&where);
      //      difference = *closest - testVector;
      closestPair[i] = where;
      difference.assignDifference(closest,&testVector);
      delete closest;
      dist = difference.sqr_length();
      // cout << "Matches image number: " << where/numPoints << "\n";
      bucketArray[i].value = dist;
      bucketArray[i].index = where;
    }

  for (int i=0; i<outTest->dim1(); i++) {
    printf("Test Image %d matches Training image %d with value %f\n",i,bucketArray[i].index, bucketArray[i].value);
  }
  printf("\n");

  // Now for every test image that returned a best match which was shitty, shitty being defined as 
  // having a value which is >= BAD_MATCH, remove the background by overlaying the training images (keep 
  // result which gives the smallest distance.  This is where stuff gets expensive.
  vector *projectedMask;
  vector maskToProject;
  where = -1;
  for (int i=0; i<outTest->dim1(); i++) {
    /*  
	Now there's different ways I can go about this...make something thats a good match an even better match or
	as I'm doing currently, which is to make a bad match into a better one.  Talk to Greg.
    */
    if (bucketArray[i].value >= BAD_MATCH) {  
      maskToProject = (*testImage)[i]; // creates a copy
      cout << "Trying to improve test image " << i << " by removing bg\n";
      for (int j=0; j<db->dim1(); j++) {
	if (useColour) {
	  for (int k=0; k<db->dim2(); k+=3) { // account for rgb
	    if (((*db)(j,k) <= BG_THRESHOLD) && ((*db)(j,k+1) <= BG_THRESHOLD) && ((*db)(j,k+2) <= BG_THRESHOLD)) {
	      // paint the potential bg black in the image to be projected
	      maskToProject[k] = (*db)(j,k);
	      maskToProject[k+1] = (*db)(j,k+1);
	      maskToProject[k+2] = (*db)(j,k+2);
	    }
	  }
	}
	else {
	  for (int k=0; k<db->dim2(); k++) { 
	    if (((*db)(j,k) <= BW_BG_THRESHOLD)) {
	      // paint the potential bg black in the image to be projected
	      maskToProject[k] = (*db)(j,k);
	    }
	  }
	}
	
	
	// project this vector onto the eigenspace and get distance to closest eigen point
	projectedMask = getProjectedVector(&maskToProject,pc,userK);
	// get the distance in eigenspace and if it is less than that in bucketArray, keep it and
	// update the outTest matrix.
	if (numNearestNeighbors)
	  closest = getNClosestVectors(kdtOutDb,projectedMask,&where);
	else
	  closest = getClosestVector(kdtOutDb,projectedMask,&where);
	//	      closest = getClosestVector(kdtOutDb,projectedMask,&where);
	// difference = *closest - *projectedMask;
	difference.assignDifference(closest,projectedMask);
	dist = difference.sqr_length();
	// reset the maskToProject
	maskToProject = (*testImage)[i];
	if (dist < bucketArray[i].value) {
	  cout << "Image " << i << " is better matched with " << where << " in " << j << "th iteration with " << dist << "\n"; 
	  cout << "This corresponds to interest point " << dbIntPts[where/dbNumIntPts][where % dbNumIntPts] << endl;
	  closestPair[i] = where; // global array to keep track of which test interestPt matches which global one.
	  // Then update the outTest matrix;
	  (*outTest)[i] = *projectedMask;
	  // free some memory
	  delete projectedMask;
	  bucketArray[i].index = where;
	  bucketArray[i].value = dist;
	  /*
	  if (dist < GOOD_MATCH) {
	    // potential speed up?! 
	    cout << "Found a good match, breaking out\n";
	    break;
	  }	
	  */
	}
	else {
	  delete projectedMask;
	}
      }
    }
  }

  int *bestIndeces = kdtClassifier(outTest,kdtOutDb);
  return bestIndeces;
  
}

matrix 
*getProjectedMatrix(matrix *in, PrincipalComponents *pc, int K)
{
  bool result;
  int numDims=K;
  
  if (K==0) {
    numDims = pc->GetK();
    userK = numDims;
  }

  cout << "Db:" << in->dim1() << " x " << in->dim2() << "\n";
  matrix *out=new matrix(in->dim1(),numDims);
  cout << "outDb:" << out->dim1() << " x " << out->dim2() << "\n";
  result = pc->Project(*in,*out);
  return out;
}


vector
*getProjectedVector(vector *in, PrincipalComponents *pc, int K)
{
  bool result;
  int numDims = K;
  
  if (K==0) {
    numDims = pc->GetK();
    userK = numDims;
  }

  vector *out = new vector(numDims);
  result = pc->Project(*in, *out);
  return out;
}


void
getPC(matrix *db, PrincipalComponents *pc, bool val)
{
  PCA pcaObject;

  //  pcaObject.SetVerbosity(4);
  //  (*pc).SetVerbosity(4);
  // the boolean value normalises the row vector to 1 (ie images) if true.
  // This has the effect of accounting for intensity
  (*pc) = pcaObject.GetSVDPrincipalComponents(*db,val);
  if (userK == 0)
    userK = pc->GetK(); // set the global value of userK
}


int
loadPC(char *path, PrincipalComponents *pc)
{
  PCA pcaObject;
  fstream principal;


  printf("Loading principal components from %s\n",path);
  principal.open(path,ios::in | ios::nocreate);
  if (!principal) 
    {
      fprintf(stderr,"Problems opening %s for reading\n",path);
      return 0;
    }
  principal.seekg(0);
  pc->read(principal);

  if (userK == 0)
    userK = pc->GetK(); // set the global value of userK

  principal.close();
  return 1;
}

// assumes that space has already been allocated for mat
int
loadMatrix(char *path, matrix *mat)
{
  fstream matrixstream;


  printf("Loading projected database from %s\n",path);
  matrixstream.open(path,ios::in | ios::nocreate);
  if (!matrixstream) 
    {
      fprintf(stderr,"Problems opening %s for reading\n",path);
      return 0;
    }
  matrixstream.seekg(0);
  mat->read(matrixstream);

  matrixstream.close();
  return 1;
}

int
savePC(char *path, PrincipalComponents *pc)
{

  fstream principal;
  
  principal.open(path,ios::out | ios::trunc);
  if (!principal) 
    {
      fprintf(stderr,"Problems opening %s for writing\n",path);
      return 0;
    }
  principal.seekg(0);

  principal << *pc;

  principal.close();
  return 1;
}


int
saveMatrix(char *path, matrix *mat)
{

  fstream matrixstream;
  
  matrixstream.open(path,ios::out | ios::trunc);
  if (!matrixstream) 
    {
      fprintf(stderr,"Problems opening %s for writing\n",path);
      return 0;
    }
  matrixstream.seekg(0);

  mat->print(matrixstream);

  matrixstream.close();
  return 1;
}



vector *
combineForColour(vector *r, vector *g, vector *b)
{
  int count=0;

  assert(r->dim() == g->dim() && g->dim() == b->dim());
  vector *combine = new vector((r->dim())*3);
  for (int i=0; i<combine->dim(); i+=3) {
    (*combine)[i] = (*r)[count];
    (*combine)[i+1] = (*g)[count];
    (*combine)[i+2] = (*b)[count];
    count++;
  }
  return combine;
}

/*
  Given a file containing the test Image to read, the number of interest points 
  required, the following loads the ppm, runs the interest operator on in, creates
  a matrix around the interest points and returns it.
*/
matrix *
processTestImage(char *testImFname,int numIntPts, int width,int height)
{
  char *delimiter;
  char pgmName[ARR_SIZE];
  char datName[ARR_SIZE];
  char opName[ARR_SIZE];
  PPM testIm;
  int retVal=0;
  matrix *result;
  int newWidth=0;
  int newHeight=0;


  if (useFFT) {
    newWidth = getNextPowerOf2(width);
    newHeight = getNextPowerOf2(height);
  }
  else {
    newWidth = width;
    newHeight = height;
  }
  

  int scale=1;
  if (overKill)
    scale = globalNumImages;

  if (!useColour) // black and white
    result = new matrix(scale*numIntPts,newWidth*newHeight);
  else            // dealing with ppms
    result = new matrix(scale*numIntPts,newWidth*newHeight*3);

  printf("Loading %s\n",testImFname);
  if (!testIm.load(testImFname))
    {
      fprintf(stderr,"Problems loading %s\n",testImFname);
      return NULL;
    }

  matrix *r;
  matrix *g;
  matrix *b;

  r = testIm.makeMatrix(testIm,'r');
  g = testIm.makeMatrix(testIm,'g');
  b = testIm.makeMatrix(testIm,'b');
  

  // Create a matrix using the three colours combined meaningfully (meaningful according to ppmtopgm)
  matrix forPGM = (*r)*0.299 + (*g)*0.587 + (*b)*0.114;

  delete r;
  delete g;
  delete b;

  PGM pgmImage(forPGM);

  convertName(testImFname,pgmName,".pgm");
  convertName(testImFname,datName,".dat");

  pgmImage.save(pgmName);


  // Assuming that the operator path is the full operator path 
  // together with executable filename, separate the name from 
  // path.  Note that if the path has no "/" in it, it assumes
  // that the binary is the same as the path
  delimiter = strrchr(intOpPath,'/');
  if (delimiter != NULL)
    strcpy(opName,delimiter+1);
  else
    strcpy(opName,intOpPath);

  
  retVal = sforkexec(intOpPath,opName,pgmName,datName);

  if (!specificSig) 
    testIntPts = getInterestPoints(testImFname,numIntPts,1);
  else
    testIntPts = getSpecificInterestPoints(testImFname,numIntPts);

  // allocate space for pair matching
  closestPair = new int[numIntPts];  // this is a global allocation

  printf("Interest Points for %s\n",pgmName);
  for (int i=0; i<scale*numIntPts; i++)
    cout << i << ":" << testIntPts[i];

  if (useCircular) 
    cout << "Creating matrix of sub windows around test interest points [circular]\n";
  else if (useFFT)
    cout << "Creating matrix of sub windows around test interest points [fft]\n";
  else if (useNothing)
    cout << "Creating matrix of sub windows around test interest points\n";

  int row=0;
  vector v;

  PGM pgmCopy;
  PPM ppmCopy;
  if (drawBoxes) {
    ppmCopy = testIm;
    pgmCopy = pgmImage;
  }
      
  for (int j=0; j<scale*numIntPts; j++)
    {
      // Colour fft
      if (useColour && useFFT) { 
	// First crop the PPM around the interest point
	matrix *redVals   = testIm.makeMatrix(testIm,*(testIntPts+j),width,height,'r');
	matrix *greenVals = testIm.makeMatrix(testIm,*(testIntPts+j),width,height,'g');
	matrix *blueVals  = testIm.makeMatrix(testIm,*(testIntPts+j),width,height,'b');
	if (drawBoxes) ppmCopy.drawWindow(ppmCopy,*(testIntPts+j),width,height);	
	// next make polar matrices for each color and multiply by a Gaussian (Hamming window)
	matrix polarRed   = redVals->makePolarMatrix(*redVals,width,height);
	matrix polarGreen = greenVals->makePolarMatrix(*greenVals,width,height);
	matrix polarBlue  = blueVals->makePolarMatrix(*blueVals,width,height);
	delete redVals;
	delete greenVals;
	delete blueVals;
	matrix redGaussian(polarRed.dim1(),polarRed.dim2());
	matrix greenGaussian(polarGreen.dim1(),polarGreen.dim2());
	matrix blueGaussian(polarBlue.dim1(),polarBlue.dim2());
	redGaussian = polarRed.makeGaussianMatrix(polarRed,sigma);
	greenGaussian = polarGreen.makeGaussianMatrix(polarGreen,sigma);
	blueGaussian = polarBlue.makeGaussianMatrix(polarBlue,sigma);
	// now scale the matrices to make them powers of 2 dimesions
	matrix redGaussianScaled(newHeight,newWidth);
	matrix greenGaussianScaled(newHeight,newWidth);
	matrix blueGaussianScaled(newHeight,newWidth);
	redGaussianScaled = redGaussian.makeBiggerMatrix(redGaussian,newHeight,newWidth);
	greenGaussianScaled = greenGaussian.makeBiggerMatrix(greenGaussian,newHeight,newWidth);
	blueGaussianScaled = blueGaussian.makeBiggerMatrix(blueGaussian,newHeight,newWidth);
	// fft each polar gaussian color matrix
	fft redFft(redGaussianScaled);
	fft greenFft(greenGaussianScaled);
	fft blueFft(blueGaussianScaled);
	// compute amplitudes for each colour 
	redFft.computeLogAmplitude();
	greenFft.computeLogAmplitude();
	blueFft.computeLogAmplitude();
	matrix redAmp = *(redFft.getAmplitude());
	matrix greenAmp = *(greenFft.getAmplitude());
	matrix blueAmp = *(blueFft.getAmplitude());
	// combine info into a big vector
	vector redV = redAmp.makeVector(redAmp);
	vector greenV = greenAmp.makeVector(greenAmp);
	vector blueV = blueAmp.makeVector(blueAmp);
	vector *temp = combineForColour(&redV,&greenV,&blueV);
	v = *temp;
	delete temp;
      } 
      // Black and White FFT
      else if (!useColour && useFFT) { 
	matrix *subMatrix = pgmImage.makeMatrix(pgmImage,*(testIntPts+j),width,height);
	if (drawBoxes) pgmCopy.drawWindow(pgmCopy,*(testIntPts+j),width,height);	
	matrix polar = subMatrix->makePolarMatrix(*subMatrix,width,height);
	delete subMatrix;
	matrix polarGaussian(polar.dim1(),polar.dim2());
	polarGaussian = polar.makeGaussianMatrix(polar,sigma);
	matrix polarGaussianScaled(newHeight,newWidth);
	polarGaussianScaled = polarGaussian.makeBiggerMatrix(polarGaussian,newHeight,newWidth);
	fft fftobj(polarGaussianScaled);
	fftobj.computeLogAmplitude();
	matrix *amp = fftobj.getAmplitude();
	v = (*amp).makeVector(*amp);
      } 
      // Colour circular
      else if (useColour && useCircular) {
	matrix *subMatrix = testIm.makeGaussianMatrix(testIm,*(testIntPts+j),newWidth,newHeight,sigma);
	if (drawBoxes) ppmCopy.drawWindow(ppmCopy,*(testIntPts+j),newWidth,newHeight);	
	matrix circle(subMatrix->dim1(), subMatrix->dim2());
	circle = (*subMatrix).makeCircularMatrix(*subMatrix); 
	delete subMatrix;
	v = circle.makeVector(circle);
      }
      // Black and White circular
      else if (!useColour && useCircular) {
	matrix *subMatrix = pgmImage.makeGaussianMatrix(pgmImage,*(testIntPts+j),newWidth,newHeight,sigma);
	if (drawBoxes) pgmCopy.drawWindow(pgmCopy,*(testIntPts+j),width,height);	
	matrix circle(subMatrix->dim1(), subMatrix->dim2());
	circle = (*subMatrix).makeCircularMatrix(*subMatrix); 
	delete subMatrix;
	v = circle.makeVector(circle);
      }      
      // Colour nothing
      else if (useColour && useNothing) {
	matrix *subMatrix = testIm.makeGaussianMatrix(testIm,*(testIntPts+j),newWidth,newHeight,sigma);
	if (drawBoxes) ppmCopy.drawWindow(ppmCopy,*(testIntPts+j),newWidth,newHeight);	
	v = (*subMatrix).makeVector(*subMatrix);
	delete subMatrix;
      }
      // Black and White nothing
      else if (!useColour && useNothing) {
	matrix *subMatrix = pgmImage.makeGaussianMatrix(pgmImage,*(testIntPts+j),newWidth,newHeight,sigma);
	if (drawBoxes) pgmCopy.drawWindow(pgmCopy,*(testIntPts+j),width,height);	
	v = (*subMatrix).makeVector(*subMatrix);
	delete subMatrix;
      }
      (*result)[row] = v;
      row++;
    }

  if (drawBoxes && !useColour) {
    strcat(pgmName,".box");
    pgmCopy.save(pgmName);
  }

  if (drawBoxes && useColour) {
    strcat(testImFname,".box");
    ppmCopy.save(testImFname);
  }

  // now have a matrix where each row vector consists of an image which was 
  // obtained around an interest point.

  return result;
}

// Given an empty testArray, allocates space for it and fills
// it with the correct filenames to load, returns the size of 
// the allocated array.  *** is ugly but neat in its ugliness
int processTestImageNames(char *PPMStoTEST, char ***testArray)
{
  FILE *fp;
  char *delimiter;
  char buffer[ARR_SIZE];
  int numTestImages=0;

  fp = fopen(PPMStoTEST,"r");
  if (fp == NULL)
    {
      fprintf(stderr,"Problems opening %s\n",PPMStoTEST);
      return -1;
    }

  // Need to remove the filename itself
  delimiter = strrchr(PPMStoTEST,'/');
  *delimiter = '\0';


  int c=0;
  while (fgets(buffer,ARR_SIZE,fp) != NULL)
    {
      if (buffer[0] == '.')
	break;
      // printf("Parsing :%s",buffer);
      if (buffer[0] == '!') 
	{
	  sscanf(buffer,"! %d",&numTestImages);
	  *testArray = new char*[numTestImages];
	  for (int i=0; i<numTestImages; i++)
	    {
	      (*testArray)[i]=new char[ARR_SIZE];
	      strcpy((*testArray)[i],"");
	    }
	}
      else if (buffer[0] != '!' && buffer[0] != '#' && c < numTestImages)
	{
	  strcat(PPMStoTEST,"/");
	  strcat(PPMStoTEST,buffer);
	  delimiter = strrchr(PPMStoTEST,'\n');
	  *delimiter = '\0';
	  strcpy((*testArray)[c],PPMStoTEST);
	  //	  printf("%d: %s\n",c,(*testArray)[c]);
	  c++;

	  // setup buffer for next ppm to be read
	  delimiter = strrchr(PPMStoTEST,'/');
	  *delimiter='\0';
	}
    }

  return numTestImages;
}

int 
isPowerOf2(int n)
{
  int copy = n;
  int copy2 = n;
  int andVal = 0;
  int numOnBits=0;

  while (copy != 0) {
    andVal = copy2 & copy;
    if (andVal)
      numOnBits++;
    copy >>= 1;
  }
  
  if (numOnBits == 1)
    return 1;

  return 0;
}


int
getNextPowerOf2(int n)
{
  int increment;
  if ((n % 2) == 0)
    increment = 2;
  else
    increment = 1;

  while(!isPowerOf2(n))
    n += increment;

  return n;
}


/*
  By the time this is called one should have ppmFiles and pgmFiles fully
  allocated and filled.  So given the testImage file name and an array containing
  the sorted indeces of the best matches, this function proceeds to load the 
  relevant sub-windows and compute the phase for a given image match.

  ABANDONED!!!!!!!!!!!!

*/
void 
correctForPhase(char *testImageName, int *order, int width, int height, int imageOrder)
{

  PGM trainingPgm;   // An array of pointers to pgms which are loaded
  PGM testPgm;         // Holds the testImage PGM
  char testPgmName[ARR_SIZE];

  // convert the testImageName from ppm to pgm
  cout << testImageName << "\n";
  convertName(testImageName,testPgmName,".pgm");
  // load the testImage pgm now
  printf("Loading test image: %s for phase fix\n",testPgmName);
  if (!(testPgm.load(testPgmName)))
    fprintf(stderr,"problems loading file");


  // load the db best pgm (i.e. recognised one)
  printf("Loading db image: %s for phase fix\n",pgmFiles[order[imageOrder]]);
  if (!(trainingPgm.load(pgmFiles[order[imageOrder]])))
    fprintf(stderr,"problems loading file");

  
  // create a 2d array of doubles to hold phases
  double *phases;
  phases = new double[testNumIntPts];
  for (int i=0; i<testNumIntPts; i++) 
    phases[i] = -1;
  
  for (int i=0; i<testNumIntPts; i++) {
    cout << i << " matches " << closestPair[i] << "\n";
  }
    

  double avePhase=0;
  for (int j=0; j<testNumIntPts; j++) {
    // crop the window in the test image and make a polar matrix
    matrix *testWin     = testPgm.makeMatrix(testPgm,testIntPts[j],width,height);
    matrix polarTestWin = testWin->makePolarMatrix(*testWin,width,height);  
    int closestMatch = closestPair[j];
    int row = closestMatch / dbNumIntPts;
    int col = closestMatch % dbNumIntPts;
    cout << "r=" << row << ",c=" << col << "\n";
    // crop the window in the training image and make a polar matrix
    matrix *dbWin       = trainingPgm.makeMatrix(trainingPgm,dbIntPts[row][col],width,height);
    matrix polarDbWin   = dbWin->makePolarMatrix(*dbWin,width,height);
    double angle = fft::recoverPhase(polarTestWin,polarDbWin);  // uses a least squares match 
    cout << "(" << j << ")" << "Angle = " << angle*180/PI << "\n";
    cout << "Comparing: " << testIntPts[j] << " to " << dbIntPts[row][col] << "\n";
    phases[j] = angle;
    avePhase += angle;
  }

  avePhase /= testNumIntPts;
  cout << "Average phase = " << avePhase*180/PI << "\n";

  // load all the subwindows for the db images

  // compute the phase shift for the test image by comparing to recognised object and 
  // second best recognised object

  // Find the sub-windows in the test image which do not have this phase.

  // Create a new matrix to project onto the eigenspace which does NOT contain subwindows which 
  // do not exhibit such a phase

  // Reproject it onto the eigenspace and reclassify.

}


// Given the path name of the files with which the training db was constructed
// this proceeds to fill in the global arrays for the names corresponding to the
// ppms which were used to create the db.
void fillTrainingImageNames(char *PPMStoRead)
{ 
  FILE *fpdb;  // contains the ppms to be read
  char *delimiter;
  char buffer[ARR_SIZE];
  int numImages=0;
  char pgmName[ARR_SIZE];

  fpdb = fopen(PPMStoRead,"r");
  if (fpdb == NULL)
    {
      fprintf(stderr,"Problems opening %s\n",PPMStoRead);
      return;
    }

  // Need to remove the filename itself
  delimiter = strrchr(PPMStoRead,'/');
  *delimiter = '\0';
  
  int c=0; 
  // preprocess the buffer to get the number of images, number of pixels in an image 
  // and convert all ppms to pgms
  while (fgets(buffer,ARR_SIZE,fpdb) != NULL)
    {
      if (buffer[0] == '.')
	break;
      if (buffer[0] == '!') 
	{
	  sscanf(buffer,"! %d",&numImages);
	  // now global pgmFiles has numImages entries with 256 chars for each entry
	  pgmFiles = new char*[numImages];
	  for (int i=0; i<numImages; i++)
	    {
	      pgmFiles[i]=new char[ARR_SIZE];
	      strcpy(pgmFiles[i],"");
	    }
	  // now global ppmFiles has numImages entries with 256 chars for each entry
	  ppmFiles = new char*[numImages];
	  for (int i=0; i<numImages; i++)
	    {
	      ppmFiles[i]=new char[ARR_SIZE];
	      strcpy(ppmFiles[i],"");
	    }
	}
      else if (buffer[0] != '!' && buffer[0] != '#')
	{
	  strcat(PPMStoRead,"/");
	  strcat(PPMStoRead,buffer);
	  delimiter = strrchr(PPMStoRead,'\n');
	  *delimiter = '\0';
	  strcpy(ppmFiles[c],PPMStoRead);


	  // fix the extension on the pgm filename
	  strcpy(pgmName,PPMStoRead);
	  delimiter = strrchr(pgmName,'.');
	  *delimiter = '\0';
	  strcat(pgmName,".pgm");
	  strcpy(pgmFiles[c],pgmName);

	  c++;

	  // setup buffer for next ppm to be read
	  delimiter = strrchr(PPMStoRead,'/');
	  *delimiter='\0';
	}
    }
}



/*
  Given a file containing the ppms to read, the number of interest points required,
  the following opens the file PPMStoRead, for every ppm within that file, it forks 
  the attention operator with two command line args.  The ppm read from PPMStoRead and
  the target file (.dat).  For each target (.dat) file, once its ready, numIntPts interest 
  points are extracted and numIntPts subwindows are cropped from that image and places as 
  rows of the database matrix which is returned.
*/
matrix *
processTrainingImages(char *PPMStoRead, int numIntPts, int width, int height)
{
  FILE *fpdb;  // contains the ppms to be read
  char *delimiter;
  char buffer[ARR_SIZE];
  char pgmName[ARR_SIZE];
  int numImages=0;
  int numPixels=0;
  int gotNumPixels=0;
  PPM **ppms=NULL;       // An array of pointers to ppms which are loaded
  PGM **pgms=NULL;       // An array of pointers to pgms which are loaded
  int newWidth=0;
  int newHeight=0;
	  
  if (useFFT) {
    newWidth = getNextPowerOf2(width);
    newHeight = getNextPowerOf2(height);
  }
  else {
    newWidth = width;
    newHeight = height;
  }

  int row=0;

  fpdb = fopen(PPMStoRead,"r");
  if (fpdb == NULL)
    {
      fprintf(stderr,"Problems opening %s\n",PPMStoRead);
      return NULL;
    }

  // Need to remove the filename itself
  delimiter = strrchr(PPMStoRead,'/');
  *delimiter = '\0';
  
  int c=0;
  // preprocess the buffer to get the number of images, number of pixels in an image 
  // and convert all ppms to pgms
  while (fgets(buffer,ARR_SIZE,fpdb) != NULL)
    {
      if (buffer[0] == '.')
	break;
      // printf("Parsing :%s",buffer);
      if (buffer[0] == '!') 
	{
	  sscanf(buffer,"! %d",&numImages);
	  // now pgmFiles has numImages entries with 256 chars for each entry
	  pgmFiles = new char*[numImages];
	  for (int i=0; i<numImages; i++)
	    {
	      pgmFiles[i]=new char[ARR_SIZE];
	      strcpy(pgmFiles[i],"");
	    }
	  // now ppmFiles has numImages entries with 256 chars for each entry
	  ppmFiles = new char*[numImages];
	  for (int i=0; i<numImages; i++)
	    {
	      ppmFiles[i]=new char[ARR_SIZE];
	      strcpy(ppmFiles[i],"");
	    }
	  // now allocate space for all the ppms which will be loaded (hog!!)
	  ppms = new PPM*[numImages];
	  for (int i=0; i<numImages; i++)
	    ppms[i] = new PPM;
	  
	  // now allocate space for all the pgms which will be loaded (hog!!)
	  pgms = new PGM*[numImages];
	  for (int i=0; i<numImages; i++)
	    pgms[i] = new PGM;
	}
      else if (buffer[0] != '!' && buffer[0] != '#')
	{
	  strcat(PPMStoRead,"/");
	  strcat(PPMStoRead,buffer);
	  delimiter = strrchr(PPMStoRead,'\n');
	  *delimiter = '\0';
	  strcpy(ppmFiles[c],PPMStoRead);

	  printf("%d: Loading: %s\n",c,PPMStoRead);
	  if (!(ppms[c]->load(PPMStoRead)))
	    fprintf(stderr,"problems loading file");

	  // First have to convert PPMs to PGMs so that I can run the attention operator
	  matrix *r;
	  matrix *g;
	  matrix *b;

	  r = ppms[c]->makeMatrix(*ppms[c],'r');
	  g = ppms[c]->makeMatrix(*ppms[c],'g');
	  b = ppms[c]->makeMatrix(*ppms[c],'b');


	  // Create a matrix using the three colours combined meaningfully (meaningful according to ppmtopgm
	  matrix forPGM = (*r)*0.299 + (*g)*0.587 + (*b)*0.114;

	  delete r;
	  delete g;
	  delete b;

	  PGM pgmImage(forPGM); // create a PGM out of the rgb values

	  // fix the extension on the pgm filename
	  strcpy(pgmName,PPMStoRead);
	  delimiter = strrchr(pgmName,'.');
	  *delimiter = '\0';
	  strcat(pgmName,".pgm");
	  pgmImage.save(pgmName);
	  strcpy(pgmFiles[c],pgmName);

	  // load the pgm into mem
	  if (!useColour)
	    *(pgms[c]) = pgmImage;
	  else
	    delete pgms[c];

	  if (!gotNumPixels) 
	    {
	      numPixels = (ppms[c]->getNumRows())*(ppms[c]->getNumCols());
	      gotNumPixels = 1;
	    }

	  c++;

	  // setup buffer for next ppm to be read
	  delimiter = strrchr(PPMStoRead,'/');
	  *delimiter='\0';
	}
    }

  // Now have the size of the matrix to contain the database images as well as all filenames and 
  // also the ppms in an array.

 
  // done with file
  fclose(fpdb);
  
  printf("The number of images = %d and number of pixels = %d\n",numImages,numPixels);

  // Assign the global value of the number of images
  globalNumImages = numImages; 


  matrix *db;
  if (!useColour) // black and white only
    db = new matrix(numImages*numIntPts,newWidth*newHeight); 
  else            // Allocate space for the db and *3 because of the rgb values
    db = new matrix(numImages*numIntPts,newWidth*newHeight*3); 

  // This is the 2d globally visible interest points array, allocate space here
  dbIntPts = new interestPt*[numImages];

  //for (int i=0; i<numImages; i++)
  //  dbIntPts[i] = new interestPt[numIntPts];

  // run off the interest operator on all the pgms
  if (pforkexec(numImages,pgmFiles,intOpPath) < 0)
    {
      fprintf(stderr,"Issues getting interest points\n");
      return NULL;
    }

  printf("Interest Points found for training set\n");
 
  if (useCircular) 
    cout << "Creating matrix of sub windows around training interest points [circular]\n";
  else if (useFFT) 
    cout << "Creating matrix of sub windows around training interest points [fft]\n";
  else if (useNothing)
    cout << "Creating matrix of sub windows around training interest points\n";


  PGM pgmCopy;
  PPM ppmCopy;
  // now have to go through, and crop subwindows around interest points
  // so for every ppm, load it 
  vector v;
  for (int i=0; i<numImages; i++)
    {
      dbIntPts[i] = getInterestPoints(ppmFiles[i],numIntPts,0);
      if (drawBoxes) {
	if (!useColour)
	  pgmCopy = *(pgms[i]);
	else
	  ppmCopy = *(ppms[i]);
      }
      // In theory could get rid of all images here to free memory


      for (int j=0; j<numIntPts; j++)
	{
	  // Colour fft
	  if (useColour && useFFT) { 
	    // First crop the PPM around the interest point
	    matrix *redVals   = ppms[i]->makeMatrix(*(ppms[i]),dbIntPts[i][j],width,height,'r');
	    matrix *greenVals = ppms[i]->makeMatrix(*(ppms[i]),dbIntPts[i][j],width,height,'g');
	    matrix *blueVals  = ppms[i]->makeMatrix(*(ppms[i]),dbIntPts[i][j],width,height,'b');
	    // next make polar matrices for each color and multiply by a Gaussian (Hamming window)
	    matrix polarRed   = redVals->makePolarMatrix(*redVals,width,height);
	    matrix polarGreen = greenVals->makePolarMatrix(*greenVals,width,height);
	    matrix polarBlue  = blueVals->makePolarMatrix(*blueVals,width,height);
	    if (drawBoxes) ppmCopy.drawWindow(ppmCopy,dbIntPts[i][j],width,height);	
	    delete redVals;
	    delete greenVals;
	    delete blueVals;
	    matrix redGaussian(polarRed.dim1(),polarRed.dim2());
	    matrix greenGaussian(polarGreen.dim1(),polarGreen.dim2());
	    matrix blueGaussian(polarBlue.dim1(),polarBlue.dim2());
	    redGaussian = polarRed.makeGaussianMatrix(polarRed,sigma);
	    greenGaussian = polarGreen.makeGaussianMatrix(polarGreen,sigma);
	    blueGaussian = polarBlue.makeGaussianMatrix(polarBlue,sigma);
	    // now scale the matrices to make them powers of 2 dimesions
	    matrix redGaussianScaled(newHeight,newWidth);
	    matrix greenGaussianScaled(newHeight,newWidth);
	    matrix blueGaussianScaled(newHeight,newWidth);
	    redGaussianScaled = redGaussian.makeBiggerMatrix(redGaussian,newHeight,newWidth);
	    greenGaussianScaled = greenGaussian.makeBiggerMatrix(greenGaussian,newHeight,newWidth);
	    blueGaussianScaled = blueGaussian.makeBiggerMatrix(blueGaussian,newHeight,newWidth);
	    // fft each polar gaussian color matrix
	    fft redFft(redGaussianScaled);
	    fft greenFft(greenGaussianScaled);
	    fft blueFft(blueGaussianScaled);
	    // compute amplitudes for each colour 
	    redFft.computeLogAmplitude();
	    greenFft.computeLogAmplitude();
	    blueFft.computeLogAmplitude();
	    matrix redAmp = *(redFft.getAmplitude());
	    matrix greenAmp = *(greenFft.getAmplitude());
	    matrix blueAmp = *(blueFft.getAmplitude());
	    // combine info into a big vector
	    vector redV = redAmp.makeVector(redAmp);
	    vector greenV = greenAmp.makeVector(greenAmp);
	    vector blueV = blueAmp.makeVector(blueAmp);
	    vector *temp = combineForColour(&redV,&greenV,&blueV);
	    v = *temp;
	    delete temp;
	  } 
	  // Black and White FFT
	  else if (!useColour && useFFT) { 
	    matrix *subMatrix = pgms[i]->makeMatrix(*(pgms[i]),dbIntPts[i][j],width,height);
	    if (drawBoxes) pgmCopy.drawWindow(pgmCopy,dbIntPts[i][j],width,height);	
	    matrix polar = subMatrix->makePolarMatrix(*subMatrix,width,height);
	    delete subMatrix;
	    matrix polarGaussian(polar.dim1(),polar.dim2());
	    polarGaussian = polar.makeGaussianMatrix(polar,sigma);
	    matrix polarGaussianScaled(newHeight,newWidth);
	    polarGaussianScaled = polarGaussian.makeBiggerMatrix(polarGaussian,newHeight,newWidth);
	    fft fftobj(polarGaussianScaled);
	    fftobj.computeLogAmplitude();
	    matrix *amp = fftobj.getAmplitude();
	    v = (*amp).makeVector(*amp);
	  } 
	  // Colour circular
	  else if (useColour && useCircular) {
	    matrix *subMatrix = ppms[i]->makeGaussianMatrix(*(ppms[i]),dbIntPts[i][j],newWidth,newHeight,sigma);
	    if (drawBoxes) ppmCopy.drawWindow(ppmCopy,dbIntPts[i][j],width,height);	
	    matrix circle(subMatrix->dim1(), subMatrix->dim2());
	    circle = (*subMatrix).makeCircularMatrix(*subMatrix); 
	    delete subMatrix;
	    v = circle.makeVector(circle);
	  }
	  // Black and White circular
	  else if (!useColour && useCircular) {
	    matrix *subMatrix = pgms[i]->makeGaussianMatrix(*(pgms[i]),dbIntPts[i][j],newWidth,newHeight,sigma);
	    if (drawBoxes) pgmCopy.drawWindow(pgmCopy,dbIntPts[i][j],newWidth,newHeight);	
	    matrix circle(subMatrix->dim1(), subMatrix->dim2());
	    circle = (*subMatrix).makeCircularMatrix(*subMatrix); 
	    delete subMatrix;
	    v = circle.makeVector(circle);
	  }      
	  // Colour nothing
	  else if (useColour && useNothing) {
	    matrix *subMatrix = ppms[i]->makeGaussianMatrix(*(ppms[i]),dbIntPts[i][j],newWidth,newHeight,sigma);
	    if (drawBoxes) ppmCopy.drawWindow(ppmCopy,dbIntPts[i][j],width,height);	
	    v = (*subMatrix).makeVector(*subMatrix);
	    delete subMatrix;
	  }
	  // Black and White nothing
	  else if (!useColour && useNothing) {
	    matrix *subMatrix = pgms[i]->makeGaussianMatrix(*(pgms[i]),dbIntPts[i][j],newWidth,newHeight,sigma);
	    if (drawBoxes) pgmCopy.drawWindow(pgmCopy,dbIntPts[i][j],newWidth,newHeight);	
	    v = (*subMatrix).makeVector(*subMatrix);
	    delete subMatrix;
	  }

	  (*db)[row] = v;
	  row++;
	}

      if (drawBoxes && !useColour) {
	char newPgmName[ARR_SIZE];
	strcpy(newPgmName,pgmFiles[i]);
	strcat(newPgmName,".box");
	pgmCopy.save(newPgmName);
      }
      
      if (drawBoxes && useColour) {
	char newPpmName[ARR_SIZE];
	strcpy(newPpmName,ppmFiles[i]);
	strcat(newPpmName,".box");
	ppmCopy.save(newPpmName);
      }
    }
  
  for (int i=0; i<numImages;i++) {
    printf("Int Pts for %s\n",pgmFiles[i]);
    for (int j=0; j<numIntPts; j++) {
      cout << j << ":" << dbIntPts[i][j];
    }
  }


  for (int i=0; i<numImages; i++)
    {
      //      delete pgmFiles[i];
      //      delete ppmFiles[i];
      delete ppms[i];
      if (!useColour)
	delete pgms[i];
    }


  delete []ppms;
  delete []pgms;


  printf("Database has %d rows and %d cols\n",db->dim1(), db->dim2());
  return db;
}


// if the array has empty interestPoints, this returns the index, else -1
int 
checkForEmpty(interestPt *array, int numPoints)
{

  for (int i=0; i<numPoints; i++)
    {
      if (array[i].getx() == -1 || array[i].gety() == -1)
	return i;
    }
  return -1;
}


/*
  Rejects any interest points whose std deviation within the sub-window cropped around 
  the interest point is too "plain".  ie with low std-dev
*/
int 
checkNeighborhood(interestPt p, PPM *image, int d)
{
  int x=0,y=0;
  int x1=0,y1=0,x2=0,y2=0;
  int aver=0,aveg=0,aveb=0;
  int averS=0,avegS=0,avebS=0;
  int count=0;

  x=p.getx();
  y=p.gety();

  if (d % 2 == 0)
    {
      x1=(int)x-(d/2);
      x2=(int)x+(d/2) - 1;
      y1=(int)y-(d/2);
      y2=(int)y+(d/2) - 1;      
    }
  else
    {
      x1=(int)x-(d/2);
      x2=(int)x+(d/2);
      y1=(int)y-(d/2);
      y2=(int)y+(d/2);      
    }

  for (int i=x1; i<=x2; i++)
    {
      for (int j=y1; j<=y2; j++)
	{
	  aver += (*image)(i,j).r;
	  aveg += (*image)(i,j).g;
	  aveb += (*image)(i,j).b;
	  averS += (*image)(i,j).r * (*image)(i,j).r;
	  avegS += (*image)(i,j).g * (*image)(i,j).g;
	  avebS += (*image)(i,j).b * (*image)(i,j).b;
	  count++;
	}
    }

  float redAve = aver/(float)count;
  float greenAve = aveg/(float)count;
  float blueAve = aveb/(float)count;
  float redAveS = averS/(float)count;
  float greenAveS = avegS/(float)count;
  float blueAveS = avebS/(float)count;
  float varRed = redAveS - (redAve*redAve);
  float varGreen = greenAveS - (greenAve*greenAve);
  float varBlue = blueAveS - (blueAve*blueAve);
  float stdRed = sqrt(varRed);
  float stdGreen = sqrt(varGreen);
  float stdBlue = sqrt(varBlue);

  // cout << "stdRed = " << stdRed << " stdGreen = " << stdGreen << " stdBlue = " << stdBlue << endl;
  // cout << "This is for interestPoint " << p << endl;

  if ((stdRed < STD_DEV_THRESHOLD) && (stdBlue < STD_DEV_THRESHOLD) && (stdGreen < STD_DEV_THRESHOLD))
    {
      // cout << "Rejecting " << p << " as too plain\n";
      return 0;
    }

  if ((redAve < BG_THRESHOLD) && (blueAve < BG_THRESHOLD) && (greenAve < BG_THRESHOLD))
    {
      //cout << "Rejecting " << p << " as too black\n";
      return 0;
    }

  if ((redAve > WHITE_THRESHOLD) && (blueAve > WHITE_THRESHOLD) && (greenAve > WHITE_THRESHOLD))
    {
      //cout << "Rejecting " << p << " as too white\n";
      return 0;
    }

  return 1;
}

// fills array with interesting points which are at MOST DIST_THRESHOLD apart
// An interesting point by definition should not have a pixel value of 0 as that
// by definition is part of the matt black background.  Doesn't matter if its slow
// as it is only used during training phase
void
recursiveFill(FILE *fp, interestPt *array, int numPoints, int dist, PPM *image)
{
  char line[ARR_SIZE];
  bool first = true;
  interestPt p;


  if (checkForEmpty(array,numPoints) == -1) {
    // ie no empty spots left
    return;
  }

  if (dist < 1) {
    // printf("DISTANCE = %d (2)\n",dist);
    return;
  }

  int i = 0;
  while(fgets(line,ARR_SIZE,fp) != NULL) 
    {
      // printf("--->%s",line);
      if (first)
	{
	  p.read(line); // p is point read at that line	  
	  // Interest point cannot be on the background, N.B. distThreshold
	  // is global and can be set by -d
	  if (checkNeighborhood(p,image,distThreshold))
	    {
	      array[i] = p;
	      first = false;
	      if (i==numPoints-1)
		break;
	      i++;
	    }
	  else
	    ;// printf("%d INTERESTPOINT at %d, %d and first\n",image(p.getx(),p.gety()) ,p.getx(), p.gety());
	}
      else
	{
	  p.read(line);
	  // Interest point cannot be on the background, N.B. distThreshold
	  // is global and can be set by -d
	  if (checkNeighborhood(p,image,distThreshold))
	    {
	      //  printf("Trying dist = %d\n",dist);
	      if (sufficientlyDifferent(p,array,dist))
		{
		  array[i] = p;
		  if (i==numPoints-1)
		    break;
		  i++;
		}
	    }
	  else
	    ;// printf("%d INTERESTPOINT at %d, %d\n",image((*p).getx(),(*p).gety()),(*p).getx(),(*p).gety());
	}
    }
  
  if (checkForEmpty(array,numPoints) != -1)
    {
      rewind(fp);
      printf("Recursing with dist = %d\n",dist-1);
      recursiveFill(fp,array,numPoints,dist-1,image);
    }
}



// fills array with interesting points which are at MOST DIST_THRESHOLD apart
// An interesting point by definition should not have a pixel value of 0 as that
// by definition is part of the matt black background.  Doesn't matter if its slow
// as it is only used during training phase
void
justFill(FILE *fp, interestPt *array, int numPoints, int dist, PPM *image)
{
  char line[ARR_SIZE];
  bool first = true;
  interestPt p;

  int i = 0;
  int index=-1;
  while ((index = checkForEmpty(array,numPoints)) != -1) {
    while(fgets(line,ARR_SIZE,fp) != NULL) {
      if (first) {
	p.read(line); // p is point read at that line	  
	// Interest point cannot be on the background for training images, N.B. distThreshold
	// is global and can be set by -d	
	if (checkNeighborhood(p,image,distThreshold)) {
	  array[i] = p;
	  first = false;
	  i++;
	  cout << "First time is always the sweetest, need " << numPoints << " have " << i << "\n";
	  if (i==numPoints) {
	    cout << "Got " << numPoints << " interest points, done\n";
	    break;
	  }
	}
      }
      else { // not first time
	p.read(line);
	// Interest point cannot be on the background, N.B. distThreshold
	// is global and can be set by -d
	if (sufficientlyDifferent(p,array,dist)) {
	  if (checkNeighborhood(p,image,distThreshold)) {
	    array[i] = p;
	    i++;
	    if (i==numPoints) {
	      cout << "Got " << numPoints << " interest points, done\n";
	      break;
	    }
	  }
	}
      }
    }
    
    if (i<=numPoints) {
      rewind(fp);
      dist--;
      cout << "Not enough points obtained " << i << " dist is now " << dist << endl;
    }
  }
}




int 
sufficientlyDifferent(interestPt p, interestPt *array, float dist)
{
  int i=0;
  int success = 1;

  while (array[i].getx() != -1 || array[i].gety() != -1)
    {
      // cout << "Comparing " << array[i] << " and " << p << "\n";
      if (!sufficientlyDifferentPoints(p,array[i],dist))
	{
	  success = 0;
	  break;
	}
      i++;
    }
  return success;
}

int 
sufficientlyDifferentPoints(interestPt p, interestPt q, float dist)
{
  float deltax = q.getx() - p.getx();
  float deltay = q.gety() - p.gety();
  float hyp = sqrt(deltax*deltax + deltay*deltay);

  if (hyp >= dist)
    return 1;

  return 0;
}


/*
  Given a ppmName, it converts it to a dat name, loads that file,
  picks numIntPts from it, creates an array of interestPts from them, 
  and returns them.  Note that it assumes that the top n interest points
  it chooses are the n best (ie pre sort them)
*/
interestPt *getInterestPoints(char *ppmName, int numIntPts, int forTest)
{
  FILE *fp;
  char datName[ARR_SIZE];

  interestPt *array;
  array = new interestPt[numIntPts];
  if (array == NULL)
    cout << "HAH!!!!!!!\n";



  convertName(ppmName,datName,".dat");

  printf("Processing: %s\n",datName);
  
  fp = fopen(datName,"r");
  if (fp == NULL)
    {
      fprintf(stderr,"Problems opening %s\n",datName);
      return NULL;
    }

  PPM image;
  if (!(image.load(ppmName)))
    {
      fprintf(stderr,"problems loading file");
      return NULL;
    }
  
  
  //  recursiveFill(fp,array,numIntPts,distThreshold,&image);
  justFill(fp,array,numIntPts,distThreshold,&image);

  return array;
}



/*
  Given a ppmName, it converts it to a dat name, loads that file,
  picks numIntPts from it, creates an array of interestPts from them, 
  and returns them.  Note that it assumes that the top n interest points
  it chooses are the n best (ie pre sort them)
*/
interestPt 
*getSpecificInterestPoints(char *ppmName, int numIntPts)
{
  FILE *fp;
  char datName[ARR_SIZE];

  interestPt *overKillArray;
  interestPt *array;

  convertName(ppmName,datName,".dat");

  printf("Processing: %s\n",datName);
  
  fp = fopen(datName,"r");
  if (fp == NULL)
    {
      fprintf(stderr,"Problems opening %s\n",datName);
      return NULL;
    }


  PPM image;
  if (!(image.load(ppmName)))
    {
      fprintf(stderr,"problems loading file");
      return NULL;
    }
  
  if (overKill) {
    overKillArray = specificFill(fp,numIntPts,&image);
    cout << "Overkill Array obtained\n";    
    return overKillArray;    
  }

  // Not overKill
  array = new interestPt[numIntPts]; 
  specificFill(fp,array,numIntPts,&image); 
  cout << "Array obtained\n";    
  return array;    
}

// changes names to ext instead of .pgm or .ppm.  Assumes everything has already been
// allocated
void 
convertName(char *in, char *out, char *ext)
{
  char *delimiter;

  strcpy(out, in);
  delimiter = strrchr(out,'.');
  *delimiter = '\0';
  strcat(out,ext);
}

// changes names to ext instead of .pgm or .ppm.  Assumes everything has already been
// allocated
void 
convertNames(char **in, char **out, int n, char *ext)
{
  char *delimiter;

  for (int i=0; i<n; i++)
    {
      strcpy(out[i], in[i]);
      delimiter = strrchr(out[i],'.');
      *delimiter = '\0';
      strcat(out[i],ext);
    }
}

//  For every pgm in array of strings pgms, fork and exec what's specified in 
//  path.
int 
pforkexec(int numPgms, char **pgms, char *opPath)
{
  char *delimiter;
  int result=-1;
  int pid;
  char **outputNames;
  pid_t childpid[numPgms];
  char opName[ARR_SIZE];

  // Assuming that the operator path is the full operator path 
  // together with executable filename, separate the name from 
  // path.  Note that if the path has no "/" in it, it assumes
  // that the binary is the same as the path
  delimiter = strrchr(opPath,'/');
  if (delimiter != NULL)
    strcpy(opName,delimiter+1);
  else
    strcpy(opName,opPath);
  


  outputNames = new char*[numPgms];
  for (int i=0; i<numPgms; i++) {
    outputNames[i] = new char[ARR_SIZE];
  }

  // all the .pgms turn into .dats in outputNames
  convertNames(pgms,outputNames,numPgms,".dat");


  
  //  if (numPgms > 0 /*CHILD_MAX*/)  // can't fork that many children
  if (runLinear)
    {
      // fork each sequentially
      // printf("Too many images [%d], hence sequential forking\n",numPgms);
      printf("Running Matlabs sequentially\n");
      for (int i=0; i<numPgms; i++) {
	result = sforkexec(opPath,opName,pgms[i],outputNames[i]);
      }	  
      return result;
    }
  else
    // do everything in parallel
    {
      printf("Forking and execing interest operator\n");
      for (int i=0; i<numPgms; i++)
	{
	  if ((pid = fork()) < 0)
	    {
	      fprintf(stderr,"Fork error\n");
	      return -1;
	    }
	  if (pid > 0) // parent
	    {
	      childpid[i] = pid; // save the child's pid
	    }
	  else // child
	    {
	      // exec interest operator with correct arguments
	      if (useRemote) // run remotely on different machines
		{
		  printf("Running Interest Operator remotely on: %s\n",remoteMachines[i%NUM_RMACHINES]);
		  if ((execl(opPath,opName,pgms[i],outputNames[i],remoteMachines[i%NUM_RMACHINES],NULL)) < 0)
		    {
		      fprintf(stderr,"Exec error\n");
		      return -1;
		    }
		}
	      else 
		{
		  if ((execl(opPath,opName,pgms[i],outputNames[i],NULL)) < 0)
		    {
		      fprintf(stderr,"Exec error\n");
		      return -1;
		    }
		}
	    }
	}

      // Wait for all of them to finish.  With threads this could be much nicer
      for (int i=0; i<numPgms; i++) {
	waitpid(childpid[i],NULL,0);
	if (childpid[i] < 0) {
	  fprintf(stderr,"child %d returned with error\n",childpid[i]);
	  return -1;
	}	  
      }
    }

  for (int i=0; i<numPgms; i++)
    delete outputNames[i];
  delete []outputNames;

  return 1;
}

// Forks and execs fname with two command line args, input and output
// Output contains the interest points in the format required by this 
// code.
int 
sforkexec(char *opPath, char *opExe, char *input, char *output)
{
  int pid;

  if ((pid=fork()) < 0)
    {
      fprintf(stderr,"Fork error\n");
      return -1;
    }

  if (pid == 0)
    {
      if (execl(opPath, opExe, input, output, NULL) < 0)
	{
	  fprintf (stderr,"execl error at %s with exe:%s\n",opPath,opExe);
	  return -1;
	}
    }
  else
    waitpid(pid,NULL,0);

  return pid;
}

// some functions which are required by my kdtree template
float signatureAt(const interestPt *p, int index)
{
  interestPt tmp = *p;
  return tmp.keyAt(index % tmp.getn());
}

float distBetween(const interestPt *p, const interestPt *q)
{
  interestPt tmp1 = *p;
  interestPt tmp2 = *q;

  return tmp1.euclideanDist(tmp2);
}

// equivalent functions for the vector data structure
float vecSignatureAt(const vector *v, int index)
{
  return (*v)[index % v->dim()];
}

float vecDistBetween(const vector *v1, const vector *v2)
{
  vector difference;
  float distance=0.0;

  //  difference = *v1 - *v2;
  difference.assignDifference(v1,v2);
  distance = difference.length();
  return distance;
}
/*
void testMatrixShit()
{
  float **arr;
  arr = new float*[10];
  for (int i=0; i<10; i++)
    arr[i] = new float[10];
  int x=0;
  for (int i=0; i<10; i++)
    for (int j=0; j<10; j++) {
      arr[i][j] = x*(i+j);
      x++;
    }
  matrix test(10,10,arr);
  cout << test << "\n";
  matrix circ = test.makeCircularMatrix(test);
  cout << circ << "\n";
  matrix gauss = test.makeGaussianMatrix(test);
  cout << gauss << "\n";
  exit(1);

}
*/
int 
main(int argc, char **argv)
{

/*
  --If already have a database [matrix of subwindows] read it, else create it:
    --Read a text file which has the list of ppms which are to be used for training
    --Determine the number of images to be dealt with and the number of subwindows 
      accordingly
    --For each ppm, fork and run attention operator.  This spews output file
    --For each output file, pick the number of subwindow with coords as centers.
    --Crop the training images at those coords and build the database.
  --Read test image, run attention operator on it, crop subwindows and project
  --Classify.
*/

  struct tms clock_start;
  struct tms clock_stop;
  unsigned long tick;

  extern char *optarg;
  int ArgChar;
  char databasePath[ARR_SIZE] = DEFAULT_DATABASE_PATH;
  char interestPointPath[ARR_SIZE] = DEFAULT_INTEREST_POINT_PATH;
  char principalCompPath[ARR_SIZE] = DEFAULT_PRINC_COMP_PATH;
  char projectedDbPath[ARR_SIZE] = DEFAULT_PROJECTED_DB_PATH;
  char PPMStoRead[ARR_SIZE] = DEFAULT_PPMSFILE;
  char PPMStoTEST[ARR_SIZE] = DEFAULT_TESTPPMSFILE;
  //  char testImFname[ARR_SIZE] = DEFAULT_TEST_IM_FNAME;  //CVPR HACK

  int readTEST=0;      // whether should read a test image
  int readPPMS=0;      // whether should read the ppms file (ie contains list of ppms)
  int readTESTS=0;     // whether should read a file of test images names

  int haveResultsPath=0;

  // keep this square unless you feel gutsy
  int windowWidth = DEFAULT_SUBWINDOW_WIDTH;   // subwindow wdth
  int windowHeight = DEFAULT_SUBWINDOW_HEIGHT; // subwindow height  
  
  int manual=0;

  int dbRows = 0;
  int dbCols = 0;

  ioData *ioObject;

  fstream dbstream; // stream associated with the database matrix
  fstream ipstream; // stream associated with the interest points

  int haveTestNumIntPts=0;

  printf("Size of PRECISION = %d\n",sizeof(PRECISION));
  printf("Size of float = %d\n",sizeof(float));
  printf("Size of double = %d\n",sizeof(double));


  tick = sysconf(_SC_CLK_TCK); // get the number of ticks per second
  if (tick < 0)
    {
      fprintf(stderr,"Sysconf error for timing\n");
      exit(-1);
    }


  while ((ArgChar=getopt(argc, argv, "N:qn:e:LOs:BcFCK:SmMLAP:RI:D:d:x:y:p:T:f:t:h"))!=EOF)
    {
      switch (ArgChar)
	{
	case 'h':
	  {
	    fprintf(stderr,"-I <interest operator binary path>\n");
	    fprintf(stderr,"-d [distance threshold] (interest points will be that far apart in pixels)\n");
	    fprintf(stderr,"-s [sigma for gaussian multiplied to subwindow]\n");
	    fprintf(stderr,"-e [epsilon for classifcation]\n");
	    fprintf(stderr,"-x [width of sub window]\n");
	    fprintf(stderr,"-y [height of sub window]\n");
	    fprintf(stderr,"-c makes it use colour images\n"); 
	    fprintf(stderr,"-p [dbNumIntPts] (dbNumIntPts is the number of interest points to be used for database creation)\n");
	    fprintf(stderr,"-n [testNumIntPts] is the number of interest points that the test image should be tested with\n");
	    fprintf(stderr,"-B draws boxes of width vs height around interest point in pgms or ppms\n");
	    fprintf(stderr,"-f <filename of ppm database> [THIS REQUIRES THE FULL PATH, NOT RELATIVE PATH]\n");
	    fprintf(stderr,"-D <path of where the database, interestPoints, principal comps and projected db are>\n"); 
	    fprintf(stderr,"-P <path of where the database, interestPoints, principal comps and projected db should be stored>\n");
	    fprintf(stderr,"-t <filename of ppm to compare> [THIS REQUIRES THE FULL PATH, NOT RELATIVE PATH]\n");
	    fprintf(stderr,"-T <filename of ppm to compare database> [THIS REQUIRES THE FULL PATH, NOT RELATIVE PATH]\n");
	    fprintf(stderr,"-R use remote machines\n");
	    fprintf(stderr,"-L run matlabs linearly (sequentially)\n");
	    fprintf(stderr,"-A don't die after recognising one image\n");
	    fprintf(stderr,"-m allows user to enter all paths manually ie queried (almost same as -D)\n");
	    fprintf(stderr,"-S look for specific signatures when processing the test image\n");
	    fprintf(stderr,"-O Overkill on -S ie will search number of interest pts X number of images interest pts\n");
	    fprintf(stderr,"-M remove background explicitly\n");
	    fprintf(stderr,"-C uses the annular operator, hence planar rotation invariant\n");
	    fprintf(stderr,"-F uses the 2d fft function and examines the amplitude images, hence planar rotation invariant\n");
	    fprintf(stderr,"-K <value> sets the number of prinicpal components to use manually\n");
	    fprintf(stderr,"-N <value> searches <value> nearest neighbors in classification\n");
	    fprintf(stderr,"-q gets rid of interest points which do not conform to recognised image phase change\n");
	    exit(0);
	  }
	case 'N':
	  numNearestNeighbors = atoi(optarg);
	  break;
	case 'q':
	  correctPhase=1;
	  break;
	case 'e':
	  epsilon = atof(optarg);
	  break;
	case 'L':
	  runLinear=1;
	  break;
	case 'O':
	  overKill=1;
	  specificSig=1;
	  break;
	case 's':
	  sigma = atof(optarg);
	  break;
	case 'B':
	  drawBoxes = 1;
	  break;
	case 'c':
	  useColour = 1;
	  break;
	case 'F':
	  useFFT=1;
	  useCircular=0;
	  break;
	case 'C':
	  useCircular=1;
	  useFFT=0;
	  break;
	case 'K':
	  userK = atoi(optarg);
	  break;
	case 'M':
	  performMasking=1;
	  break;
	case 'S':
	  specificSig=1;
	  break;
	case 'm':
	  manual=1;
	  haveDatabase=1;
	  break;
	case 'A':
	  stayAlive=1;
	  break;	
	case 'P':	  
	  strcpy(globalResultsPath,optarg);
	  strcat(globalResultsPath,"/");
	  strcpy(databasePath,optarg);
	  strcat(databasePath,"/database");
	  strcpy(interestPointPath,optarg);
	  strcat(interestPointPath,"/interestPoints");
	  strcpy(principalCompPath,optarg);
	  strcat(principalCompPath,"/principalComps");
	  strcpy(projectedDbPath,optarg);
	  strcat(projectedDbPath,"/projectedDb");
	  haveDatabase=0;
	  haveResultsPath=1;
	  break;
	case 'R':
	  useRemote=1;
	  break;
	case 'I':
	  // note changing a global array here
	  strcpy(intOpPath,optarg);
	  break;
	case 'D':
	  // this is a bit fishy.  If you have the database, then you necessarily have the interest points 
	  // the principal componenets and the projected database at the same place.  So do some internal 
	  // parsing.
	  strcpy(globalResultsPath,optarg);
	  strcat(globalResultsPath,"/");
	  strcpy(databasePath,optarg);
	  strcat(databasePath,"/database");
	  strcpy(interestPointPath,optarg);
	  strcat(interestPointPath,"/interestPoints");
	  strcpy(principalCompPath,optarg);
	  strcat(principalCompPath,"/principalComps");
	  strcpy(projectedDbPath,optarg);
	  strcat(projectedDbPath,"/projectedDb");
	  haveDatabase = 1;
	  break;
	case 'd':
	  distThreshold = atoi(optarg);
	  break;
	case 't':  // only process one test image (readTESTS=0)
	  strcpy(globalTestImFname,optarg);
	  readTEST=1;
	  readTESTS=0;
	  break;
	case 'f':
	  strcpy(PPMStoRead,optarg);
	  readPPMS = 1;
	  break;
	case 'T': // process a whole bunch of test images (both set to one)
	  strcpy(PPMStoTEST,optarg);
	  readTESTS = 1;
	  readTEST=0;
	  stayAlive=0;
	  break;
	case 'x':
	  windowWidth = atoi(optarg);
	  break;
	case 'y':
	  windowHeight = atoi(optarg);
	  break;
	case 'p':
	  dbNumIntPts = atoi(optarg);
	  break;
	case 'n':
	  haveTestNumIntPts=1;
	  testNumIntPts = atoi(optarg);
	  break;
	default:
	  fprintf(stderr,"Unrecognised option, try -h\n");
	  exit(0);
	}
    }
  
  if (useFFT && useCircular) {
    fprintf(stderr,"Cannot perform FFT and Circular windows\n");
    exit(1);
  }

  if (useFFT==0 && useCircular==0)
    useNothing=1;

  matrix *database;  // will contain all the test images subwindows
  matrix *testImage; // will contain the test image subwindows
  matrix *outDb;     // will contain the eigenspace
  matrix *outTest;   // will contain the projected test image
  PrincipalComponents pc; // used by runPCA

  // used for classification of projected image points
  kdtree<vector,vecSignatureAt,vecDistBetween> *kdtOutDb = new kdtree<vector,vecSignatureAt,vecDistBetween>; 


  if (manual)
    {
      printf("Please enter the path of the database\n");
      scanf("%s",databasePath);
      printf("Please enter the path of the interest points\n");
      scanf("%s",interestPointPath);
      printf("Please enter the path of the principal components\n");
      scanf("%s",principalCompPath);
      printf("Please enter the path of the projected database\n");
      scanf("%s",projectedDbPath);
    }

  

  if (!haveDatabase) // build the database
    {
      // set up path names properly
      if (!haveResultsPath) {
	strcat(databasePath,"database");
	strcat(interestPointPath,"interestPoints");
	strcat(principalCompPath,"principalComps");
	strcat(projectedDbPath,"projectedDb");
      }

      //  Create a read/write stream for storing the database
      dbstream.open(databasePath,ios::in | ios::out | ios::trunc);
      if (!dbstream)
	fprintf(stderr,"Problems opening %s\n",databasePath);
      
      //  Create a read/write stream for storing the interest points
      ipstream.open(interestPointPath,ios::in | ios::out | ios::trunc);
      if (!ipstream)
	fprintf(stderr,"Problems opening %s\n",interestPointPath);


      dbstream.seekg(0);
      dbstream << "Created: " << PPMStoRead << "\n";


      if (readPPMS)
	database = processTrainingImages(PPMStoRead, dbNumIntPts, windowWidth, windowHeight);
      else
	{
	  printf("Please enter the FULL path of file containing ppm's to build the database with:\n");
	  scanf("%s",PPMStoRead);
	  database = processTrainingImages(PPMStoRead, dbNumIntPts, windowWidth, windowHeight);
	}

      /* 
	 getPC calls PCA::GetSVDPrincipalComponents which unfortunately needs to alter
	 database (one of its arguments).  This is to save memory since this (database)
	 can be huge.  So save it to disk first, then reload after the call.
      */

      //  bookeeping purposes, write to disk
      dbstream << "numRows " << (*database).dim1() << "\n";
      dbstream << "numCols " << (*database).dim2() << "\n";
      dbstream << "numIntPts " << dbNumIntPts << "\n";
      dbstream << "windowWidth " << windowWidth << "\n";
      dbstream << "windowHeight " << windowHeight << "\n";      
      dbstream << "circular " << useCircular << "\n";
      dbstream << "fft " << useFFT << "\n";
      dbstream << "userK " << userK << "\n";
      dbstream << "useColour " << useColour << "\n";
      
      // need to save database to disk and also the global interest points array

      ioObject = new ioData(database,dbIntPts,globalNumImages,dbNumIntPts);
      printf("Saving database\n");
      ioObject->dbWrite(dbstream);
      printf("Database created and stored in %s\n",databasePath);
      printf("Saving interest points\n");
      ipstream.seekg(0);
      ioObject->arrWrite(ipstream);
      printf("Int Pts created and stored in %s\n",interestPointPath);


      printf("Computing the eigenspace...\n");      
      // Now can call getPC to get the principal componenets
      getPC(database,&pc,true);  
      // save PC to disk
      if (!savePC(principalCompPath,&pc))
	fprintf(stderr,"Problems saving principal components");

      printf("PC created and stored in %s\n",principalCompPath);
      
      delete database;
      // Since database has been altered, reload from the disk
      printf("Reloading database from disk\n");
      dbstream.seekg(0);
      char temp[ARR_SIZE];
      dbstream >> temp >> PPMStoRead;
      dbstream >> temp >> dbRows; 
      dbstream >> temp >> dbCols; 
      dbstream >> temp >> dbNumIntPts;
      dbstream >> temp >> windowWidth;
      dbstream >> temp >> windowHeight;
      dbstream >> temp >> useCircular;
      dbstream >> temp >> useFFT;
      dbstream >> temp >> userK;
      dbstream >> temp >> useColour;

      database = new matrix(dbRows,dbCols);
      (*database).read(dbstream);
      printf("Database reloaded\n");

      // project the database onto the eigenspace
      printf("Projecting database onto eigenspace\n");
      outDb = getProjectedMatrix(database,&pc,userK);
      // build the kdtree from that
      printf("Building kdtree of %d projected images\n",outDb->dim1());
      for (int i=0; i<outDb->dim1(); i++)
	kdtOutDb->insert(&(*outDb)[i],i);

      printf("Saving projected database to disk\n");
      if (!saveMatrix(projectedDbPath,outDb))
	fprintf(stderr,"Problems saving projected database\n");
      printf("Projected database stored in %s\n",projectedDbPath);


      dbstream.close();
      ipstream.close();

      // due to constructor nightmares ie no space allocated here but be clean
      // about it.
      ioObject->clearIoData();         
    }
  else // have the database and interest points
    {
      //  Create a read stream for storing the database
      dbstream.open(databasePath,ios::in | ios::nocreate);
      if (!dbstream)
	fprintf(stderr,"Problems opening %s\n",databasePath);
      
      //  Create a read stream for storing the interest points
      ipstream.open(interestPointPath,ios::in | ios::nocreate);
      if (!ipstream)
	fprintf(stderr,"Problems opening %s\n",interestPointPath);      


      printf("Loading components\n");
      char temp[ARR_SIZE];
      dbstream >> temp >> PPMStoRead;
      dbstream >> temp >> dbRows;
      dbstream >> temp >> dbCols;
      dbstream >> temp >> dbNumIntPts;
      dbstream >> temp >> windowWidth;
      dbstream >> temp >> windowHeight;
      dbstream >> temp >> useCircular;    
      dbstream >> temp >> useFFT;
      dbstream >> temp >> userK;
      dbstream >> temp >> useColour;

      globalNumImages = dbRows/dbNumIntPts;       // Initialise the number of image (global)

      // Load the filenames for the ppms and pgms into the global ppmFiles and pgmFiles
      fillTrainingImageNames(PPMStoRead);

      // read of disk and build database and interest points array     
      // ioObject contains all the database and also the interest points
      ioObject = new ioData(dbRows, dbCols, globalNumImages, dbNumIntPts, dbstream, ipstream);
      dbstream.close();
      ipstream.close();
      database = new matrix(dbRows,dbCols);
      (*database) = ioObject->getDatabase(); // uses copy constructor
      printf("Database loaded\n");

      // allocate space for the interest points
      dbIntPts = new interestPt*[globalNumImages];
      for (int i=0; i<globalNumImages; i++)
	dbIntPts[i] = new interestPt[dbNumIntPts];
            
      for (int i=0; i<globalNumImages; i++)
	for (int j=0; j<dbNumIntPts; j++)
	  dbIntPts[i][j] = ioObject->getInterestpoint(i,j);
      printf("Interest Points loaded\n");

      if (!loadPC(principalCompPath,&pc))
	fprintf(stderr,"Problems loading principal components\n");
      printf("Eigenspace loaded\n");

      //      outDb = new matrix(database->dim1(),pc.GetK());
      outDb = new matrix(database->dim1(),userK);
      if (!loadMatrix(projectedDbPath,outDb))
	fprintf(stderr,"Problems loading projected database\n");
      // build the kdtree out of that
      printf("Building kdtree of %d projected images\n",outDb->dim1());
      for (int i=0; i<outDb->dim1(); i++)
	kdtOutDb->insert(&(*outDb)[i],i);

    }

  // free some much needed memory
  if (!performMasking) {
    delete ioObject;
    delete database;
  }  // if masking is to be done need the database           
  

  if (!haveTestNumIntPts) {
    testNumIntPts = dbNumIntPts;
  }

  // some feedback to user
  printf("The number of interest points for db creation is: %d\n",dbNumIntPts);
  printf("The number of interest points for test image is: %d\n",testNumIntPts);
  printf("The sub window width is: %d\n",windowWidth);
  printf("The sub window height is: %d\n",windowHeight);
  printf("Database dims are %d x %d\n",database->dim1(), database->dim2());  
  printf("Principal Comps:K=%d, N=%d\n",userK, pc.GetN());
  printf("useCircular = %d\n",useCircular);
  printf("useFFT = %d\n",useFFT);
  printf("useColour = %d\n",useColour);
  
  if (useCircular)
    printf("Using circular matrices as database was built with them\n");
  else if (useFFT)
    printf("Using fft amplitude images as database was built with them\n");


  char** testArray;
  int numToTest=0;
  if (readTESTS) // have a whole bunch of testPPMS specified in PPMStoTEST
    {
      numToTest = processTestImageNames(PPMStoTEST, &testArray);      
      for (int i=0; i<numToTest; i++)
	{
	  strcpy(globalTestImFname,testArray[i]);
	  times(&clock_start);
	  testImage = processTestImage(testArray[i],testNumIntPts,windowWidth,windowHeight);	  
	  // Now testImage is a matrix where each row corresponds to a subwindow around an 
	  // interest point.  
	  outTest = getProjectedMatrix(testImage,&pc,userK);	  
	  int *order=NULL;
	  if (performMasking) {
	    order  = maskAndClassify(testImage, database, outTest, kdtOutDb, &pc);
	    delete testImage;
	  }
	  else {
	    delete testImage;
	    //	    classifier(outTest,outDb);	  
	    order = kdtClassifier(outTest,kdtOutDb);  // has a new in there
	    //CORRECTIONS!!!!!
	    if (correctPhase)
	      correctForPhase(testArray[i],order,windowWidth,windowHeight,0);
	  }
	  if (order != NULL)
	    delete order;
	    


	  // stop the clock
	  times(&clock_stop);
	  printf("User time: %f secs\n",(clock_stop.tms_utime-clock_start.tms_utime)/(double)tick);
	  printf("System time: %f secs\n",(clock_stop.tms_stime-clock_start.tms_stime)/(double)tick);
	  printf("User time (children): %f secs\n",(clock_stop.tms_cutime-clock_start.tms_cutime)/(double)tick);
	  printf("System time (children): %f secs\n",(clock_stop.tms_cstime-clock_start.tms_cstime)/(double)tick);
	  delete outTest;
	  delete []testIntPts;   
	  delete []closestPair;
	}
	  
      for (int i=0; i<numToTest; i++)
	{
	  delete[] testArray[i];
	  testArray[i]=NULL;
	}
      delete[] testArray;      
    }
  else  // loop reading images one at a time
    {
      while (1)
	{
	  
	  if (!readTEST) 
	    {
	      printf("Please enter the FULL path of ppm file which you wish to compare\n");
	      scanf("%s",globalTestImFname);
	    }
	  
	  // start the clock
	  times(&clock_start);
	  testImage = processTestImage(globalTestImFname,testNumIntPts,windowWidth,windowHeight);
	  
	  // Perform PCA, memory is turning into a real fucking nightmare, be stingy with it.
	  outTest = getProjectedMatrix(testImage,&pc,userK);	  

	  int *order = NULL;
	  if (performMasking) {
	    order = maskAndClassify(testImage, database, outTest, kdtOutDb, &pc);
	    delete testImage;
	  }
	  else {
	    delete testImage;
	    //	    cout << "Non kdt classifier\n:";
	    //	    classifier(outTest,outDb);	  
	    order = kdtClassifier(outTest,kdtOutDb);
	    //CORRECTIONS!!!!!
	    if (correctPhase)
	      correctForPhase(globalTestImFname,order,windowWidth,windowHeight,0);
	  }
	  if (order != NULL)
	    delete order;


	  // stop the clock
	  times(&clock_stop);
	  printf("User time: %f secs\n",(clock_stop.tms_utime-clock_start.tms_utime)/(double)tick);
	  printf("System time: %f secs\n",(clock_stop.tms_stime-clock_start.tms_stime)/(double)tick);
	  printf("User time (children): %f secs\n",(clock_stop.tms_cutime-clock_start.tms_cutime)/(double)tick);
	  printf("System time (children): %f secs\n",(clock_stop.tms_cstime-clock_start.tms_cstime)/(double)tick);
	  delete outTest;
	  delete []testIntPts;   
	  delete []closestPair;

	  readTEST=0;
	  
	  if (!stayAlive)
	    break;
	}
    }

  // clean up


  for (int i=0; i<globalNumImages; i++)
    {
      delete pgmFiles[i];
      delete ppmFiles[i];
    }

  delete []pgmFiles;
  delete []ppmFiles;

  // free some much needed memory
  if (performMasking) {
    delete ioObject;
    delete database;
  }  // if masking is to be done need the database


  for (int i=0; i<globalNumImages; i++)
    {
      delete[] dbIntPts[i];
      dbIntPts[i]=NULL;
    }
  delete[] dbIntPts;

}

