/*

  ISSUES:

  - Why does the color seem unimportant?
    Try looking at the eigenvectors!!!!
  - Interest Points are chosen differently at times and sometimes the
    same ones are chosen in a different order.  Partial Fix would be
    to force vpselect to choose them according to the stored values of
    the interest points found for the images in the database.
  - Understand the orientation aspect of vpselect
  - How does the size of the gaussian affect results?
  - How does the size of the overlap affect results?
  - How does the size of the subwindow affect results?  <-- more
    weight to center of window!!
  - How does one account for backgrounds which are not plain?
  - What should the relationship be (if any) between the gaussian
    window and the subwindow?


    31/10/98
    ISSUE OF SCALE WITH PCA ie too big and performing on background, too small and
    too little PCA information


    3/11/98
    
    --floating point errors in qsort

    4/11/98
    
    --Have to change classifier weighting mechanism
    --Have to make it so it can take pgms too.
    --I should just use the variation values of density and orientation for corellating interest pts.
    --Have to code foveated makeMatrix


    29/11/98  COOL IDEA

    FOR BETTER CLASSIFICATION:
    
    For points that are close in the manifold, construct a new eigenspace
    with all the points, project the testimage into the new eigenspace and 
    the narrow down which image it matches to

    
  */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <fstream.h>
#include "/home/tsunami/deepu/pcaproject/deeppca/src/sharedSrc/PPM.h"
#include "/home/tsunami/deepu/pcaproject/deeppca/src/sharedSrc/PCA.h"
#include "/home/tsunami/deepu/pcaproject/deeppca/src/sharedSrc/PrincipalComponents.h"
#include "/home/tsunami/deepu/pcaproject/deeppca/src/sharedSrc/ioData.h"

#define DEFAULT_PPMSFILE "/home/tsunami/deepu/pcaproject/deeppca/ppms/ppmread"

#define DEFAULT_INPUTPPM_PATH "./input.ppm"
#define DEFAULT_OUTPUTPPM_PATH "./output.ppm"

#define DEFAULT_IMAGE_TO_COMAPARE_FILE "/home/tsunami/deepu/pcaproject/deeppca/ppms/test0.ppm"
#define DEFAULT_CANNY_PATH "/home/tsunami/deepu/vpsel/edge/src/canny"
#define DEFAULT_CANNY_INPUTFILE_PATH "/home/tsunami/deepu/vpsel/vpselect/src/face1.pgm"
#define DEFAULT_VPSELECT_PATH "/home/tsunami/deepu/vpsel/vpselect/src/vpselect" 
#define DEFAULT_VPSELECT_INPUTFILE_PATH "/home/tsunami/deepu/pcaproject/deeppca/ppms/inDen" 
#define DEFAULT_DATABASE_PATH "/home/tsunami/deepu/pcaproject/deeppca/ppms/database"
#define DEFAULT_INTEREST_POINT_PATH "/home/tsunami/deepu/pcaproject/deeppca/ppms/interestPoints"
#define DEFAULT_NUM_INTEREST_PTS 5
#define DEFAULT_SUBWINDOW_WIDTH 40
#define DEFAULT_SUBWINDOW_HEIGHT 40
#define DEFAULT_GAUSSIAN_WIDTH 40
#define DEFAULT_GAUSSIAN_HEIGHT 40
#define DEFAULT_GAUSSIAN_VARIANCE 30 // typically 2/3 value of widht or height

#define DIST_THRESHOLD 30
#define BG_THRESHOLD 25

#define ARR_SIZE 256

// Some globals
int numberOfImages = 0;
int densityOnly = 1;      // By default vpselect uses only density
int orientationOnly = 0;
int densityAndOrientation = 0;
interestPoint **globalArray = NULL;


// Prototypes;
void runPCA(matrix *database, matrix *testImage);
void testPCA();
int classifier(vector in, matrix m);
int classifier(matrix testImage, matrix database);
interestPoint* getPoint(char *buffer);
int sufficientlyDifferent(interestPoint p, interestPoint *array, double dist);
int sufficientlyDifferentPoints(interestPoint p, interestPoint q, double dist);
matrix *setupDatabase(char *PPMStoRead, char *cannyPath, char *vpselPath, char *vpselInputPath, 
		      int numPoints, int windowWidth, int windowHeight, int gaussianWidth, 
		      int gaussianHeight, int gaussianVariance);
matrix *processTestImage(char *inputFile, char *cannyPath, char *vpselPath, char *vpselInputPath, 
			 int numPoints, int width, int height, int gaussianWidth, 
			 int gaussianHeight, int gaussianVariance);
int checkForEmpty(interestPoint *array, int numPoints);
void recursiveFill(FILE *fp, interestPoint *array, int numPoints, int dist, PGM image);
void specificFill(FILE *fp, interestPoint *array, int numPoints);
int sortInterest(interestPoint *p1, interestPoint *p2);
interestPoint* getClosestPoint(FILE *fp, interestPoint* point);
interestPoint* getInterestPoints(char *vpselExe, int numPoints, PGM image);
interestPoint* getSpecificInterestPoints(char *vpselExe, int numPoints);
int getMaxIndex(int *array, int arrSize);

void 
runPCA(matrix *database, matrix *testImage)
{

  PCA pcaObject;
  PrincipalComponents result;
  
  // pcaObject.SetVerbosity(4);
  bool projectResult;
  // the boolean value normalises the row vector to 1 (ie images) if true.
  // This has the effect of accounting for intensity
  result = pcaObject.GetSVDPrincipalComponents(*database,true);
  // result = pcaObject.GetPrincipalComponents(*database,true);
  // cout << "K=" << result.GetK() << "\n";
  // cout << "N=" << result.GetN() << "\n";

  // in_matrix=MxN, out_matrix = MxK
  matrix outDb((*database).dim1(),result.GetK());      // this becomes the matrix of the projected images
  matrix outTest((*testImage).dim1(),result.GetK());   // this becomes the matrix of the projected test image;

  projectResult = result.Project(*database, outDb);
  projectResult = result.Project(*testImage, outTest);

  int index = classifier(outTest,outDb);
  printf("index = %d\n",index);
}


void 
testPCA()
{
  bool projectResult;
  PCA pcaObject;
  PrincipalComponents result;
  PPM square;
  PPM circle;
  PPM triangle;
  PPM ssquare;
  
  if (!square.load("./ppms/shape0.ppm"))
    fprintf(stderr,"problems loading file");

  if (!circle.load("./ppms/shape1.ppm"))
    fprintf(stderr,"problems loading file");

  if (!triangle.load("./ppms/shape2.ppm"))
    fprintf(stderr,"problems loading file");
  
  if (!ssquare.load("./ppms/test0.ppm"))
    fprintf(stderr,"problems loading file");


  matrix *s;
  matrix *c;
  matrix *t;
  matrix *ss;

  vector v;

  s = square.makeMatrix(square);
  c = circle.makeMatrix(circle);
  t = triangle.makeMatrix(triangle);
  ss = ssquare.makeMatrix(ssquare);

  v = (*s).makeVector(*s);

  matrix database(3,v.dim());

  database[0] = v;

  v = (*c).makeVector(*c);

  database[1] = v;

  v = (*t).makeVector(*t);

  database[2] = v;

  v = (*ss).makeVector(*ss);

  //cout << database;

  result = pcaObject.GetSVDPrincipalComponents(database,true);

  cout << "K=" << result.GetK() << "\n";
  cout << "N=" << result.GetN() << "\n";
  cout << "Number of pixels=" << v.dim() << "\n";

  matrix outm(3,3); // this becomes the matrix of the projected images
  vector outv(3);   // this becomes the vector of the projected test image;

  projectResult = result.Project(database, outm);
  projectResult = result.Project(v, outv);

  cout << "outv=" << outv;
  cout << "outm=" << outm;


  int index = classifier(outv,outm);
  printf("index = %d\n",index);

}


int 
getMaxIndex(int *array, int arrSize)
{
  int max=-99999;
  int index;
  
  for (int i=0; i<arrSize; i++)
    {
      if (array[i] > max)
	{
	  max = array[i];
	  index = i;
	}
    }
  return index;
}
  

int 
classifier(vector in, matrix m)
{
  double minDistance=99999999;
  int index;
  double distance;
  vector difference;

  for (int i=0; i < m.dim1(); i++)
    {
      // cout << "Vector:" << in << ", Matrix" << m; 
      difference = m[i] - in;
      distance = difference.sqr_length();
      // cout << i << ":" << distance << "\n";
      if (distance < minDistance)
	{
	  minDistance = distance;
	  index = i;
	}
    }

  cout << "Min Distance = " << minDistance << "\n";
  cout << "Index = " << index << "\n";

  return index;
}


int 
classifier(matrix testImage, matrix database)
{
  int index;
  vector difference;
  int buckets[database.dim1()];
  int numImages = (int)(database.dim1()/testImage.dim1());
  int values[numImages];

  // Initialise the buckets to 0
  for (int i=0; i<database.dim1(); i++)
    buckets[i] = 0;

  for (int i=0; i<numImages; i++)
    values[i] = 0;

  int numPoints = testImage.dim1(); //number of row == number of matrices around sub-points

  // For every test Image (ie matrix around vp select point) look for the closest row in 
  // the database and increment a bucket value at that index     
  for (int i=0; i<testImage.dim1(); i++)
    {
      vector testVector = testImage[i];
      int where = classifier(testVector,database);
      // Weight the viewpoints such that the most importance is placed on the first
      // point and least for last point.  This is entirely linear
      buckets[where] += (numPoints - i); 
    }

 
  for (int i=0; i<database.dim1(); i++)
    {
      // printf("BUCKET VALUE %d=%d\n",i,buckets[i]);
      if (buckets[i] != 0)
	values[(int)(i/numPoints)] += buckets[i];
    }

    
  for (int i=0; i<numImages; i++)
    printf("VALUE[%d]=%d\n",i,values[i]);

  index = getMaxIndex(values,numImages);

  printf("Best Match is with image %d\n",index);

  /*

  for (int i=0; i < database.dim1(); i+=numPoints)
    {
      // cout << "Vector:" << in << ", Matrix" << m; 
      
      // create subMatrices of the database as each image is represented by numPoints rows
      matrix subMatrix(numPoints,database.dim2());
      for (int j=0; j<numPoints;j++) {
	// printf("row = %d\n",i+j);
	subMatrix[j] = database[i+j];
      }

      // subtract the two matrices
      matrix difference = subMatrix - testImage;
      
      // convert the result into a vector
      vector diffVec = difference.makeVector(difference);
      
      distance = diffVec.sqr_length();
      printf("Distance to image %d = %f\n",(int)(i/numPoints),distance);
      if (distance < minDistance)
	{
	  minDistance = distance;
	  index = (int)i/numPoints;
	}
    }
    
  cout << "Min Distance = " << minDistance << "\n";
  cout << "Index = " << index << "\n";
  

  */


  return index;
}

interestPoint*
getPoint(char *buffer)
{
  char filename[ARR_SIZE];
  char scraps[ARR_SIZE];


  int x=-1;
  int y=-1;
  double densityVal = -1;
  double densityVar = -1;
  double orientationVal = -1;
  double orientationVar = -1;
  double orientationStr = -1;

  if (densityOnly)
    sscanf(buffer,"%[^\t]%d\t%d\t%lf\t%lf%[^\n]"
	   ,filename,&x,&y,&densityVal,&densityVar,scraps);
  else if (orientationOnly)
    sscanf(buffer,"%[^\t]%d\t%d\t%lf\t%lf\t(%lf)%[^\n]"
	   ,filename,&x,&y,&orientationVal,&orientationVar, &orientationStr, scraps);
  else if (densityAndOrientation)    
    sscanf(buffer,"%[^\t]%d\t%d\t(%lf,%lf)\t(%lf,%lf)%[^\n]"
	   ,filename,&x,&y,&densityVal,&orientationVal,&densityVar,&orientationVar, scraps); 

  interestPoint *p = new interestPoint(x,y,densityVal,orientationVal,densityVar,orientationVar,orientationStr);


  return p;
}



// if the array has empty interestPoints, this returns the index, else -1
int 
checkForEmpty(interestPoint *array, int numPoints)
{

  for (int i=0; i<numPoints; i++)
    {
      if (array[i].getx() == -1 || array[i].gety() == -1)
	return i;
    }
  return -1;
}


// fills array with interesting points which are at MOST DIST_THRESHOLD apart
// An interesting point by definition should not have a pixel value of 0 as that
// by definition is part of the matt black background.
void
recursiveFill(FILE *fp, interestPoint *array, int numPoints, int dist, PGM image)
{
  char line[ARR_SIZE];
  bool first = true;
  interestPoint *p;

  if (checkForEmpty(array,numPoints) == -1) {
    // printf("DISTANCE = %d\n",dist);
    return;
  }

  if (dist < 1) {
    // printf("DISTANCE = %d (2)\n",dist);
    return;
  }

  int i = 0;
  while(fgets(line,ARR_SIZE,fp) != NULL) 
    {
      // printf("--->%s",line);
      if (line[0] != '!' && line[0] != '#' && first)
	{
	  p = getPoint(line); // p is point read at that line	  
	  //  Maybe I should check the surrounding values around the pixel too
	  if (image((*p).getx(),(*p).gety()) > BG_THRESHOLD) // Interest point cannot be on the background
	    {
	      array[i] = *p;
	      first = false;
	      if (i==numPoints-1)
		break;
	      i++;
	    }
	  else
	    ;// printf("%d INTERESTPOINT at %d, %d and first\n",image((*p).getx(),(*p).gety()) ,(*p).getx(), (*p).gety());
	}
      else if (line[0] != '!' && line[0] != '#')
	{
	  p = getPoint(line);
	  //  Maybe I should check the surrounding values around the pixel too
	  if (image((*p).getx(),(*p).gety()) > BG_THRESHOLD) // Interest point cannot be on the background
	    {
	      //  printf("Trying dist = %d\n",dist);
	      if (sufficientlyDifferent(*p,array,dist))
		{
		  array[i] = *p;
		  if (i==numPoints-1)
		    break;
		  i++;
		}
	    }
	  else
	    ;// printf("%d INTERESTPOINT at %d, %d\n",image((*p).getx(),(*p).gety()),(*p).getx(),(*p).gety());
	}
    }
  
  if (checkForEmpty(array,numPoints) != -1)
    {
      rewind(fp);
      // printf("Recursing with dist = %d\n",dist-1);
      recursiveFill(fp,array,numPoints,dist-1,image);
    }
}

// Given a file of interest points and a point, this searches the file for the closest
// corresponding point in the file to the specified point by examining L2 norms.
interestPoint *
getClosestPoint(FILE *fp, interestPoint *point)
{
  char line[ARR_SIZE];
  interestPoint *p;
  interestPoint *closestPoint = new interestPoint();
  double density = 0.0;
  double varDensity = 0.0;
  double aveDensity = 0.0;      // since have two comparison fields, compare to their average
  double minDenCompare = 999999.0;
  double orientation = 0.0;
  double varOrientation = 0.0;
  double aveOrientation = 0.0;  // since have two comparison fields, compare to their average
  double minOrtCompare = 999999.0;
  double aveAll = 0.0;
  double minAllCompare = 999999.0;

  // initialise the max values to those of point as when finding max's for 
  // a specific fp, its possible that the max values are greater in point.
  double maxDensity = (*point).getDensity();
  double maxOrientation = (*point).getOrientation();
  double maxVarDensity = (*point).getVarDensity();
  double maxVarOrientation = (*point).getVarOrientation();

  // Go through once and get all the max values in fp

  while(fgets(line, ARR_SIZE, fp) != NULL)
    {
      if (line[0] != '!' && line[0] != '#')
	{
	  p = getPoint(line);
	  density = (*p).getDensity();
	  varDensity = (*p).getVarDensity();
	  orientation = (*p).getOrientation();
	  varOrientation = (*p).getVarOrientation();
	  if (density >= maxDensity)
	    maxDensity = density;
	  if (orientation >= maxOrientation)
	    maxOrientation = orientation;
	  if (varDensity >= maxVarDensity)
	    maxVarDensity = varDensity;
	  if (varOrientation >= maxVarOrientation)
	    maxVarOrientation = varOrientation;
	}
    }
  
  // cout << "Max values are (d,o,vd,vo) " << maxDensity << "," << maxOrientation << "," << maxVarDensity << "," << maxVarOrientation << "\n"; 

  rewind(fp);
  density=varDensity=orientation=varOrientation=0.0;

  double maxDenSquare = maxDensity * maxDensity;
  double maxOrtSquare = maxOrientation * maxOrientation;
  double maxVarDenSquare = maxVarDensity * maxVarDensity;
  double maxVarOrtSquare = maxVarOrientation * maxVarOrientation;

  cout << "Comparing to "; (*point).printDetail(); cout << "\n";

  while(fgets(line, ARR_SIZE, fp) != NULL)
    {
      if (line[0] != '!' && line[0] != '#')
	{
	  p = getPoint(line);
	  if (densityOnly) 
	    {
	      density = (((*p).getDensity() - (*point).getDensity()) * 
			 ((*p).getDensity() - (*point).getDensity())) / maxDenSquare;
	      varDensity = (((*p).getVarDensity() - (*point).getVarDensity()) *  
			    ((*p).getVarDensity() - (*point).getVarDensity())) / maxVarDenSquare;
	      aveDensity = density + varDensity;
	      
	      if (aveDensity < minDenCompare)
		{
		  minDenCompare = aveDensity;
		  *closestPoint = *p;
		  (*closestPoint).setConfidence(aveDensity);		  
		}
	    }
	  else if (orientationOnly)
	    {
	      orientation = (((*p).getOrientation() - (*point).getOrientation()) * 
			     ((*p).getOrientation() - (*point).getOrientation())) / maxOrtSquare;
	      varOrientation = (((*p).getVarOrientation() - (*point).getVarOrientation()) * 
				((*p).getVarOrientation() - (*point).getVarOrientation())) / maxVarOrtSquare ;
	      aveOrientation = orientation + varOrientation;
	      
	      if (aveOrientation < minOrtCompare)
		{
		  minOrtCompare = aveOrientation;
		  *closestPoint = *p;
		  (*closestPoint).setConfidence(aveOrientation);
		}	  
	    }
	  else if (densityAndOrientation)
	    {
	      density = (((*p).getDensity() - (*point).getDensity()) * 
			 ((*p).getDensity() - (*point).getDensity())) / maxDenSquare;
	      varDensity = (((*p).getVarDensity() - (*point).getVarDensity()) *  
			    ((*p).getVarDensity() - (*point).getVarDensity())) / maxVarDenSquare;
	      orientation = (((*p).getOrientation() - (*point).getOrientation()) * 
			     ((*p).getOrientation() - (*point).getOrientation())) / maxOrtSquare;
	      varOrientation = (((*p).getVarOrientation() - (*point).getVarOrientation()) * 
				((*p).getVarOrientation() - (*point).getVarOrientation())) / maxVarOrtSquare ;
	      
	      aveAll = density + varDensity + orientation + varOrientation;
	      
	      if (aveAll < minAllCompare)
		{
		  minAllCompare = aveAll;
		  *closestPoint = *p;
		  (*closestPoint).setConfidence(aveAll);
		}
	    }
	}
    }

  cout << "Closest is "; (*closestPoint).printDetail(); cout << "\n";

  return closestPoint;
}



// fills array with interesting points which are similar to those specified in globalArray
void
specificFill(FILE *fp, interestPoint *array, int numPoints)
{
  interestPoint *p;
  interestPoint bigArray[numPoints*numberOfImages];  // contains all the best matching interest points
  int k=0;

  rewind(fp);

  for (int i=0; i<numberOfImages; i++) 
    for (int j=0; j<numPoints; j++)
      {
	p = getClosestPoint(fp,&globalArray[i][j]);
	// printf("Closest Match to %d: ",i); globalArray[i][j].printDetail(); printf(" is "); (*p).printDetail(); printf("\n");
	bigArray[k] = *p;
	k++;
	rewind(fp);
      }
  
  qsort(bigArray, numPoints*numberOfImages, sizeof(interestPoint), sortInterest);

  
  printf("SORTED\n");
  for (int i=0; i<numPoints*numberOfImages; i++)
    {
      printf("%d:",i);
      bigArray[i].printDetail();
      printf("\n");
    }
  

  // copy the best matches to the array required
  for (int i=0; i<numPoints; i++)
    array[i] = bigArray[i];
  
}


int 
sortInterest(interestPoint *p1, interestPoint *p2)
{
  double lowerBound = -0.0000001;
  double upperBound =  0.0000001;
  double difference = (*p1).getConfidence() - (*p2).getConfidence();

  if (difference > upperBound)
    return 1;

  if (difference < lowerBound)
    return -1;

  if ((difference > lowerBound) && (difference < upperBound))
    return 0;

  return -2;
}


// IF I DON"T GET ENOUGH INTEREST POINTS I'M FUCKED OVER HERE!!!!
// This is the function I use to build the database.  If at anypoint 
// I pre segment the image, it is in that segmented place that I should 
// find the view points.
interestPoint* 
getInterestPoints(char *vpselExe, int numPoints, PGM image)
{
  FILE *fpout;  // for the pipe from vpselect to here
  FILE *output; // for output file to dump pipe info into
  char line[ARR_SIZE];
  char vpselOutput[ARR_SIZE];


  interestPoint *array;
  array = new interestPoint[numPoints];

  //  for (int i=0; i<numPoints; i++)
  //  printf("[%d,%d]\n",array[i].getx(),array[i].gety());

  //  printf("popening %s\n",vpselExe);

  if ((fpout = popen(vpselExe,"r")) == NULL) {
    perror("popen eror");
  }

  strcpy(vpselOutput,"vpselOutput");

  // creates a file to dump vpselect output into
  output = fopen(vpselOutput,"w+");
  if (output == NULL)
    {
      fprintf(stderr,"Problems opening %s\n",vpselOutput);
      return NULL;
    }

  // suck from pipe and dump into file
  while(fgets(line,ARR_SIZE,fpout) != NULL) {
    // printf("%s\n",line);
    fputs(line,output);
  }

  rewind(output);
  recursiveFill(output, array, numPoints, DIST_THRESHOLD, image);

  fclose(output);
  pclose(fpout);
  return array;
}




interestPoint* 
getSpecificInterestPoints(char *vpselExe, int numPoints, PGM image)
{
  FILE *fpout;  // for the pipe from vpselect to here
  FILE *output; // for output file to dump pipe info into
  char line[ARR_SIZE];
  char vpselOutput[ARR_SIZE];



  interestPoint *array;
  array = new interestPoint[numPoints];

  //  for (int i=0; i<numPoints; i++)
  //  printf("[%d,%d]\n",array[i].getx(),array[i].gety());

  //  printf("popening %s\n",vpselExe);

  if ((fpout = popen(vpselExe,"r")) == NULL) {
    perror("popen eror");
  }

  strcpy(vpselOutput,"vpselOutput");

  // creates a file to dump vpselect output into
  output = fopen(vpselOutput,"w+");
  if (output == NULL)
    {
      fprintf(stderr,"Problems opening %s\n",vpselOutput);
      return NULL;
    }

  // suck from pipe and dump into file
  while(fgets(line,ARR_SIZE,fpout) != NULL) {
    // printf("%s\n",line);
    fputs(line,output);
  }

  rewind(output);
  //  Want to get points that are specified in inArray
  specificFill(output, array, numPoints);

  fclose(output);
  pclose(fpout);
  return array;
}



int 
sufficientlyDifferent(interestPoint p, interestPoint *array, double dist)
{
  int i=0;
  int success = 1;

  while (array[i].getx() != -1 || array[i].gety() != -1)
    {
      // cout << "Comparing " << array[i] << " and " << p << "\n";
      if (!sufficientlyDifferentPoints(p,array[i],dist))
	{
	  success = 0;
	  break;
	}
      i++;
    }
  return success;
}

int 
sufficientlyDifferentPoints(interestPoint p, interestPoint q, double dist)
{
  double deltax = q.getx() - p.getx();
  double deltay = q.gety() - p.gety();
  double hyp = sqrt(deltax*deltax + deltay*deltay);

  if (hyp >= dist)
    return 1;

  return 0;
}



matrix *
processTestImage(char *inputFile, char *cannyPath, char *vpselPath, char *vpselInputPath, 
		 int numPoints, int width, int height, int gaussianWidth, 
		 int gaussianHeight, int gaussianVariance)
{
  FILE *fpvp;
  char cannyExe[ARR_SIZE];
  char vpselExe[ARR_SIZE];
  char openVpFile[ARR_SIZE];
  char *delimiter;
  char pgmName[ARR_SIZE];
  int row=0;

  matrix *result = new matrix(numPoints,width*height*3); // 3 as rgb values

  PPM image;

  printf("loading %s\n",inputFile);
  if (!image.load(inputFile))
    fprintf(stderr,"problems loading file");


  // First have to convert PPMs to PGMs so that I can run canny and vpselect on them
  matrix *r;
  matrix *g;
  matrix *b;
	
  r = image.makeMatrix(image,'r');
  g = image.makeMatrix(image,'g');
  b = image.makeMatrix(image,'b');

  // Create a matrix using the three colours combined meaningfully
  matrix forPGM = *r + *g + *b;

  PGM pgmImage(forPGM); // create a PGM out of the rgb values

  // fix the extension on the pgm filename
  strcpy(pgmName,inputFile);
  delimiter = strrchr(pgmName,'.');
  *delimiter = '\0';
  strcat(pgmName,".pgm");
  pgmImage.save(pgmName);

  // Run canny on the pgm image to get .th and .edg files
  strcpy(cannyExe,""); // clear the buffer
  strcat(cannyExe,cannyPath);
  strcat(cannyExe," ");
  strcat(cannyExe,pgmName);

  // printf("pgmName=%s\ncannyExe=%s\n",pgmName,cannyExe);
  system(cannyExe);

  // Now write info to created input file for vpselect
  fpvp = fopen(vpselInputPath,"w+");
  if (fpvp == NULL)
    {
      fprintf(stderr,"Problems opening %s\n",vpselInputPath);
      return NULL;
    }

  // fputs("set verbosity 2\n",fpvp);
  strcpy(openVpFile,"");
  strcat(openVpFile,"open ");
  strcat(openVpFile,pgmName);
  strcat(openVpFile,"\n");
  fputs(openVpFile,fpvp);
  fputs("overlap 0.9\n",fpvp);
  char temp[ARR_SIZE];
  sprintf(temp,"gaussian %d %d %d\n",gaussianWidth,gaussianHeight,gaussianVariance);
  fputs(temp,fpvp);
  if (densityOnly)
    fputs("density 1000 0 0 0 0\n",fpvp);
  else if (orientationOnly)
    fputs("orientation 1000 0 0 0 0\n",fpvp);
  else if (densityAndOrientation)
    fputs("both 1000 0 0 0 0\n",fpvp);
  fputs("quit\n",fpvp);
  
  fclose(fpvp);
	  

  // Now run vpselect with this as the input file and grab interest points
  strcpy(vpselExe,"");
  strcat(vpselExe,"cat ");
  strcat(vpselExe,vpselInputPath);
  strcat(vpselExe," | ");
  strcat(vpselExe,vpselPath);

  interestPoint *array;
  array = getSpecificInterestPoints(vpselExe,numPoints,pgmImage);
  for (int i=0; i<numPoints;i++)
    cout << array[i];
  cout << "\n";

  for (int np=0; np<numPoints && ((array[np].getx() != -1) && (array[np].gety() != -1)); np++)
    {
      // For each interest point make a sub matrix, then proceed to convert that sub matrix 
      // into a row vector which is then placed as a row into the database matrix
      matrix *subMatrix;
      subMatrix = image.makeGaussianMatrix(image,array[np],width,height);
      //	      cout << *subMatrix << "\n";
      vector v;
      v = (*subMatrix).makeVector(*subMatrix);	      
      (*result)[row] = v;
      row++;
      
    }

  return result;
}
  
// DATABASE size is dependent on the number of interest points, number of images, and size of submatrix.
// In fact the number of columns is the size of the submatrix, and the number of rows is the number of 
// images*numPoints
matrix *
setupDatabase(char *PPMStoRead, char *cannyPath, char *vpselPath, char *vpselInputPath, 
	      int numPoints, int width, int height, int gaussianWidth, 
	      int gaussianHeight, int gaussianVariance)
{
  FILE *fpdb;    // databse of PPMS to read
  FILE *fpvp;    // vp select input file
  char buffer[ARR_SIZE];
  char cannyExe[ARR_SIZE];
  char vpselExe[ARR_SIZE];
  char openVpFile[ARR_SIZE];
  char *delimiter;
  char pgmName[ARR_SIZE];
  int numImages=0;
  int numPixels=0;
  int row = 0;
  int imageNumber = 0;

  fpdb = fopen(PPMStoRead,"r");
  if (fpdb == NULL)
    {
      fprintf(stderr,"Problems opening %s\n",PPMStoRead);
      return NULL;
    }



  // Need to remove the filename itself
  delimiter = strrchr(PPMStoRead,'/');
  *delimiter = '\0';

  
  // preprocess the buffer to get the number of images and an example ppm
  while (fgets(buffer,ARR_SIZE,fpdb) != NULL)
    {
      if (buffer[0] == '!')
	sscanf(buffer,"! %d",&numImages);
      else if (buffer[0] != '!' && buffer[0] != '#')
	{
	  strcat(PPMStoRead,"/");
	  strcat(PPMStoRead,buffer);
	  delimiter = strrchr(PPMStoRead,'\n');
	  *delimiter = '\0';
	  
	  // printf("Trying to load %s\n",PPMStoRead);
	  PPM testImage;
	  
	  if (!testImage.load(PPMStoRead))
	    fprintf(stderr,"problems loading file");
	  
	  numPixels = testImage.getNumRows()*testImage.getNumCols();
	  // setup buffer for next ppm to be read
	  delimiter = strrchr(PPMStoRead,'/');
	  *delimiter='\0';
	  break;
	}
    }
  // Now have the size of the matrix to contain the database images
  
  // reset the file pointer
  rewind(fpdb);
  
  // printf("The number of images=%d and number of pixels=%d\n",numImages,numPixels);

  numberOfImages = numImages; // Assign the global value of the number of images
  matrix *database = new matrix(numImages*numPoints,width*height*3); // 3 as rgb values


  //  This is the 2d global array of interestPoints which is allocated here
  globalArray = new interestPoint*[numImages];
  for (int i=0; i<numImages; i++)
    globalArray[i] = new interestPoint[numPoints];

  while(fgets(buffer,ARR_SIZE,fpdb) != NULL)   
    {
      if (buffer[0] != '!' && buffer[0] != '#')
	{
	  strcat(PPMStoRead,"/");
	  strcat(PPMStoRead,buffer);
	  delimiter = strrchr(PPMStoRead,'\n');
	  *delimiter = '\0';
	  
	  printf("loading %s\n",PPMStoRead);

	  PPM ppmImage;
	  
	  if (!ppmImage.load(PPMStoRead))
	    fprintf(stderr,"problems loading file");


	  // CORE OF CODE

	  // First have to convert PPMs to PGMs so that I can run canny and vpselect on them
	  matrix *r;
	  matrix *g;
	  matrix *b;
	
	  r = ppmImage.makeMatrix(ppmImage,'r');
	  g = ppmImage.makeMatrix(ppmImage,'g');
	  b = ppmImage.makeMatrix(ppmImage,'b');

	  // Create a matrix using the three colours combined meaningfully
	  matrix forPGM = *r + *g + *b;

	  PGM pgmImage(forPGM); // create a PGM out of the rgb values

	  // fix the extension on the pgm filename
	  strcpy(pgmName,PPMStoRead);
	  delimiter = strrchr(pgmName,'.');
	  *delimiter = '\0';
	  strcat(pgmName,".pgm");
	  pgmImage.save(pgmName);

	  // Run canny on the pgm image to get .th and .edg files
	  strcpy(cannyExe,""); // clear the buffer
	  strcat(cannyExe,cannyPath);
	  strcat(cannyExe," ");
	  strcat(cannyExe,pgmName);

	  // printf("pgmName=%s\ncannyExe=%s\n",pgmName,cannyExe);
	  system(cannyExe);

	  // Now write info to created input file for vpselect
	  fpvp = fopen(vpselInputPath,"w+");
	  if (fpvp == NULL)
	    {
	      fprintf(stderr,"Problems opening %s\n",vpselInputPath);
	      return NULL;
	    }

	  // fputs("set verbosity 2\n",fpvp);
	  strcpy(openVpFile,"");
	  strcat(openVpFile,"open ");
	  strcat(openVpFile,pgmName);
	  strcat(openVpFile,"\n");
	  fputs(openVpFile,fpvp);
	  fputs("overlap 0.9\n",fpvp);
	  char temp[ARR_SIZE];
	  sprintf(temp,"gaussian %d %d %d\n",gaussianWidth,gaussianHeight,gaussianVariance);
	  fputs(temp,fpvp);
	  if (densityOnly)
	    fputs("density 1000 0 0 0 0\n",fpvp);
	  else if (orientationOnly)
	    fputs("orientation 1000 0 0 0 0\n",fpvp);
	  else if (densityAndOrientation)
	    fputs("both 1000 0 0 0 0\n",fpvp);
	  fputs("quit\n",fpvp);

	  fclose(fpvp);
	  

	  // Now run vpselect with this as the input file and grab interest points
	  strcpy(vpselExe,"");
	  strcat(vpselExe,"cat ");
	  strcat(vpselExe,vpselInputPath);
	  strcat(vpselExe," | ");
	  strcat(vpselExe,vpselPath);

	  //  Needs the PGM file so it can check out bg values
	  globalArray[imageNumber] = getInterestPoints(vpselExe,numPoints,pgmImage);

	  cout << "Interest Points for " << pgmName << "\n";
	  for (int i=0; i<numPoints;i++)
	    cout << globalArray[imageNumber][i];
	  cout << "\n";

	  for (int np=0; np<numPoints && ((globalArray[imageNumber][np].getx() != -1) && (globalArray[imageNumber][np].gety() != -1)); np++)
	    {
	      // For each interest point make a sub matrix, then proceed to convert that sub matrix 
	      // into a row vector which is then placed as a row into the database matrix
	      matrix *subMatrix;
	      subMatrix = ppmImage.makeGaussianMatrix(ppmImage,globalArray[imageNumber][np],width,height);
	      //	      cout << *subMatrix << "\n";
	      vector v;
	      v = (*subMatrix).makeVector(*subMatrix);	      
	      (*database)[row] = v;
	      row++;
	      
	    }
	  delimiter = strrchr(PPMStoRead,'/');
	  *delimiter='\0';
	  imageNumber++;
	}
    }

  fclose(fpdb);
  printf("NumRows = %d, row = %d\n",(*database).dim1(),row);
  return database;
}
/* 
   -- Read in a file of PPMs (determine size of database matrix accordingly)
   -- Convert the PPMs to PGMs (for erics code)
   -- Run canny on all the PGMs
   -- Send PGMs to vpselect (done by creating a file for each PGM)
   -- get 4 to 5 interesting viewpoints that are far apart
   -- Build database matrix using sub windows obtained from vpselect
 

   To Do:

   -- First write clever makematrix
   -- Combine rgb values meaningfully.
   -- Want to make code background invariant...this is accomplished by
      training on a black background and looking for interest points
      that lie strictly within the object ie no black pixels in sub
      window around the image.  Then the values for these interest
      points are looked for when using vpselect on the test images.
   -- Want to add rotation invariance.   


   

   */

int main(int argc, char **argv)
{

  extern char *optarg;
  int ArgChar;
  int numPoints = DEFAULT_NUM_INTEREST_PTS;
  char PPMStoRead[ARR_SIZE] = DEFAULT_PPMSFILE;
  int readPPMS = 0;
  int haveDatabase = 0;
  int readTEST = 0;
  char inputPPMFilename[ARR_SIZE] = DEFAULT_INPUTPPM_PATH;
  char outputPPMFilename[ARR_SIZE] = DEFAULT_OUTPUTPPM_PATH;

  char databasePath[ARR_SIZE] = DEFAULT_DATABASE_PATH;
  char interestPointPath[ARR_SIZE] = DEFAULT_INTEREST_POINT_PATH;

  int windowWidth = DEFAULT_SUBWINDOW_WIDTH;
  int windowHeight = DEFAULT_SUBWINDOW_HEIGHT;  
  int dbRows = 0;
  int dbCols = 0;

  ioData *ioObject;

  fstream dbstream;  // stream associated with the database
  fstream ipstream;  // stream associated with the interestPoints 2d array
  
  char imageToCompareFilename[ARR_SIZE] = DEFAULT_IMAGE_TO_COMAPARE_FILE;

  while ((ArgChar=getopt(argc, argv, "123D:d:x:y:X:Y:S:p:f:F:V:C:c:v:i:o:h")) != EOF)
    {
      switch (ArgChar)
	{
	case 'h':
	  fprintf(stderr,"-D <filename of matrix database to be read>\n"); 
	  fprintf(stderr,"-d <filename of matrix database created> (include path if necessary)\n");
	  fprintf(stderr,"-x [width of sub window]\n");
	  fprintf(stderr,"-y [height of sub window]\n");
	  fprintf(stderr,"-X [gaussian width]\n");
	  fprintf(stderr,"-Y [gaussian height]\n");
	  fprintf(stderr,"-S [gaussian variance]\n");
	  fprintf(stderr,"-1 Only use density in vpselect\n");
	  fprintf(stderr,"-2 Only use orientation in vpselect\n");
	  fprintf(stderr,"-3 Use both density and orientation in vpselect\n");
	  fprintf(stderr,"-p [numPoints] (numPoints is the number of interest points to be used)\n");
	  fprintf(stderr,"-f <filename of ppm database> [THIS REQUIRES THE FULL PATH, NOT RELATIVE PATH]\n");
	  fprintf(stderr,"-F <filename of ppm to compare> [THIS REQUIRES THE FULL PATH, NOT RELATIVE PATH]\n");
	  fprintf(stderr,"-i <input ppm filename> (include path if necessary)\n");
	  fprintf(stderr,"-o <output ppm filename> (include path if necessary)\n");
	  fprintf(stderr,"-c <canny executable filename> (include path if necessary)\n");
	  fprintf(stderr,"-C <PGM for canny> (include path if necessary)\n");
	  fprintf(stderr,"-v <vpselect executable filename> (include path if necessary)\n");
	  fprintf(stderr,"-V <input file for vpselect> (include path if necessary)\n");
	  fprintf(stderr,"-h: gives this message\n");
	  exit(0);
	case 'D':
	  strcpy(databasePath,optarg);
	  haveDatabase = 1;
	  break;
	case 'd':
	  strcpy(databasePath,optarg);
	  break;
	case 'F':
	  strcpy(imageToCompareFilename,optarg);
	  readTEST=1;
	  break;
	case 'x':
	  windowWidth = atoi(optarg);
	  break;
	case 'y':
	  windowHeight = atoi(optarg);
	  break;
	case 'X':
	  gaussianWidth = atoi(optarg);
	  break;
	case 'Y':
	  gaussianHeight = atoi(optarg);
	  break;
	case 'S':
	  gaussianVariance = atoi(optarg);
	  break;
	case '1':
	  densityOnly = 1;
	  orientationOnly = 0;
	  densityAndOrientation = 0;
	  break;
	case '2':
	  densityOnly = 0;
	  orientationOnly = 1;
	  densityAndOrientation = 0;
	  break;
	case '3':
	  densityOnly = 0;
	  orientationOnly = 0;
	  densityAndOrientation = 1;
	  break;
	case 'p':
	  numPoints = atoi(optarg);
	  break;
	case 'f':
	  strcpy(PPMStoRead,optarg);
	  readPPMS = 1;
	  break;
	case 'i':
	  strcpy(inputPPMFilename,optarg);
	  break;
	case 'v':
	  strcpy(vpselPath,optarg);
	  break;
	case 'V':
	  strcpy(vpselInputPath,optarg);
	  break;
	case 'c':
	  strcpy(cannyPath,optarg);
	  break;
	case 'C':
	  strcpy(cannyInputPath,optarg);
	  break;
	case 'o':
	  strcpy(outputPPMFilename,optarg);
	  break;
	default:
	  fprintf(stderr,"Unrecognised option, try -h\n");
	  exit(0);
	}
    }

  matrix *database;
  matrix *testImage;
  matrix *databaseCopy;


  if (haveDatabase)
    {
      //  Create a read stream for storing the database
      dbstream.open(databasePath,ios::in | ios::nocreate);
      if (!dbstream)
	fprintf(stderr,"Problems opening %s\n",databasePath);
      
      //  Create a read stream for storing the interest points
      ipstream.open(interestPointPath,ios::in | ios::nocreate);
      if (!ipstream)
	fprintf(stderr,"Problems opening %s\n",interestPointPath);
      
    }
  else
    {
      //  Create a write stream for storing the database
      dbstream.open(databasePath,ios::out | ios::trunc);
      if (!dbstream)
	fprintf(stderr,"Problems opening %s\n",databasePath);
      
      //  Create a write stream for storing the interest points
      ipstream.open(interestPointPath,ios::out | ios::trunc);
      if (!ipstream)
	fprintf(stderr,"Problems opening %s\n",interestPointPath);
      
    }


  printf("The number of view points is: %d\n",numPoints);
  printf("The sub window width is: %d\n",windowWidth);
  printf("The sub window heightis: %d\n",windowHeight);
  printf("The guassian width is: %d\n",gaussianWidth);
  printf("The guassian height is: %d\n",gaussianWidth);
  printf("The gaussian variance is: %d\n",gaussianVariance);
  if (densityOnly)
    printf("vpselect using density only\n");
  if (orientationOnly)
    printf("vpselect using orientation only\n");
  if (densityAndOrientation)
    printf("vpselect using density and orientation\n");



  if (!haveDatabase)  // build the database
    {
      if (readPPMS)
	{
	  database = 
	    setupDatabase(PPMStoRead,cannyPath,vpselPath,vpselInputPath,
			  numPoints,windowWidth,windowHeight,gaussianWidth,
			  gaussianHeight,gaussianVariance);
	  dbstream << "NumRows " << (*database).dim1() << "\n";
	  dbstream << "NumCols " << (*database).dim2() << "\n";
	  dbstream << "NumPoints " << numPoints << "\n";
	  dbstream << "windowWidth " << windowWidth << "\n";
	  dbstream << "windowHeight " << windowHeight << "\n";
	  // (*database).print(dbstream); // save database to disk
	} 
      else
	{
	  printf("Please enter the FULL path of file containing ppm's to build the database with:\n");
	  scanf("%s",PPMStoRead);
	  database = 
	    setupDatabase(PPMStoRead,cannyPath,vpselPath,vpselInputPath,
			  numPoints,windowWidth,windowHeight,gaussianWidth,
			  gaussianHeight,gaussianVariance);
	  dbstream << "NumRows " << (*database).dim1() << "\n";
	  dbstream << "NumCols " << (*database).dim2() << "\n";
	  dbstream << "NumPoints " << numPoints << "\n";
	  dbstream << "windowWidth " << windowWidth << "\n";
	  dbstream << "windowHeight " << windowHeight << "\n";
	  // (*database).print(dbstream); // save database to disk
	}


      for (int i=0; i<numberOfImages; i++)
	for (int j=0; j<numPoints; j++)
	  {
	    printf("(%d,%d):",i,j);
	    globalArray[i][j].printDetail();
	    printf("\n");
	  }

      
      ioObject = new ioData(*database, globalArray, numberOfImages, numPoints);
      (*ioObject).dbWrite(dbstream);
      (*ioObject).arrWrite(ipstream);
      printf("Database created and stored in %s\n",databasePath);
      printf("Interest Points created and stored in %s\n",interestPointPath);
      dbstream.close();
      ipstream.close();
    }
  else // have database and  have interestPoints array
    {
      // get the dimensions of the matrix
      char temp[ARR_SIZE];
      dbstream >> temp >> dbRows;
      dbstream >> temp >> dbCols;
      dbstream >> temp >> numPoints;
      dbstream >> temp >> windowWidth;
      dbstream >> temp >> windowHeight;
      numberOfImages = dbRows/numPoints;       // Initialise the number of image (global)

      // ioObject contains all the database and also the interest points
      ioObject = new ioData(dbRows, dbCols, numberOfImages, numPoints,
			    dbstream, ipstream);
      dbstream.close();
      ipstream.close();

      databaseCopy = (*ioObject).getDatabase();
      
      globalArray = new interestPoint*[numberOfImages];
      for (int i=0; i<numberOfImages; i++)
	globalArray[i] = new interestPoint[numPoints];
      
      
      for (int i=0; i<numberOfImages; i++)
	for (int j=0; j<numPoints; j++)
	  {
	    globalArray[i][j] = (*ioObject).getInterestpoint(i,j);
	    printf("(%d,%d):",i,j);
	    globalArray[i][j].printDetail();
	    printf("\n");
	  }

      /*
      databaseCopy = new matrix(dbRows, dbCols);
      (*databaseCopy).read(dbstream); // load database from disk
      cout << "Database loaded from " << databasePath << "\n";
      */     
    }


  // ioObject should have all the necessary information now

  while (1)
    {
      if (readTEST)
	{
	  testImage = 
	    processTestImage(imageToCompareFilename,cannyPath,vpselPath,vpselInputPath,
			     numPoints,windowWidth,windowHeight,gaussianWidth,
			     gaussianHeight,gaussianVariance);
	  if (haveDatabase)
	    runPCA(databaseCopy,testImage);
	  else
	    runPCA(database,testImage);
	  break;
	}
      else
	{
	  printf("Please enter the FULL path of ppm file which you wish to compare\n");
	  scanf("%s",imageToCompareFilename);
	  testImage = 
	    processTestImage(imageToCompareFilename,cannyPath,vpselPath,vpselInputPath,
			     numPoints,windowWidth,windowHeight,gaussianWidth,
			     gaussianHeight,gaussianVariance);
	}   
      
      if (haveDatabase)
	runPCA(databaseCopy,testImage);
      else 
	runPCA(database,testImage);
    }


  delete ioObject;
  
  for (int i=0; i<numberOfImages; i++)
    {
      delete[] globalArray[i];
      globalArray[i] = NULL;
    }
  
  delete[] globalArray;
}








